﻿//--------------------------------------------------------------------------
// 
//  Copyright (c) Chili Software.  All rights reserved. 
// 
//  File: FirebirdSqlStorage.cs
//
//  Description: Provides a storage class for Firebird.
// 
//--------------------------------------------------------------------------

using System;
using System.Reflection;
using System.Data;
using FirebirdSql.Data.FirebirdClient;
using Chili.Opf3.Query;

namespace Chili.Opf3.Storages.Firebird
{
	/// <summary>Storage class that encapsulates a pyhsical PervasiveSQL database.</summary>
	/// <remarks>
	/// 	<para>This class encapsulates a physical PervasiveSQL database. It is used to access to
	///     an PervasiveSQL database that is installed somewhere on your or on a remote machine. The storage
	///     implements the basic interfaces, such as
	///     <see cref="Opf3.Storages.IStorage">IStorage</see> for the basic storage IO
	///     functinality and
	///     <see cref="Opf3.Storages.ITransactionStorage">ITransactionStorage</see> for the
	///     transaction management.</para>
	/// 	<para><b>Attention:</b> You should not access the methods of the storage directly
	///     but instead register the storage with your instances of the
	///     <see cref="Opf3.ObjectContext">ObjectContext</see>. The
	///     <see cref="Opf3.ObjectContext">ObjectContext</see> accesses the methods internally
	///     to get the objects from the storage.</para>
	/// </remarks>
	/// <example>
	///     The following example shows how you should use the VistaDB storage in your
	///     applications. 
	///     <code lang="CS">
	/// // ... Other code
	///  
	/// // Creates a new ObjectContext that uses a Firebird database as storage.
	/// // You should only pass an instance of the storage to your ObjectContext.
	/// // That's it! Nothing more. The methods of the storage are accessed
	/// // by the context internally.
	/// ObjectContext context = new ObjectContext(
	///     new FirebirdStorage("... some ConnectionString ..."));
	///  
	/// // ... Other code
	///     </code>
	/// </example>
	public class FirebirdStorage : SqlStorageBase, ICustomDataTypesStorage
	{
		private const int _constraintError = 335544558;
		private const int _foreignKeyViolation = 335544466;
		private const int _primaryKeyViolation = 335544665;
		private const int _indexViolation = 335544349;
		private string _generatorformat = "{0}Generator";

		/// <summary>
		/// Creates a new instance of the <see cref="FirebirdStorage">FirebirdStorage Class</see>.
		/// </summary>
		/// <param name="connectionString">
		/// Specifies a string that contains the information (for ADO.NET) to connect to storage.
		/// </param>
		/// <exception cref="ArgumentNullException">ConnectionString is null.</exception>
		public FirebirdStorage(string connectionString) : base(connectionString)
		{
			base.StorageCommandBuilders.Add(new FirebirdCommandBuilder());
			base.StorageCommandBuilders.Add(new SqlQueryCommandBuilder());
		}

		/// <summary>
		/// Specifies the format string for the generator for auto increment values.
		/// </summary>
		/// <remarks>
		/// The GeneratorFormat string allows you to specify a format string for the generator names.
		/// The default value for a generator is {0}Generator. {0} stands for the name of the entity
		/// for which this generator is generated. You may also add {1} which is then the name of 
		/// the field for which the generator is created. 
		/// </remarks>
		/// <example>
		/// This example shows the sequence format strings for an entity named "USER". The generator is generated
		/// for the field "ID".
		/// <code>
		///		{0}Generator = USERGenerator
		///     {0}_s1 = USER_s1
		///     {0}{1}Generator = USERIDGenerator
		///     {0}_s{1} = USER_sID
		/// </code>
		/// </example>
		public string GeneratorFormat
		{
			get { return _generatorformat; }
			set
			{
				if (value == null)
					throw new ArgumentNullException("value", "Value is null.");
				_generatorformat = value;
			}
		}

		#region Overridden Members

		/// <summary>
		/// Executes a command on the storage that does not expect any results. It the method
		/// returns an argument it is used by the framework to set the
		/// <see cref="Opf3.FieldAttribute.AutoNumber">AutoNumber</see> property of the persistent
		/// object.
		/// </summary>
		/// <param name="query">
		/// 	<see cref="Opf3.Query.IQuery">IQuery</see> that is compiled and executed on the
		/// storage.
		/// </param>
		/// <exception cref="Opf3.Storages.StorageConstraintException">A constraint of the storage 
		/// has been violated.</exception>
		/// <exception cref="Opf3.Storages.ForeignKeyViolationException">A foreign key in the storage 
		/// has been violated.</exception>
		/// <exception cref="Opf3.Storages.PrimaryKeyViolationException">A duplicate primary key 
		/// is inserted in the storage.</exception>
		/// <exception cref="Opf3.Storages.PrimaryKeyViolationException">An index has been violated.</exception>
		/// <exception cref="Opf3.Storages.StorageException">Any other unexpected storage exception.</exception>
		public override object ExecuteCommand(IQuery query)
		{
			try
			{
				object autoValue = null;

				CompiledQuery compiledQuery = query.ToCompiledQuery(this);
				Tracing.Current.TraceStorageIO(compiledQuery);

				// Prepare the command for execution.
				using (IDbCommand command = this.PrepareCommand(compiledQuery))
				{
					command.ExecuteNonQuery();

					if (query.QueryType == QueryTypes.Insert)
					{
						// Check if the query's parameter contain an autonumber parameter.
						if (compiledQuery.AutoNumberItems != null)
						{
							// Get the value of the auto increment.
							autoValue = compiledQuery.AutoNumberItems[0].Value;

							Tracing.Current.TraceStorageIO("Autonumber value returned: " + autoValue);
						}
					}

					// If no transaction is active we can close the connection after executing the command.
					if (!IsTransactionActive)
						command.Connection.Close();
				}
				return autoValue;
			}
			catch (FbException ex)
			{
				if (ex.Errors[0].Number == _constraintError)
					throw new StorageConstraintException(ex.Message, ex, query);
				else if (ex.Errors[0].Number == _foreignKeyViolation)
					throw new ForeignKeyViolationException(ex.Message, ex, query);
				else if (ex.Errors[0].Number == _primaryKeyViolation)
					throw new PrimaryKeyViolationException(ex.Message, ex, query);
				else if (ex.Errors[0].Number == _indexViolation)
					throw new IndexViolationException(ex.Message, ex, query);
				else
					throw new StorageException(ex.Message, ex, query);
			}
		}

		/// <summary>
		/// Executes and compiles the <see cref="Opf3.Query.IQuery">IQuery</see> and returns
		/// an <see cref="Opf3.ObjectReader{T}">ObjectReader</see> with the resultsset.
		/// </summary>
		/// <param name="context">
		/// 	<see cref="Opf3.ObjectContext">ObjectContext</see> that is connected with the
		/// <see cref="Opf3.ObjectReader{T}">ObjectReader</see>.
		/// </param>
		/// <param name="query">
		/// 	<see cref="Opf3.Query.IQuery">IQuery</see> that is compiled and executed on the
		/// storage.
		/// </param>
		/// <exception cref="Opf3.Storages.StorageException">An unexpected storage exception has happened.</exception>
		public override ObjectReader<T> ExecuteReader<T>(ObjectContext context, IQuery query)
		{
			try
			{
				return base.ExecuteReader<T>(context, query);
			}
			catch (FbException ex)
			{
				throw new StorageException(ex.Message, ex, query);
			}
		}

		/// <summary>Creates a new storage dependent connection object.</summary>
		/// <returns>Returns an instance of a storage dependent connection object.</returns>
		public override IDbConnection CreateConnection()
		{
			FbConnection connection = new FbConnection(this.ConnectionString);
			connection.Open();
			return connection;
		}

		/// <summary>Creates a new storage dependent command object.</summary>
		/// <returns>Returns an instance of a storage dependent command object.</returns>
		public override IDbCommand CreateCommand()
		{
			return new FbCommand();
		}

		/// <summary>Creates and populates a storage dependent parameter with data.</summary>
		/// <returns>
		/// The name of the storage parameter should be set to a valid parameter name (one
		/// that is supported by the storage) while creating the parameter.
		/// </returns>
		/// <param name="index">Index of the storage parameter.</param>
		/// <param name="name">Name of the storage parameter.</param>
		/// <param name="value">Value of the storage parameter.</param>
		/// <param name="type">The type of the value argument. This information is useful if the value is null.</param>
		/// <exception cref="ArgumentNullException">Name is null.</exception>
		public override IDataParameter CreateParameter(int index, ref string name, object value, Type type)
		{
			if (name == null)
				throw new ArgumentNullException("name", "Name is null.");

			// Set the name for the parameter.
			name = "@" + name;

			return new FbParameter(name, value == null ? DBNull.Value : value);
		}

		/// <summary>Transforms the parameter to a valid field.</summary>
		/// <returns>Returns a string that represents a valid field name.</returns>
		/// <remarks>
		/// This method returns a field name that is valid for the current storage. For some
		/// storages it is a better choice to start each field name with a '[' and end with a ']'
		/// (Ms Access) - This method could wrap each field name passed with '[]'.
		/// </remarks>
		/// <param name="name">Name for the field that is transformed.</param>
		/// <exception cref="ArgumentNullException">Name is null.</exception>
		public override string GetValidFieldName(string name)
		{
			if (name == null)
				throw new ArgumentNullException("name", "Name is null.");

			return name;
		}

		/// <summary>
		/// Transforms the name of the entity to a valid entity name.
		/// </summary>
		/// <returns>Returns a string that represents a valid entity name.</returns>
		/// <remarks>
		/// This method returns an entity name that is valid for the current storage. For
		/// some storages it is a better choice to start each entity name with a '[' and end with a
		/// ']' (Ms Access) - This method could wrap each entity name passed with '[]'.
		/// </remarks>
		/// <param name="name">Name of the entity.</param>
		/// <exception cref="ArgumentNullException">Name is null.</exception>
		public override string GetValidEntityName(string name)
		{
			if (name == null)
				throw new ArgumentNullException("name", "Name is null.");

			return name;
		}

		#endregion

		#region ICustomDataTypesStorage Members

		/// <summary>
		/// Allows to convert from the type in the storage to the target type of the object.
		/// </summary>
		/// <param name="value">Value loaded from the storage.</param>
		/// <param name="targetType">The type of the member of the persistent object.</param>
		/// <returns>The value converted to the target type.</returns>
		object ICustomDataTypesStorage.ChangeFromDataType(object value, Type targetType)
		{
			if (value == null)
				return value;

			if (value is int && targetType == typeof(bool))
				return Convert.ToBoolean(value);
//			if (value is short && targetType.IsEnum)
//				return Enum.ToObject(targetType, (short)value);
			return value;
		}

		/// <summary>
		/// Allows to convert from the type of the object to a type that is supported by the storage.
		/// </summary>
		/// <param name="value">Value from the persistent object that is converted to a type supported by the storage.</param>
		/// <returns>The value converted to a type that is supported by the storage.</returns>
		object ICustomDataTypesStorage.ChangeToDataType(object value)
		{
			if (value is bool)
				return Convert.ToInt32((bool)value);
//			if (value != null && value.GetType().IsEnum)
//				return Convert.ToInt16(value);
			return value;
		}

		#endregion
	}
}
