﻿//--------------------------------------------------------------------------
// 
//  Copyright (c) Chili Software.  All rights reserved. 
// 
//  File: PostgreSqlStorage.cs
//
//  Description: Provides a storage class for PostgreSQL.
// 
//--------------------------------------------------------------------------

using System;
using System.Data;
using Chili.Opf3.Query;
using Npgsql;

namespace Chili.Opf3.Storages.PostgreSql
{
	/// <summary>Storage class that encapsulates a pyhsical PostgreSql database.</summary>
	/// <remarks>
	/// 	<para>This class encapsulates a physical PostgreSql database. It is used to access to
	///     an PostgreSql database that is installed somewhere on your or on a remote machine. The storage
	///     implements the basic interfaces, such as
	///     <see cref="Opf3.Storages.IStorage">IStorage</see> for the basic storage IO
	///     functinality and
	///     <see cref="Opf3.Storages.ITransactionStorage">ITransactionStorage</see> for the
	///     transaction management.</para>
	/// 	<para><b>Attention:</b> You should not access the methods of the storage directly
	///     but instead register the storage with your instances of the
	///     <see cref="Opf3.ObjectContext">ObjectContext</see>. The
	///     <see cref="Opf3.ObjectContext">ObjectContext</see> accesses the methods internally
	///     to get the objects from the storage.</para>
	/// </remarks>
	/// <example>
	///     The following example shows how you should use the VistaDB storage in your
	///     applications. 
	///     <code lang="CS">
	/// // ... Other code
	///  
	/// // Creates a new ObjectContext that uses an VistaDB database as storage.
	/// // You should only pass an instance of the storage to your ObjectContext.
	/// // That's it! Nothing more. The methods of the storage are accessed
	/// // by the context internally.
	/// ObjectContext context = new ObjectContext(
	///     new PostgreSqlStorage("... some ConnectionString ..."));
	///  
	/// // ... Other code
	///     </code>
	/// </example>
	public class PostgreSqlStorage : SqlStorageBase
	{
		private const string _constraintError1 = "23000";
		private const string _constraintError2 = "23501";
		private const string _constraintError3 = "23002";
		private const string _constraintError4 = "23514";
		private const string _foreignKeyViolation = "23503";
		private const string _indexViolation = "23505";
		private string _sequenceFormat = "{0}_{1}_seq";

		/// <summary>
		/// Creates a new instance of the <see cref="PostgreSqlStorage">PostgreSqlStorage Class</see>.
		/// </summary>
		/// <param name="connectionString">Specifies a string that contains the information (for ADO.Net) to connect to storage.</param>
		/// <exception cref="ArgumentNullException">ConnectionString is null.</exception>
		public PostgreSqlStorage(string connectionString) : base(connectionString)
		{
			base.StorageCommandBuilders.Add(new PostgreSqlCommandBuilder());
			base.StorageCommandBuilders.Add(new SqlQueryCommandBuilder());
		}

		/// <summary>
		/// Creates a new instance of the <see cref="PostgreSqlStorage">PostgreSqlStorage Class</see>.
		/// </summary>
		/// <param name="user">Specifies the name of the user that connects to the storage.</param>
		/// <param name="password">Specifies the password of the user that connects to the storage.</param>
		/// <param name="host">Specifies the hostname of the machine where the database is installed.</param>
		/// <param name="database">Specifies the name of the database that you want connect to.</param>
		/// <exception cref="ArgumentNullException">User is null, password is null and database is null.</exception>
		public PostgreSqlStorage(string user, string password, string host, string database)
			: this(string.Format(@"Server={0};User Id={1};Password={2};Database={3}", host, user, password, database))
		{
			if (user == null)
				throw new ArgumentNullException("user", "User is null.");
			if (password == null)
				throw new ArgumentNullException("password", "Password is null.");
			if (database == null)
				throw new ArgumentNullException("database", "Database is null.");
		}

		/// <summary>
		/// Specifies the format string for sequences.
		/// </summary>
		/// <remarks>
		/// The SequenceFormat string allows you to specify a format string for the sequence names.
		/// The default value for a sequence is {0}_{1}_seq (follows the schema of generated sequences with 
		/// pgAdminIII). {0} stands for the name of the entity for which this sequence is generated. {1} 
		/// is then the name of the field for which the sequence is created. 
		/// </remarks>
		/// <example>
		/// This example shows the sequence format strings for an entity named "USER". The sequence is generated
		/// for the field "ID".
		/// <code>
		///		{0}Sequence = USERSequence
		///     {0}_s1 = USER_s1
		///     {0}{1}Sequence = USERIDSequence
		///     {0}_s{1} = USER_sID
		/// </code>
		/// </example>
		public string SequenceFormat
		{
			get { return _sequenceFormat; }
			set
			{
				if (value == null)
					throw new ArgumentNullException("value", "Value is null.");
				_sequenceFormat = value;
			}
		}

		#region Abstract Members

		/// <summary>
		/// Executes a command on the storage that does not expect any results. It the method
		/// returns an argument it is used by the framework to set the
		/// <see cref="Opf3.FieldAttribute.AutoNumber">AutoNumber</see> property of the persistent
		/// object. 
		/// </summary>
		/// <param name="query">
		/// 	<see cref="Opf3.Query.IQuery">IQuery</see> that is compiled and executed on the
		/// storage.
		/// </param>
		/// <exception cref="Opf3.Storages.StorageConstraintException">A constraint of the storage 
		/// has been violated.</exception>
		/// <exception cref="Opf3.Storages.ForeignKeyViolationException">A foreign key in the storage 
		/// has been violated.</exception>
		/// <exception cref="Opf3.Storages.PrimaryKeyViolationException">A duplicate primary key 
		/// is inserted in the storage.</exception>
		/// <exception cref="Opf3.Storages.PrimaryKeyViolationException">An index has been violated.</exception>
		/// <exception cref="Opf3.Storages.StorageException">Any other unexpected storage exception.</exception>
		public override object ExecuteCommand(IQuery query)
		{
			try
			{
				object autoValue = null;
				CompiledQuery compiledQuery = query.ToCompiledQuery(this);
				Tracing.Current.TraceStorageIO(compiledQuery);

				// Prepare the command for execution.
				using (IDbCommand dbCommand = PrepareCommand(compiledQuery))
				{
					dbCommand.ExecuteNonQuery();

					if (query.QueryType == QueryTypes.Insert)
					{
						// Check if the query's parameter contain an autonumber parameter.
						if (compiledQuery.AutoNumberItems != null)
						{
							// Get the first field. This once autonumber is returned.
							FieldAttribute item = compiledQuery.AutoNumberItems[0].FieldAttribute;

							// Create the query executed to get the sequence.
							string sql = string.Format("select * from currval('{0}');", _sequenceFormat);
							sql = string.Format(sql, compiledQuery.AutoNumberItems.PersistentAttribute.Entity, item.FieldName);

							NpgsqlCommand idCommand = new NpgsqlCommand(sql, (NpgsqlConnection)dbCommand.Connection);
							idCommand.Transaction = (NpgsqlTransaction)dbCommand.Transaction;
							autoValue = idCommand.ExecuteScalar();

							Tracing.Current.TraceStorageIO("Autonumber value returned: " + autoValue);
						}
					}

					// If no transaction is active we can close the connection after executing the command.
					if (!IsTransactionActive)
						dbCommand.Connection.Close();
				}
				return autoValue;
			}
			catch (NpgsqlException ex) {
				if (ex.Code == _constraintError1 || ex.Code == _constraintError2 ||
					ex.Code == _constraintError3 || ex.Code == _constraintError4)
				{
					throw new StorageConstraintException(ex.Message, ex, query);
				}
				if (ex.Code == _foreignKeyViolation)
					throw new ForeignKeyViolationException(ex.Message, ex, query);
				if (ex.Code == _indexViolation)
					throw new IndexViolationException(ex.Message, ex, query);
				throw new StorageException(ex.Message, ex, query);
			}
		}

		/// <summary>
		/// Executes and compiles the <see cref="Opf3.Query.IQuery">IQuery</see> and returns
		/// an <see cref="Opf3.ObjectReader{T}">ObjectReader</see> with the resultsset.
		/// </summary>
		/// <param name="context">
		/// 	<see cref="Opf3.ObjectContext">ObjectContext</see> that is connected with the
		/// <see cref="Opf3.ObjectReader{T}">ObjectReader</see>.
		/// </param>
		/// <param name="query">
		/// 	<see cref="Opf3.Query.IQuery">IQuery</see> that is compiled and executed on the
		/// storage.
		/// </param>
		/// <exception cref="Opf3.Storages.StorageException">An unexpected storage exception has happened.</exception>
		public override ObjectReader<T> ExecuteReader<T>(ObjectContext context, IQuery query)
		{
			try
			{
				return base.ExecuteReader<T>(context, query);
			}
			catch (NpgsqlException ex)
			{
				throw new StorageException(ex.Message, ex, query);
			}
		}

		/// <summary>Creates a new storage dependent connection object.</summary>
		/// <returns>Returns an instance of a storage dependent connection object.</returns>
		public override IDbConnection CreateConnection()
		{
			NpgsqlConnection connection = new NpgsqlConnection(ConnectionString);
			connection.Open();
			return connection;
		}

		/// <summary>Creates a new storage dependent command object.</summary>
		/// <returns>Returns an instance of a storage dependent command object.</returns>
		public override IDbCommand CreateCommand()
		{
			return new NpgsqlCommand();
		}

		/// <summary>Creates and populates a storage dependent parameter with data.</summary>
		/// <returns>
		/// The name of the storage parameter should be set to a valid parameter name (one
		/// that is supported by the storage) while creating the parameter.
		/// </returns>
		/// <param name="index">Index of the storage parameter.</param>
		/// <param name="name">Name of the storage parameter.</param>
		/// <param name="value">Value of the storage parameter.</param>
		/// <param name="type">The type of the value argument. This information is useful if the value is null.</param>
		/// <exception cref="ArgumentNullException">Name is null.</exception>
		public override IDataParameter CreateParameter(int index, ref string name, object value, Type type)
		{
			if (name == null)
				throw new ArgumentNullException("name", "Name is null.");

			// Set the name for the parameter.
			name = ":" + name;

			return new NpgsqlParameter(name, value ?? DBNull.Value);
		}

		/// <summary>Transforms the parameter to a valid field.</summary>
		/// <returns>Returns a string that represents a valid field name.</returns>
		/// <remarks>
		/// This method returns a field name that is valid for the current storage. For some
		/// storages it is a better choice to start each field name with a '[' and end with a ']'
		/// (Ms Access) - This method could wrap each field name passed with '[]'.
		/// </remarks>
		/// <param name="name">Name for the field that is transformed.</param>
		/// <exception cref="ArgumentNullException">Name is null.</exception>
		public override string GetValidFieldName(string name)
		{
			if (name == null)
				throw new ArgumentNullException("name", "Name is null.");

			return ProperPostgreSQLName(name);
		}

		/// <summary>
		/// Transforms the name of the entity to a valid entity name.
		/// </summary>
		/// <returns>Returns a string that represents a valid entity name.</returns>
		/// <remarks>
		/// This method returns an entity name that is valid for the current storage. For
		/// some storages it is a better choice to start each entity name with a '[' and end with a
		/// ']' (Ms Access) - This method could wrap each entity name passed with '[]'.
		/// </remarks>
		/// <param name="name">Name of the entity.</param>
		/// <exception cref="ArgumentNullException">Name is null.</exception>
		public override string GetValidEntityName(string name)
		{
			if (name == null)
				throw new ArgumentNullException("name", "Name is null.");

			return ProperPostgreSQLName(name);
		}

		#endregion

		/// <summary>
		/// Encloses the field or entity name in quotes if it contains uppercase characters.
		/// </summary>
		/// <param name="name">The field or entity name that is transformed.</param>
		/// <returns>Quoted mixed-case name, or the original lowercase name.</returns>
		private string ProperPostgreSQLName(string name)
		{
			bool mixedCase = name.ToLower() != name;
			return mixedCase ? string.Format("\"{0}\"", name) : name;
		}
	}
}
