﻿//--------------------------------------------------------------------------
// 
//  Copyright (c) Chili Software.  All rights reserved. 
// 
//  File: SqliteStorage.cs
//
//  Description: Provides a storage class for Sqlite.
// 
//--------------------------------------------------------------------------

using System;
using System.Reflection;
using System.Data;
using System.Data.Common;
using System.Data.SqlClient;
using System.Data.SQLite;
using Chili.Opf3.Query;
using Chili.Opf3.Query.OPathParser;

namespace Chili.Opf3.Storages.Sqlite
{
	/// <summary>Storage class that encapsulates a pyhsical Sqlite database.</summary>
	/// <remarks>
	/// 	<para>This class encapsulates a physical Sqlite database. It is used to access to
	///     an Sqlite database that is installed somewhere on your or on a remote machine. The storage
	///     implements the basic interfaces, such as
	///     <see cref="Opf3.Storages.IStorage">IStorage</see> for the basic storage IO
	///     functinality and
	///     <see cref="Opf3.Storages.ITransactionStorage">ITransactionStorage</see> for the
	///     transaction management.</para>
	/// 	<para><b>Attention:</b> You should not access the methods of the storage directly
	///     but instead register the storage with your instances of the
	///     <see cref="Opf3.ObjectContext">ObjectContext</see>. The
	///     <see cref="Opf3.ObjectContext">ObjectContext</see> accesses the methods internally
	///     to get the objects from the storage.</para>
	/// </remarks>
	/// <example>
	///     The following example shows how you should use the VistaDB storage in your
	///     applications. 
	///     <code lang="CS">
	/// // ... Other code
	///  
	/// // Creates a new ObjectContext that uses a My SQL database as storage.
	/// // You should only pass an instance of the storage to your ObjectContext.
	/// // That's it! Nothing more. The methods of the storage are accessed
	/// // by the context internally.
	/// ObjectContext context = new ObjectContext(
	///     new SqliteStorage("server", "user", "password", "database"));
	///  
	/// // ... Other code
	///     </code>
	/// </example>
	public class SqliteStorage : SqlStorageBase, ICustomDataTypesStorage
	{
		/// <summary>
		/// Creates a new instance of the <see cref="SqliteStorage">SqliteStorage Class</see>.
		/// </summary>
		/// <param name="connectionString">Specifies a string that contains the information (for ADO.NET) to connect to storage.</param>
		/// <exception cref="ArgumentNullException">ConnectionString is null.</exception>
		public SqliteStorage(string connectionString) : base(connectionString)
		{
			base.StorageCommandBuilders.Add(new SqliteCommandBuilder());
			base.StorageCommandBuilders.Add(new SqlQueryCommandBuilder());
		}

		/// <summary>
        /// Creates a new instance of the <see cref="SqliteStorage">SqliteStorage Class</see>.
		/// </summary>
		/// <param name="user">Specifies the name of the user that connects to the storage.</param>
		/// <param name="password">Specifies the password of the user that connects to the storage.</param>
		/// <param name="server">Specifies the server where the database is installed.</param>
		/// <param name="database">Specifies the name of the database that you want connect to.</param>
		/// <exception cref="ArgumentNullException">FileName is null.</exception>
        public SqliteStorage(string user, string password, string server, string database)
			: this(string.Format(@"Server={0};UserId={1};Password={2};Database={3}", server, user, password, database))
		{
			if (user == null)
				throw new ArgumentNullException("user", "User is null.");
			if (password == null)
				throw new ArgumentNullException("password", "Password is null.");
			if (database == null)
				throw new ArgumentNullException("database", "Database is null.");
			if (server == null)
				throw new ArgumentNullException("server", "Server is null.");
		}

		#region Overridden Member

		/// <summary>
		/// Executes a command on the storage that does not expect any results. It the method
		/// returns an argument it is used by the framework to set the
		/// <see cref="Opf3.FieldAttribute.AutoNumber">AutoNumber</see> property of the persistent
		/// object.
		/// </summary>
		/// <param name="query">
		/// 	<see cref="Opf3.Query.IQuery">IQuery</see> that is compiled and executed on the
		/// storage.
		/// </param>
		/// <exception cref="Opf3.Storages.StorageConstraintException">A constraint of the storage 
		/// has been violated.</exception>
		/// <exception cref="Opf3.Storages.ForeignKeyViolationException">A foreign key in the storage 
		/// has been violated.</exception>
		/// <exception cref="Opf3.Storages.PrimaryKeyViolationException">A duplicate primary key 
		/// is inserted in the storage.</exception>
		/// <exception cref="Opf3.Storages.PrimaryKeyViolationException">An index has been violated.</exception>
		/// <exception cref="Opf3.Storages.StorageException">Any other unexpected storage exception.</exception>
		public override object ExecuteCommand(IQuery query)
		{
			if (query == null)
				throw new ArgumentNullException("query", "Query is null.");

			try
			{
				object autoValue = null;
				CompiledQuery compiledQuery = query.ToCompiledQuery(this);
				Tracing.Current.TraceStorageIO(compiledQuery);

				// Prepare the command for execution.
				using (IDbCommand dbCommand = this.PrepareCommand(compiledQuery))
				{
					dbCommand.ExecuteNonQuery();

					if (query.QueryType == QueryTypes.Insert)
					{
						// Check if the query's parameter contain an autonumber parameter.
						if (compiledQuery.AutoNumberItems != null)
						{
                            // last_insert_rowid() taken from: http://sqlite.phxsoftware.com/forums/p/1022/4424.aspx#4424
                            SQLiteCommand idCommand = new SQLiteCommand("SELECT last_insert_rowid();", (SQLiteConnection)dbCommand.Connection);
							idCommand.Transaction = (SQLiteTransaction)dbCommand.Transaction;
							autoValue = idCommand.ExecuteScalar();

							Tracing.Current.TraceStorageIO("Autonumber value returned: " + autoValue);
						}
					}
					// If no transaction is active we can close the connection after executing the command.
					if (!IsTransactionActive)
						dbCommand.Connection.Close();
				}
				return autoValue;
			}
			catch (Exception ex)
			{
                throw new StorageException(ex.Message, ex, query);
			}
		}

		/// <summary>
		/// Executes and compiles the <see cref="Opf3.Query.IQuery">IQuery</see> and returns
		/// an <see cref="Opf3.ObjectReader{T}">ObjectReader</see> with the resultsset.
		/// </summary>
		/// <param name="context">
		/// 	<see cref="Opf3.ObjectContext">ObjectContext</see> that is connected with the
		/// <see cref="Opf3.ObjectReader{T}">ObjectReader</see>.
		/// </param>
		/// <param name="query">
		/// 	<see cref="Opf3.Query.IQuery">IQuery</see> that is compiled and executed on the
		/// storage.
		/// </param>
		/// <exception cref="Opf3.Storages.StorageException">An unexpected storage exception has happened.</exception>
		public override ObjectReader<T> ExecuteReader<T>(ObjectContext context, IQuery query)
		{
			try
			{
				return base.ExecuteReader<T>(context, query);
			}
			catch (Exception ex)
			{
				throw new StorageException(ex.Message, ex, query);
			}
		}

		/// <summary>
		/// Routine that prepares the command. Associates the command with a connection and
		/// populates it with data from the <see cref="Opf3.Query.IQuery">IQuery</see>.
		/// </summary>
		/// <param name="query"><see cref="Opf3.Query.IQuery">IQuery</see> that is compiled and executed on the storage.</param>
		protected override IDbCommand PrepareCommand(CompiledQuery query)
		{
			// Get the transaction connection.
			IDbConnection connection = (this.ActiveTransaction != null) ? ((SqlTransactionBase)this.ActiveTransaction).InnerTransaction.Connection : CreateConnection();

			// Create a new command object.
			IDbCommand command = query.Command;
			command.Connection = connection;
			if (this.ActiveTransaction != null)
				command.Transaction = ((SqlTransactionBase)this.ActiveTransaction).InnerTransaction;

			// Throw an exception if the connection is null!
			if (command.Connection == null)
			{
				throw new InvalidOperationException(string.Format("Connection for command is null. Transaction state: {0}",
					((SqlTransactionBase)this.ActiveTransaction).InnerTransaction));
			}

			return command;
		}

		/// <summary>Creates a new storage dependent connection object.</summary>
		/// <returns>Returns an instance of a storage dependent connection object.</returns>
		public override IDbConnection CreateConnection()
		{
            SQLiteConnection connection = new SQLiteConnection(this.ConnectionString);
			connection.Open();
			return connection;
		}

		/// <summary>Creates a new storage dependent command object.</summary>
		/// <returns>Returns an instance of a storage dependent command object.</returns>
		public override IDbCommand CreateCommand()
		{
			return new SQLiteCommand();
		}

		/// <summary>Creates and populates a storage dependent parameter with data.</summary>
		/// <returns>
		/// The name of the storage parameter should be set to a valid parameter name (one
		/// that is supported by the storage) while creating the parameter.
		/// </returns>
		/// <param name="index">Index of the storage parameter.</param>
		/// <param name="name">Name of the storage parameter.</param>
		/// <param name="value">Value of the storage parameter.</param>
		/// <param name="type">The type of the value argument. This information is useful if the value is null.</param>
		/// <exception cref="ArgumentNullException">Name is null.</exception>
		public override IDataParameter CreateParameter(int index, ref string name, object value, Type type)
		{
			if (name == null)
				throw new ArgumentNullException("name", "Name is null.");

			// Set the name for the parameter.
			name = "@" + name;

			return new SQLiteParameter(name, value == null ? DBNull.Value : value);
		}

		/// <summary>Transforms the parameter to a valid field.</summary>
		/// <returns>Returns a string that represents a valid field name.</returns>
		/// <remarks>
		/// This method returns a field name that is valid for the current storage. For some
		/// storages it is a better choice to start each field name with a '[' and end with a ']'
		/// (Ms Access) - This method could wrap each field name passed with '[]'.
		/// </remarks>
		/// <param name="name">Name for the field that is transformed.</param>
		/// <exception cref="ArgumentNullException">Name is null.</exception>
		public override string GetValidFieldName(string name)
		{
			if (name == null)
				throw new ArgumentNullException("name", "Name is null.");

			return name;
		}

		/// <summary>
		/// Transforms the name of the entity to a valid entity name.
		/// </summary>
		/// <returns>Returns a string that represents a valid entity name.</returns>
		/// <remarks>
		/// This method returns an entity name that is valid for the current storage. For
		/// some storages it is a better choice to start each entity name with a '[' and end with a
		/// ']' (Ms Access) - This method could wrap each entity name passed with '[]'.
		/// </remarks>
		/// <param name="name">Name of the entity.</param>
		/// <exception cref="ArgumentNullException">Name is null.</exception>
		public override string GetValidEntityName(string name)
		{
			if (name == null)
				throw new ArgumentNullException("name", "Name is null.");

			return name;
		}

		#endregion

		#region ICustomDataTypesStorage Members

		/// <summary>
		/// Allows to convert from the type in the storage to the target type of the object.
		/// </summary>
		/// <param name="value">Value loaded from the storage.</param>
		/// <param name="targetType">The type of the member of the persistent object.</param>
		/// <returns>The value converted to the target type.</returns>
		object ICustomDataTypesStorage.ChangeFromDataType(object value, Type targetType)
		{
			if (targetType == typeof(bool) && value is sbyte)
				return Convert.ToBoolean(value);
			return value;
		}

		/// <summary>
		/// Allows to convert from the type of the object to a type that is supported by the storage.
		/// </summary>
		/// <param name="value">Value from the persistent object that is converted to a type supported by the storage.</param>
		/// <returns>The value converted to a type that is supported by the storage.</returns>
		object ICustomDataTypesStorage.ChangeToDataType(object value)
		{
			if (value is bool)
				return Convert.ToSByte(value);
			return value;
		}

		#endregion
	}
}
