#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.Text;
using Chili.Opf3.Wizard.Shared;
using Chili.Opf3.Wizard.Shared.Storages;
using Chili.Opf3.Wizard.CSharp.NUnitUnitTest.Properties;

#endregion

namespace Chili.Opf3.Wizard.CSharp.NUnitUnitTest
{
	/// <summary>
	/// This class generates the IO test for the persistent class.
	/// </summary>
	internal sealed class PersistentIOTestCreator
	{
		public const string OPENING_BRACKETS = "{";
		public const string CLOSING_BRACKETS = "}";
		private Template _template;
		private UnitTest _test;
		private Project _project;
		private StateBag _stateBag;

		/// <summary>
		/// Creates a new instance of the <see cref="PersistentIOTestCreator">PersistentIOTestCreator Class</see>.
		/// </summary>
		/// <param name="template">The template for the unit test creator.</param>
		/// <param name="project">The project of the wizard.</param>
		/// <param name="test">The current test.</param>
		/// <param name="stateBag">The state bag containing the settings of the plug-in.</param>
		public PersistentIOTestCreator(Template template, UnitTest test, Project project, StateBag stateBag)
		{
			if (template == null)
				throw new ArgumentNullException("template", "Template is null.");
			if (project == null)
				throw new ArgumentNullException("project", "Project is null.");
			if (test == null)
				throw new ArgumentNullException("test", "Test is null.");
			if (stateBag == null)
				throw new ArgumentNullException("stateBag", "StateBag is null.");

			_template = template;
			_test = test;
			_project = project;
			_stateBag = stateBag;
		}

		/// <summary>
		/// Creates the persistent IO test.
		/// </summary>
		/// <returns></returns>
		public string CreatePersistentIOTest()
		{
			Field firstField = GetFirstField();
			string propertyName = "??";
			if (firstField != null)
				propertyName = firstField.PropertyName;

			return string.Format(_template.GetValue("//persistentIOTest"), OPENING_BRACKETS, CLOSING_BRACKETS, _test.Table.PersistentName,
				CreateLoadCondition(), CreateAssertAreEqual("//assertAreEqual1"), propertyName, GetMethodCall(firstField),
				CreateAssertAreEqual("//assertAreEqual2"));
		}

		/// <summary>
		/// Returns the first valid field that is no identifier and generated.
		/// </summary>
		/// <returns></returns>
		private Field GetFirstField()
		{
			foreach (Field field in _test.Table.Fields)
			{
				// Skip fields that are not created, autonumber or foreign key in a relation.
				if (!field.Generate || field.Identifier || UnitTestCreator.IsForeignKeyInRelation(field, _test, _project))
					continue;
				return field;
			}
			return null;
		}

		/// <summary>
		/// Creates the assert are equal for the comparisons in the test.
		/// </summary>
		/// <param name="templateKey">The key in the template to load the item from the xml.</param>
		private object CreateAssertAreEqual(string templateKey)
		{
			string result = string.Empty;

			foreach (Field field in _test.Table.Fields)
			{
				// Skip fields that are not created, autonumber or foreign key in a relation.
				if (!field.Generate || UnitTestCreator.IsForeignKeyInRelation(field, _test, _project))
					continue;

				if (result.Length > 0)
					result += "\r\n";

				result += string.Format(_template.GetValue(templateKey), field.PropertyName, _project.Namespace, _test.Table.PersistentName);
			}

			return result;
		}

		/// <summary>
		/// Creates the load condidition.
		/// </summary>
		/// <returns></returns>
		private string CreateLoadCondition()
		{
			string condition = string.Empty;
			string parameter = string.Empty;

			int index = 0;
			foreach (Field field in _test.Table.Fields)
			{
				if (!field.Generate)
					continue;

				if (field.Identifier)
				{
					if (condition.Length > 0)
						condition += " And ";

					condition += string.Format("{0} = {1}", field.PropertyName, "{" + index.ToString() + "}");
					parameter += string.Format(", target.{0}", field.PropertyName);

					index++;
				}
			}

			return string.Format("\"{0}\"{1}", condition, parameter);
		}

		/// <summary>
		/// Gets the method call for the given field.
		/// </summary>
		/// <param name="field"></param>
		private string GetMethodCall(Field field)
		{
			if (field == null)
				return "null";

			string call = _template.GetValue(string.Format("//methodCalls/call[@type = '{0}']", field.Type));
			if (field.Type == typeof(string))
				return string.Format(call, field.Length.ToString());
			return call;
		}
	}
}
