#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.Text;
using Chili.Opf3.Wizard.Shared;
using Chili.Opf3.Wizard.Shared.Storages;
using Chili.Opf3.Wizard.CSharp.NUnitUnitTest.Properties;

#endregion

namespace Chili.Opf3.Wizard.CSharp.NUnitUnitTest
{
	/// <summary>
	/// This class creates the tests for the relations.
	/// </summary>
	internal sealed class RelationsTestCreator
	{
		public const string OPENING_BRACKETS = "{";
		public const string CLOSING_BRACKETS = "}";
		private Template _template;
		private UnitTest _test;
		private Project _project;
		private StateBag _stateBag;

		/// <summary>
		/// Creates a new instance of the <see cref="RelationsTestCreator">RelationsTestCreator Class</see>.
		/// </summary>
		/// <param name="template">The template for the unit test creator.</param>
		/// <param name="project">The project of the wizard.</param>
		/// <param name="test">The current test.</param>
		/// <param name="stateBag">The state bag containing the settings of the plug-in.</param>
		public RelationsTestCreator(Template template, UnitTest test, Project project, StateBag stateBag)
		{
			if (template == null)
				throw new ArgumentNullException("template", "Template is null.");
			if (project == null)
				throw new ArgumentNullException("project", "Project is null.");
			if (test == null)
				throw new ArgumentNullException("test", "Test is null.");
			if (stateBag == null)
				throw new ArgumentNullException("stateBag", "StateBag is null.");

			_template = template;
			_test = test;
			_project = project;
			_stateBag = stateBag;
		}

		/// <summary>
		/// Creates the relations tests for the persistent.
		/// </summary>
		/// <returns></returns>
		public string CreateRelationsTests()
		{
			string result = string.Empty;

			// Loop over all relations.
			foreach (Relation relation in _project.Relations)
			{
				// Skip relations that should not be generated.
				if (!relation.Generate)
					continue;

				// Check whether we have to create the relation.
				if ((relation.ParentTable.TableName == _test.Table.TableName && relation.CreateInParentPersistent) ||
					(relation.ChildTable.TableName == _test.Table.TableName && relation.CreateInChildPersistent))
				{
					// Create the relation.
					string rel = CreateRelationTest(relation);
					// Skip if null.
					if (rel == null)
						continue;

					if (result.Length > 0)
						result += "\r\n";

					// Add the relation to the relations string.
					result += rel;
				}
			}

			if (result == string.Empty)
				return result;
			return string.Format(_template.GetValue("//relationRegion"), result);
		}

		#region CreateRelationTest

		/// <summary>
		/// Creates the test for a single relation.
		/// </summary>
		/// <param name="relation"></param>
		private string CreateRelationTest(Relation relation)
		{
			// Check which type of relation is given and create the appropriate relation test.
			switch (relation.RelationType)
			{
				case RelationTypes.OneToMany:
					{
						if (relation.ParentTable.TableName == _test.Table.TableName)
							return CreateObjectSetRelationTest(relation);
						else
							return GetObjectHolderRelationTest(relation);
					}
				case RelationTypes.OneToOne:
					return GetObjectHolderRelationTest(relation);
				case RelationTypes.ManyToMany:
					return CreateObjectSetRelationTest(relation);
				default:
					throw new NotSupportedException("Relationtype is not supported.");
			}
		}

		#endregion

		#region GetObjectHolderRelationTest

		/// <summary>
		/// Create a simple ObjectHolder relation test.
		/// </summary>
		/// <param name="relation">The relation object.</param>
		private string GetObjectHolderRelationTest(Relation relation)
		{
			string relationTestName = _stateBag.MethodNameTemplate.Replace(Resources.MethodNameVariable, 
				relation.ParentTable.PersistentName + relation.ChildTable.PersistentName);
			string parentTestName = _stateBag.ClassNameTemplate.Replace(Resources.ClassNameVariable, relation.ParentTable.PersistentName);

			return string.Format(_template.GetValue("//objectHolderTest"), OPENING_BRACKETS, CLOSING_BRACKETS, relationTestName,
				relation.ChildTable.PersistentName, relation.ParentTable.PersistentName, parentTestName, relation.ChildProperty);
		}

		#endregion

		#region CreateObjectSetRelationTest

		/// <summary>
		/// Create a One-To-Many ObjectSetHolder relation t est.
		/// </summary>
		/// <param name="relation">The relation object.</param>
		private string CreateObjectSetRelationTest(Relation relation)
		{
			string relationTestName = _stateBag.MethodNameTemplate.Replace(Resources.MethodNameVariable, 
				relation.ChildTable.PersistentName + relation.ParentTable.PersistentName);
			string childTestName = _stateBag.ClassNameTemplate.Replace(Resources.ClassNameVariable, relation.ChildTable.PersistentName);

			return string.Format(_template.GetValue("//objectSetHolderTest"), OPENING_BRACKETS, CLOSING_BRACKETS, relationTestName,
				_test.Table.PersistentName, relation.ChildTable.PersistentName, childTestName, relation.ParentProperty);
		}

		#endregion
	}
}
