#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.Text;
using Chili.Opf3.Wizard.Shared;
using Chili.Opf3.Wizard.Shared.Storages;

#endregion

namespace Chili.Opf3.Wizard.CSharp.NUnitUnitTest
{
	/// <summary>
	/// This class generates the unit test for a persistent object.
	/// </summary>
	internal sealed class UnitTestCreator
	{
		public const string OPENING_BRACKETS = "{";
		public const string CLOSING_BRACKETS = "}";
		private Template _template;
		private UnitTest _test;
		private Project _project;
		private StateBag _stateBag;

		/// <summary>
		/// Creates a new instance of the <see cref="UnitTestCreator">UnitTestCreator Class</see>.
		/// </summary>
		/// <param name="template">The template for the unit test creator.</param>
		/// <param name="project">The project of the wizard.</param>
		/// <param name="test">The current test.</param>
		/// <param name="stateBag">The state bag containing the settings of the plug-in.</param>
		public UnitTestCreator(Template template, UnitTest test, Project project, StateBag stateBag)
		{
			if (template == null)
				throw new ArgumentNullException("template", "Template is null.");
			if (project == null)
				throw new ArgumentNullException("project", "Project is null.");
			if (test == null)
				throw new ArgumentNullException("test", "Test is null.");
			if (stateBag == null)
				throw new ArgumentNullException("stateBag", "StateBag is null.");

			_template = template;
			_test = test;
			_project = project;
			_stateBag = stateBag;
		}

		/// <summary>
		/// Creates and returns the unit test.
		/// </summary>
		public string CreateUnitTest()
		{
			string testObjectInstance = new TestObjectInstanceCreator(_template, _test, _project, _stateBag).CreateTestObjectInstanceCode();

			// Create the test methods.
			string testMethods = string.Empty;
			if (_stateBag.CreateMethodTests)
				testMethods = new MethodTestsCreator(_template, _test, _project, _stateBag).CreateTestMethods();
			
			// Create the relations tests.
			string relationTests = string.Empty;
			if (_stateBag.CreateRelationTests)
				relationTests = new RelationsTestCreator(_template, _test, _project, _stateBag).CreateRelationsTests();

			string ioTest = string.Empty;
			if (_stateBag.CreatePersistentIOTest)
				ioTest = new PersistentIOTestCreator(_template, _test, _project, _stateBag).CreatePersistentIOTest();

			// Return the compiled template.
			return string.Format(_template.GetValue("//unitTest"), OPENING_BRACKETS, CLOSING_BRACKETS, _stateBag.TestNamespace, 
				_test.Name, testMethods, testObjectInstance, relationTests, _project.Namespace, ioTest, _test.Table.PersistentName);
		}

		#region GetCSharpType

		/// <summary>
		/// Returns the C# type for a given field.
		/// </summary>
		/// <param name="field">The field that's type is returned.</param>
		/// <param name="template">The template in use.</param>
		public static string GetCSharpType(Template template, Field field)
		{
			string type = template.GetValue(string.Format("//types/type[@id = '{0}']", field.Type));
			// Create the type as nullable type if not a string and null is allowed.
			if (IsNullable(field))
				type += "?";

			return type;
		}

		/// <summary>
		/// Returns whether the field is nullable.
		/// </summary>
		/// <param name="field">Field that is checked.</param>
		public static bool IsNullable(Field field)
		{
			return (field.Type != typeof(string) && field.Type != typeof(byte[]) && field.AllowDBNull && !field.Identifier);
		}

		#endregion

		#region IsForeignKeyInRelation

		/// <summary>
		/// Returns whether the field is in a relation.
		/// </summary>
		/// <param name="field">The field that is inspected.</param>
		/// <param name="test">The test that is used to check whether the field is in a foreign key relation ship.</param>
		/// <param name="project">The project that is used to check for the foreign key. The project contains the list of relations.</param>
		public static bool IsForeignKeyInRelation(Field field, UnitTest test, Project project)
		{
			// Loop over all relations.
			foreach (Relation relation in project.Relations)
			{
				// Check if the relation's child table name is the name of the current table.
				if (relation.ChildTable.TableName == test.Table.TableName)
				{
					foreach (RelationItem relationItem in relation.Relations)
					{
						if (relationItem.ChildField.FieldName == field.FieldName)
							return true;
					}
				}

				// We have a many-to-many relation.
				if (relation.WeakTable != null && relation.WeakTable.TableName == test.Table.TableName)
				{
					// Check if in the many-to-many relation one parent field is the same as the fieldname of the field to check.
					foreach (RelationItem relationItem in relation.WeakRelations)
					{
						if (relationItem.ParentField.FieldName == field.FieldName)
							return true;
					}
				}
			}

			return false;
		}

		#endregion
	}
}
