#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.Text;
using Chili.Opf3.Wizard.Shared;
using System.IO;
using System.Xml;
using Chili.Opf3.Wizard.CSharp.NUnitUnitTest.Properties;
using Chili.Opf3.Wizard.Shared.UI;
using Chili.Opf3.Wizard.CSharp.NUnitUnitTest.UI;
using System.Windows.Forms;
using Chili.Opf3.Wizard.Shared.Storages;

#endregion

namespace Chili.Opf3.Wizard.CSharp.NUnitUnitTest
{
	/// <summary>
	/// Creates an unit test (for NUnit) for each persistent object.
	/// </summary>
	public class UnitTestProvider : IObjectsProvider
	{
		private const string UNIT_TEST_ID = "unitTestProvider";
		private const string TEMPLATE_FILE_NAME = "Chili.Opf3.Wizard.CSharp.NUnitUnitTestTemplate.xml";
		private List<Page> _pages = new List<Page>();
		private StateBag _statebag = new StateBag();
		private List<UnitTest> _unitTests;
		private Project _project;

		/// <summary>
		/// Creates a new instance of the <see cref="UnitTestProvider">UnitTestProvider Class.</see>
		/// </summary>
		public UnitTestProvider()
		{
            _pages.Add(new UnitTestPage(this, _statebag));
        }

		#region IObjectsProvider Members

		/// <summary>
		/// Generates the objects form the project structure.
		/// </summary>
		/// <param name="project">The project containing all selected tables and fields of the database.</param>
		public void GenerateObjects(Project project)
		{
			foreach (UnitTest test in _unitTests)
			{
				// Invoke the event generating the object.
				OnGeneratingObject(new GeneratingObjectEventArgs(test.ID));

				try
				{
					// Create the template.
					Template template = new Template(string.Format("{0}\\{1}", Application.StartupPath, TEMPLATE_FILE_NAME));

					// Create the unit test creator that creates one single unit test class.
					UnitTestCreator creator = new UnitTestCreator(template, test, project, _statebag);
					// Get the content of the unit test.
					string content = creator.CreateUnitTest();

					WriteUnitTestFile(project.OutputDirectoy, content, test);

					// Invoke the event to signalize to the wizard that the object has been written.
					OnGeneratedObject(new GeneratedObjectEventArgs(test.ID));
				}
				catch (Exception ex)
				{
					// Invoke the event to tell the wizard that the object couldn't be created.
					OnGeneratedObject(new GeneratedObjectEventArgs(test.ID, ex));
				}
			}
		}

		/// <summary>
		/// Event is fired before the object is generated from the data of the <see cref="Project">Project</see> class.
		/// </summary>
		public event EventHandler<GeneratingObjectEventArgs> GeneratingObject;

		/// <summary>
		/// Event is fired after the object is generated from the data of the <see cref="Project">Project</see> class.
		/// </summary>
		public event EventHandler<GeneratedObjectEventArgs> GeneratedObject;

		/// <summary>
		/// Returns the name for the tool.
		/// </summary>
		public string Name
		{
			get { return Resources.Name; }
		}

		/// <summary>
		/// Returns the description for the tool.
		/// </summary>
		public string Description
		{
			get { return Resources.Description; }
		}

		/// <summary>
		/// Gets the language name supported by this tool. Or null if no explicit language
		/// is supported.
		/// </summary>
		public string LanguageName
		{
			get { return Resources.LanguageName; }
		}

		/// <summary>
		/// Returns the version of the plug-in.
		/// </summary>
		public string Version
		{
			get { return this.GetType().Assembly.GetName().Version.ToString(); }
		}

		/// <summary>
		/// Display the about dialog for the plug-in.
		/// </summary>
		public void ShowAboutDialog()
		{
			MessageBox.Show(string.Format(Resources.AboutText, "\r\n", this.Version), Resources.AboutTitle, MessageBoxButtons.OK, MessageBoxIcon.Information);
		}

        /// <summary>
        /// Allow the provider to build internal data from Storage.
        /// </summary>
        public void ProcessStorage(Project project)
        {
            _project = project;

            // Creates the list of unit tests.
            _unitTests = new List<UnitTest>();

            foreach (Table table in _project.Tables)
            {
                if (table.Generate)
                    _unitTests.Add(new UnitTest(_statebag, table));
            }
        }

        /// <summary>
        /// Deserialize the plug-in state from xml.
        /// </summary>
        public void Deserialize(Project project, XmlDocument xml, XmlNode d)
        {
            // Load the StateBag from the xml.
            _statebag.Deserialize(project,xml, d); 
		}

        /// <summary>
        /// Initialize defaults in a new project.
        /// </summary>
        public void InitializeDefaults(Project project)
        {
        }

        /// <summary>
        /// Serialize the plug-in state in xml.
        /// </summary>
        public void Serialize(Project project, XmlDocument xml, XmlNode d)
        {
            d.AppendChild(_statebag.Serialize(xml));
		}

		/// <summary>
		/// Returns a list of pages displayed by the class.
		/// </summary>
		public List<Page> Pages
		{
			get { return _pages; }
		}

		/// <summary>
		/// Returns the list of objects that are generated.
		/// </summary>
		public List<IGeneratedObject> GetObjectsList()
		{
			List<IGeneratedObject> list = new List<IGeneratedObject>();
			foreach (UnitTest test in _unitTests)
			{
				list.Add(test);
			}
			return list;
		}

		#endregion

		#region WriteUnitTestFile

		/// <summary>
		/// Writes the unit test object to a file in the output directory.
		/// </summary>
		/// <param name="outputDirectory">The output directory.</param>
		/// <param name="content">The unit test's content that is written to the file.</param>
		/// <param name="test">The test object itself.</param>
        private void WriteUnitTestFile(string outputDirectory, string content, UnitTest test)
        {
#if DEMO
			throw new NotSupportedException("Output not supported in Demo.");
#else
            // Get the path for the interface object.
            string path = string.Format("{0}\\NUnit Unit Tests\\{1}", outputDirectory, _statebag.FileNameTemplate.Replace(Resources.FileNameVariable, test.Table.PersistentName));

            // Create the directory if not created earlier.
            Directory.CreateDirectory(Path.GetDirectoryName(path));
            // Tell the wizard that the file couldn't be overwritten.
            if (File.Exists(path))
                throw new NotSupportedException("File already exists. Please delete the existing file frist.");

            // Write the file.
            using (TextWriter writer = File.CreateText(path))
            {
                writer.WriteLine(content);
            }
#endif
        }


		#endregion

		/// <summary>
		/// Invoked to fire the <see cref="GeneratingObject">GeneratingObject</see> event.
		/// </summary>
		/// <param name="e">The event args class containing further information of the event.</param>
		private void OnGeneratingObject(GeneratingObjectEventArgs e)
		{
			if (GeneratingObject != null)
				GeneratingObject(this, e);
		}

		/// <summary>
		/// Invoked to fire the <see cref="GeneratedObject">GeneratedObject</see> event.
		/// </summary>
		/// <param name="e">The event args class containing further information of the event.</param>
		private void OnGeneratedObject(GeneratedObjectEventArgs e)
		{
			if (GeneratedObject != null)
				GeneratedObject(this, e);
		}
	}
}
