#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.Text;
using Chili.Opf3.Wizard.Shared;
using System.Xml;
using Chili.Opf3.Wizard.Shared.Storages;

#endregion

namespace Chili.Opf3.Wizard.CSharp.Searcher
{
	/// <summary>
	/// This class serializes and deserializes a list of interface objects.
	/// </summary>
	internal sealed class SearcherSerializer
	{
		private const string TABLE_REFERENCE = "tableReference";
        private const string FIELD_REFERENCE = "fieldReference";

		#region Serialize

		/// <summary>
        /// Serializes the list of searchers given.
		/// </summary>
        /// <param name="searchers">The searchers that are serialized.</param>
        public XmlElement Serialize(XmlDocument xml, List<Searcher> searchers)
		{
            if (xml == null)
                throw new ArgumentNullException("xml", "Xml is null.");
            if (searchers == null)
                throw new ArgumentNullException("searchers", "Searchers is null.");

            XmlElement root = xml.CreateElement("searchers");

            // Loops over all searchers and serializes them.
            foreach (Searcher s in searchers)
			{
				root.AppendChild(SerializeSearcher(xml, s));
			}

            return root;
        }

        #region SerializeSearcher

        /// <summary>
		/// Serializes one single Searcher.
		/// </summary>
		/// <param name="iface">The interface that is serialized.</param>
        private XmlNode SerializeSearcher(XmlDocument xml, Searcher searcher)
		{
            XmlElement s = xml.CreateElement("searcher");
            s.AppendChild(CreateXmlElement(xml, "generate", searcher.Generate.ToString()));
            s.AppendChild(CreateXmlElement(xml, "searcherName", searcher.SearcherName));
            s.AppendChild(CreateTableReferenceXml(xml, searcher.Table));

            // Creates the xml for the method's of the searcher.
            s.AppendChild(SerializeMethods(xml, searcher.Methods, searcher.Table));

			return s;
		}

        /// <summary>
        /// Creates the xml for the method's searcher.
        /// </summary>
        /// <param name="list">The list containing the method for the searcher.</param>
        /// <param name="table">The table containing the fields for the method.</param>
        private XmlNode SerializeMethods(XmlDocument xml, List<SearcherMethod> methods, Table table)
        {
            XmlElement me = xml.CreateElement("methods");

            foreach (SearcherMethod method in methods)
            {
                XmlElement m = xml.CreateElement("method");
                m.AppendChild(CreateXmlElement(xml, "generate", method.Generate.ToString()));
                m.AppendChild(CreateXmlElement(xml, "name", method.Name));
                m.AppendChild(CreateXmlElement(xml, "returntype", method.ReturnType));

                // Creates the xml for the fields's of the method.
                XmlElement fields = xml.CreateElement("fields");

                foreach (SearcherField field in method.Fields)
                {
                    fields.AppendChild(CreateFieldReferenceXml(xml, table, field));
                }
                m.AppendChild(fields);
                me.AppendChild(m);
            }

            return me;
        }

        /// <summary>
        /// Creates the xml for a field reference.
        /// </summary>
        /// <param name="table">The table of the field.</param>
        /// <param name="field">The field that is referenced.</param>
        private XmlNode CreateFieldReferenceXml(XmlDocument xml, Table table, SearcherField field)
        {
            XmlElement f = xml.CreateElement(FIELD_REFERENCE);
            XmlAttribute attribute = xml.CreateAttribute("table");
            attribute.Value = table.TableName;
            f.Attributes.Append(attribute);

            attribute = xml.CreateAttribute("field");
            attribute.Value = field.Field.FieldName;
            f.Attributes.Append(attribute);

			attribute = xml.CreateAttribute("comparer");
			attribute.Value = field.Comparer.Replace("<", "&lt;").Replace(">", "&gt;");
			f.Attributes.Append(attribute);

            return f;
        }

		#endregion

		#region Utility methods

		/// <summary>
		/// Creates the xml for a table reference.
		/// </summary>
		/// <param name="table">The table that is referenced.</param>
		private XmlNode CreateTableReferenceXml(XmlDocument xml, Table table)
		{
			XmlElement f = xml.CreateElement(TABLE_REFERENCE);
			XmlAttribute attribute = xml.CreateAttribute("table");
			attribute.Value = table.TableName;
			f.Attributes.Append(attribute);

			return f;
		}

		/// <summary>
		/// Creates an xml element with the given name and value.
		/// </summary>
		/// <param name="name">The name of the element.</param>
		/// <param name="value">The value of the element.</param>
		private XmlNode CreateXmlElement(XmlDocument xml, string name, string value)
		{
			XmlElement e = xml.CreateElement(name);
			e.AppendChild(xml.CreateTextNode(value));

			return e;
		}

		#endregion

		#endregion

		#region Deserialize

		/// <summary>
		/// Deserializes the given xml to a list of interfaces.
		/// </summary>
		/// <param name="value">The xml that is deserialized.</param>
        /// <param name="node">The node that is deserialized.</param>
		/// <param name="p">The project reference to dereference the tables.</param>
        public List<Searcher> Deserialize(XmlDocument xml, XmlNode d,Project p)
		{
            List<Searcher> list = new List<Searcher>();

            XmlNodeList nodes = d.SelectNodes("./searchers/searcher");
			foreach (XmlNode node in nodes)
			{
				DeserializeSearcher(node, p, list);
			}

			return list;
		}

		/// <summary>
		/// Deserializes the given node to a searcher.
		/// </summary>
		/// <param name="node">The node that is deserialized.</param>
		/// <param name="p">The project used to dereference list.</param>
		/// <param name="searchers">The list of searchers.</param>
        private void DeserializeSearcher(XmlNode node, Project p, List<Searcher> searchers)
		{
            string tablename = node.SelectSingleNode("./" + TABLE_REFERENCE).Attributes["table"].Value;
            Table t = p.GetTableByName(tablename);
            if (t == null)
                return;

            Searcher s = new Searcher(t);
			s.Generate = bool.Parse(node.SelectSingleNode("./generate").InnerText);
            s.SearcherName = node.SelectSingleNode("./searcherName").InnerText;
            s.Methods = DeserializeMethods(node, p);

            searchers.Add(s);
		}

        /// <summary>
        /// Deserializes the given node to a searcher's methods list.
        /// </summary>
        /// <param name="node">The node that is deserialized.</param>
        /// <param name="p">The project used to dereference list.</param>
        private List<SearcherMethod> DeserializeMethods(XmlNode d, Project p)
        {
            List<SearcherMethod> methods = new List<SearcherMethod>();

            XmlNodeList nodes = d.SelectNodes("./methods/method");
            foreach (XmlNode node in nodes)
            {
                SearcherMethod method = new SearcherMethod();
                method.Generate = bool.Parse(node.SelectSingleNode("./generate").InnerText);
                method.Name = node.SelectSingleNode("./name").InnerText;
                method.ReturnType = node.SelectSingleNode("./returntype").InnerText;
                method.Fields = DeserializeFields(node, p);
                methods.Add(method);
    
            }
            return methods;
        }

        /// <summary>
        /// Deserializes the given node to a method's fields list.
        /// </summary>
        /// <param name="node">The node that is deserialized.</param>
        /// <param name="p">The project used to dereference list.</param>
        private List<SearcherField> DeserializeFields(XmlNode d, Project project)
        {
			List<SearcherField> fields = new List<SearcherField>();

            XmlNodeList nodes = d.SelectNodes("./fields/" + FIELD_REFERENCE);
            foreach (XmlNode node in nodes)
            {
                string t = node.Attributes["table"].Value;
                string f = node.Attributes["field"].Value;
				// This attribute might not be compiled in old projects. In that case fall back to the equals comparer.
				string c = node.Attributes["comparer"] != null ? node.Attributes["comparer"].Value.Replace("&lt;", "<").Replace("&gt;", ">") : "=";

                Field field = project.GetFieldByName(t,f);
                if (field != null)
                    fields.Add(new SearcherField(field, c)); 
            }
            return fields;
        }

		#endregion
	}
}
