#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.Text;
using Chili.Opf3.Wizard.Shared;
using Chili.Opf3.Wizard.Shared.Storages;

#endregion

namespace Chili.Opf3.Wizard.CSharp.Searcher
{
	/// <summary>
	/// This class generates the searcher for one persistent object.
	/// </summary>
	public sealed class SearcherCreator
	{
		public const string OPENING_BRACKETS = "{";
		public const string CLOSING_BRACKETS = "}";

		private Project _project;
		private Searcher _searcher;
		private Template _template;
        private StateBag _statebag;

		/// <summary>
        /// Creates a new instance of the <see cref="SearcherCreator">SearcherCreator Class</see>.
		/// </summary>
		/// <param name="template">The template for the persistent creator.</param>
		/// <param name="project">The project of the wizard.</param>
        /// <param name="searcher">The current seacher that should be cmpiled with the template.</param>
        /// <param name="statebag">The statebag for the ObjectProvider creator.</param>
        public SearcherCreator(Project project, Searcher searcher, Template template, StateBag statebag)
		{
			if (template == null)
				throw new ArgumentNullException("template", "Template is null.");
			if (project == null)
				throw new ArgumentNullException("project", "Project is null.");
            if (searcher == null)
                throw new ArgumentNullException("searcher", "Searcher is null.");
            if (statebag == null)
                throw new ArgumentNullException("statebag", "Statebag is null.");

			_template = template;
			_project = project;
            _searcher = searcher;
            _statebag = statebag;
		}

		
        /// <summary>
        /// Returns the searcher ojbect.
        /// </summary>
        public string GetDerivedSearcher()
        {
            string body = "";

            foreach (SearcherMethod method in _searcher.Methods)
                body += GetDerivedMethod(method);

            // Get the compiled template.
            string searcher = string.Format(_template.GetValue("//derivedseacher"), _searcher.SearcherName, _project.AutorInfo.Name,
                _project.AutorInfo.Copyright, DateTime.Now.ToString(), _project.Namespace, OPENING_BRACKETS,
                _searcher.Table.PersistentName, _searcher.SearcherName, body, CLOSING_BRACKETS,
                String.Format(_statebag.DeriveFrom, _searcher.Table.PersistentName), _statebag.UsingDirective);

            return searcher.Replace("<t>", "\t"); 
        }

        /// <summary>
        /// Returns the method ojbect.
        /// </summary>
        private string GetDerivedMethod(SearcherMethod method)
        {
			// Object<{0}> is there for compatibility reason. It's no longer used in the wizard.
            string returntype = (method.ReturnType == "Object<{0}>" ? _searcher.Table.PersistentName : String.Format(method.ReturnType, _searcher.Table.PersistentName));
            string getter = String.Format(method.ReturnType, _searcher.Table.PersistentName);

            string parameters = "";
            string comments = "";
            string fields = "";
            string query = "";
            for (int i = 0; i < method.Fields.Count;i++)
            {
                SearcherField searcherField = method.Fields[i];
				Field field = searcherField.Field;

                parameters += String.Format("{0}{1} {2}", (i > 0 ? ", " : ""), GetCSharpType(field), field.PropertyName.ToLower());
                comments += String.Format(_template.GetValue("//comment"), field.PropertyName.ToLower(), field.Description);
				query += String.Format(_template.GetValue(i == 0 ? "//firstquerymethod" : "//querymethod"), field.PropertyName, searcherField.Comparer, OPENING_BRACKETS, i, CLOSING_BRACKETS);
                fields += String.Format(", {0}", field.PropertyName.ToLower());
            }

            string body = string.Format(_template.GetValue("//derivedfindby"), _searcher.Table.PersistentName, returntype, method.Name, parameters, OPENING_BRACKETS,
                                        CLOSING_BRACKETS,comments,fields,query,getter);

            return body;
        }

        /// <summary>
        /// Returns the searcher ojbect.
        /// </summary>
        public string GetStaticSearcher()
        {
            string body = "";

            foreach (SearcherMethod method in _searcher.Methods)
                body += GetStaticMethod(method);

            // Get the compiled template.
            string searcher = string.Format(_template.GetValue("//staticseacher"), _searcher.SearcherName, _project.AutorInfo.Name,
                _project.AutorInfo.Copyright, DateTime.Now.ToString(), _project.Namespace, OPENING_BRACKETS,
                _searcher.Table.PersistentName, body, _statebag.BusinessServer, CLOSING_BRACKETS, _statebag.UsingDirective);

            return searcher.Replace("<t>", "\t");
        }

        /// <summary>
        /// Returns the method ojbect.
        /// </summary>
        private string GetStaticMethod(SearcherMethod method)
        {
			// Object<{0}> is there for compatibility reason. It's no longer used in the wizard.
			string returntype = (method.ReturnType == "Object<{0}>" ? _searcher.Table.PersistentName : String.Format(method.ReturnType, _searcher.Table.PersistentName));
            string getter = String.Format(method.ReturnType, _searcher.Table.PersistentName);

            string parameters = "";
            string comments = "";
            string fields = "";
            string query = "";
            for (int i = 0; i < method.Fields.Count; i++)
            {
                SearcherField searcherField = method.Fields[i];
				Field field = searcherField.Field;

                parameters += String.Format("{0}{1} {2}", (i > 0 ? ", " : ""), GetCSharpType(field), field.PropertyName.ToLower());
                comments += String.Format(_template.GetValue("//comment"), field.PropertyName.ToLower(), field.Description);
                query += String.Format(_template.GetValue(i == 0 ? "//firstquerymethod" : "//querymethod"), field.PropertyName, searcherField.Comparer, OPENING_BRACKETS, i, CLOSING_BRACKETS);
                fields += String.Format(", {0}", field.PropertyName.ToLower());
            }

            string body = string.Format(_template.GetValue("//staticfindby"), _searcher.Table.PersistentName, returntype, method.Name, parameters, OPENING_BRACKETS,
                                        CLOSING_BRACKETS, comments, fields, query, _statebag.BusinessServer,getter);

            return body;
        }


		#region GetCSharpType

		/// <summary>
		/// Returns the C# type for a given field.
		/// </summary>
		/// <param name="field">The field that's type is returned.</param>
		private string GetCSharpType(Field field)
		{
			string type = _template.GetValue(string.Format("//types/type[@id = '{0}']", field.Type));
			// Create the type as nullable type if not a string and null is allowed.
			if (IsNullable(field))
				type += "?";

			return type;
		}

		/// <summary>
		/// Returns whether the field is nullable.
		/// </summary>
		/// <param name="field">Field that is checked.</param>
        private bool IsNullable(Field field)
		{
			return (field.Type != typeof(string) && field.Type != typeof(byte[]) && field.AllowDBNull && !field.Identifier);
		}

		#endregion
	}
}
