#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Windows.Forms;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using Chili.Opf3.Wizard.Shared;
using System.IO;
using System.Reflection;
using Chili.Opf3.Wizard.CSharp.Searcher.UI;
using Chili.Opf3.Wizard.Shared.Storages;
using Chili.Opf3.Wizard.Shared.UI;
using Chili.Opf3.Wizard.CSharp.Searcher.Properties;

#endregion

namespace Chili.Opf3.Wizard.CSharp.Searcher
{
	/// <summary>
	/// Creates the interface representation for each persistent object.
	/// </summary>
	public sealed class SearcherProvider : IObjectsProvider
	{
		private const string TEMPLATE_FILE_NAME = "Chili.Opf3.Wizard.CSharp.SearcherTemplate.xml";
		private const string SEARCHER_FILE_NAME = "{0}\\{1}.cs";

		private List<Page> _pages = new List<Page>();
		private List<Searcher> _searchers = new List<Searcher>();
        private Project _project;
        private StateBag _statebag = new StateBag();

		/// <summary>
		/// Creates a new instance of the <see cref="SearcherProvider">SearcherProvider Class.</see>
		/// </summary>
		public SearcherProvider()
		{
			_pages.Add(new SearcherPage(this));
            _pages.Add(new AdditionalDataPage(this, _statebag));
        }

        #region IObjectsProvider Members

        /// <summary>
		/// Generates the objects form the project structure.
		/// </summary>
		/// <param name="project">The project containing all selected tables and fields of the database.</param>
		public void GenerateObjects(Project project)
		{
			foreach (Searcher searcher in _searchers)
			{
				// Skip if the seacher is not to generate.
				if (!searcher.Generate)
					continue;

				// Invoke the event generating the object.
				OnGeneratingObject(new GeneratingObjectEventArgs(searcher.ID));

				try
				{
					string directory = Path.GetDirectoryName(Assembly.GetAssembly(this.GetType()).Location);
					// Create the template.
					Template template = new Template(string.Format("{0}\\{1}", directory, TEMPLATE_FILE_NAME));

					// Create the searcher creator that creates one single searcher.
					SearcherCreator creator = new SearcherCreator(project, searcher, template, _statebag);

                    string content;
					// Get the content of the searcher.
                    if(_statebag.CreateStatic)
                        content = creator.GetStaticSearcher();
                    else
                        content = creator.GetDerivedSearcher();


					WriteSearcherFile(content, project.OutputDirectoy, searcher.SearcherName, searcher.ID);

					// Invoke the event to signalize to the wizard that the object has been written.
					OnGeneratedObject(new GeneratedObjectEventArgs(searcher.ID));
 				}
				catch (Exception ex)
				{
					// Invoke the event to tell the wizard that the object couldn't be created.
					OnGeneratedObject(new GeneratedObjectEventArgs(searcher.ID, ex));
				}
			}
		}

		/// <summary>
		/// Event is fired before the object is generated from the data of the <see cref="Project">Project</see> class.
		/// </summary>
		public event EventHandler<GeneratingObjectEventArgs> GeneratingObject;

		/// <summary>
		/// Event is fired after the object is generated from the data of the <see cref="Project">Project</see> class.
		/// </summary>
		public event EventHandler<GeneratedObjectEventArgs> GeneratedObject;

		/// <summary>
		/// Returns the name for the tool.
		/// </summary>
		public string Name
		{
			get { return Resources.Name; }
		}

		/// <summary>
		/// Returns the description for the tool.
		/// </summary>
		public string Description
		{
			get { return Resources.Description; }
		}

		/// <summary>
		/// Gets the language name supported by this tool. Or null if no explicit language
		/// is supported.
		/// </summary>
		public string LanguageName
		{
			get { return Resources.LanguageName; }
		}

		/// <summary>
		/// Returns a list of pages displayed by the class.
		/// </summary>
		public List<Page> Pages
		{
			get { return _pages; }
		}

		/// <summary>
		/// Returns the list of objects that are generated.
		/// </summary>
		public List<IGeneratedObject> GetObjectsList()
		{
			List<IGeneratedObject> list = new List<IGeneratedObject>();
			foreach (Searcher i in _searchers)
			{
				if (!i.Generate)
					continue;

				list.Add((IGeneratedObject)i);
			}

			return list;
		}

		/// <summary>
		/// Returns the version of the plug-in.
		/// </summary>
		public string Version
		{
			get { return this.GetType().Assembly.GetName().Version.ToString(); }
		}

		/// <summary>
		/// Display the about dialog for the plug-in.
		/// </summary>
		public void ShowAboutDialog()
		{
            MessageBox.Show(string.Format(Resources.SearcherProviderAboutText, "\r\n", this.Version), Resources.SearcherProviderAboutTitle,
                MessageBoxButtons.OK, MessageBoxIcon.Information);
        }

        /// <summary>
        /// Allow the provider to build internal data from Storage.
        /// </summary>
        public void ProcessStorage(Project project)
        {
            _project = project;

            List<Searcher> temp = new List<Searcher>();

            foreach (Table table in _project.Tables)
            {
                temp.Add(GetSearcher(table));
            }
            _searchers.Clear();
            _searchers.AddRange(temp);
        }

        /// <summary>
        /// Retrieve the searcher associed with the given table or create a new.
        /// </summary>
        private Searcher GetSearcher(Table table)
        {
            foreach (Searcher s in _searchers)
            {
                if (s.Table.PersistentName == table.PersistentName)
                    return s;
            }
            Searcher searcher = new Searcher(table); 
            searcher.BuildMethods();
            return searcher;
        }

        /// <summary>
        /// Initialize defaults in a new project.
        /// </summary>
        public void InitializeDefaults(Project project)
        {
            // append the typeconversion pair List.
            // extract from the template.
            string directory = Path.GetDirectoryName(Assembly.GetAssembly(this.GetType()).Location);
            Template template = new Template(string.Format("{0}\\{1}", directory, TEMPLATE_FILE_NAME));

            XmlNodeList types = template.Document.SelectNodes("//types/type");
            foreach (XmlNode node in types)
            {
                TypeConversionPair pair = new TypeConversionPair(node.Attributes[0].Value, node.InnerText);
                pair.Language = Resources.LanguageName;

                project.TypeConverter.AppendPair(pair);
            }
        }

		/// <summary>
		/// Retrieves the plug-in state from project.
		/// </summary>
		public void Deserialize(Project project, XmlDocument xml, XmlNode d)
		{
            // Load the interfaces items from the project's xml.
            List<Searcher> work = new SearcherSerializer().Deserialize(xml, d, project);
            if (work != null)
            {
                _searchers = work;
            }
            // Load the StateBag from the xml.
            _statebag.Deserialize(project, xml, d); 
		}

		/// <summary>
		/// Persists the plug-in state .
		/// </summary>
        public void Serialize(Project project, XmlDocument xml, XmlNode d)
		{
            d.AppendChild(new SearcherSerializer().Serialize(xml, _searchers));
            d.AppendChild(_statebag.Serialize(xml)); 
		}


		#endregion

		/// <summary>
		/// Writes the searcher object to a file in the output directory.
		/// </summary>
        /// <param name="searcherContent">The searcher content that is written to the file.</param>
		/// <param name="outputDirectory">The output directory.</param>
        /// <param name="searcherName">The name of the searcher object.</param>
        /// <param name="id">The id of the searcher.</param>
		private void WriteSearcherFile(string searcherContent, string outputDirectory, string searcherName, string id)
		{
#if DEMO
			throw new NotSupportedException("Output not supported in Demo.");
#else
			// Get the path for the searcher object.
			string path = string.Format(SEARCHER_FILE_NAME, outputDirectory, searcherName);

			// Tell the wizard that the file couldn't be overwritten.
			if (File.Exists(path))
				throw new NotSupportedException("File already exists. Please delete the existing file frist.");

			// Write the file.
			using (TextWriter writer = File.CreateText(path))
			{
				writer.WriteLine(searcherContent);
			}
#endif
		}

		/// <summary>
		/// Invoked to fire the <see cref="GeneratingObject">GeneratingObject</see> event.
		/// </summary>
		/// <param name="e">The event args class containing further information of the event.</param>
		private void OnGeneratingObject(GeneratingObjectEventArgs e)
		{
			if (GeneratingObject != null)
				GeneratingObject(this, e);
		}

		/// <summary>
		/// Invoked to fire the <see cref="GeneratedObject">GeneratedObject</see> event.
		/// </summary>
		/// <param name="e">The event args class containing further information of the event.</param>
		private void OnGeneratedObject(GeneratedObjectEventArgs e)
		{
			if (GeneratedObject != null)
				GeneratedObject(this, e);
		}

		/// <summary>
		/// Returns the list of seachers manages by this instance.
		/// </summary>
		internal List<Searcher> Seachers
		{
			get { return _searchers; }
		}
    }
}
