#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Chili.Opf3.Wizard.Shared.UI;
using Chili.Opf3.Wizard.Shared.Storages;
using Chili.Opf3.Wizard.CSharp.Searcher.Properties;

#endregion

namespace Chili.Opf3.Wizard.CSharp.Searcher.UI
{
	/// <summary>
	/// This dialog allows to edit one relation item.
	/// </summary>
	public partial class MethodItemDialog : BaseDialog
	{
        private SearcherMethod _method;
        private Table _table;

		/// <summary>
        /// Creates a new instance of the <see cref="RelationsItemDialog">MethodItemDialog Class</see>.
		/// </summary>
		/// <param name="mode">The mode of the dialog.</param>
        /// <param name="method">The method that is edited.</param>
        /// <param name="table">The base table.</param>
        public MethodItemDialog(DialogModes mode, SearcherMethod method, Table table)
		{
			InitializeComponent();

            if (method == null)
                throw new ArgumentNullException("method", "Method is null.");
            if (table == null)
                throw new ArgumentNullException("table", "Table is null.");

            _method = method;
            _table = table;

            PopulateAvailableFieldsList(_table, availablefieldsListBox);

			// Set the labels with the persistent names.
            methodNameTextBox.Text = _method.Name;
            returntypeCombobox.Text = _method.ReturnType;

            if (mode == DialogModes.New)
                this.Text = Resources.NewMethodItemDialogTitle;
            else
            {
                this.Text = Resources.EditMethodItemDialogTitle;
                PopulateControls();
            }
		}

		/// <summary>
        /// Populates the given listbox with the fields of the given table.
		/// </summary>
		/// <param name="table">The table that's fields are populated.</param>
		/// <param name="combobox">The combobox.</param>
        private void PopulateAvailableFieldsList(Table table, ListBox listbox)
		{
            // Clear all items in the listbox.
            listbox.Items.Clear();

			foreach (Field field in table.Fields)
			{
                listbox.Items.Add(field);
			}
            SetOperationButtonsStatus();
		}

		/// <summary>
		/// Populates the UI controls.
		/// </summary>
		private void PopulateControls()
		{
            foreach (SearcherField field in _method.Fields)
            {
                AddField(field);
            }
            SetOperationButtonsStatus();
		}

		
		/// <summary>
		/// Invoked each time the form is about to close.
		/// </summary>
		private void IndexItemDialog_FormClosing(object sender, FormClosingEventArgs e)
		{
			if (this.DialogResult != DialogResult.OK)
				return;

			e.Cancel = (!PopulateObject());
		}


        /// <summary>
        /// Add a field to the fields list and remove it from availables.
        /// </summary>
        private void AddField(SearcherField field)
        {
            if (field == null) 
				return;
            
			if (fieldsListBox.Items.Contains(field)) 
				return;
            
			fieldsListBox.Items.Add(field);
            availablefieldsListBox.Items.Remove(field.Field);  
        }

        /// <summary>
        /// Remove a field from the fields list and add it from availables.
        /// </summary>
        private void RemoveField(SearcherField field)
        {
            if (field == null) 
				return;

            if (!fieldsListBox.Items.Contains(field)) 
				return;
            
			fieldsListBox.Items.Remove(field);
            availablefieldsListBox.Items.Add(field.Field); 
        }


		/// <summary>
		/// Populates the object with data.
		/// </summary>
		private bool PopulateObject()
		{
			if (methodNameTextBox.Text.Length == 0)
			{
                MessageBox.Show(Resources.MissingMethodNameDescription, Resources.ProgramName, MessageBoxButtons.OK, MessageBoxIcon.Error);
				methodNameTextBox.Focus();
				return false;
			}

            if (returntypeCombobox.Text.Length == 0)
            {
                MessageBox.Show(Resources.MissingMethodReturnType, Resources.ProgramName, MessageBoxButtons.OK, MessageBoxIcon.Error);
                returntypeCombobox.Focus();
                return false;
            }

            _method.Name=methodNameTextBox.Text;
            _method.ReturnType = returntypeCombobox.Text;

            // fill the fields list
            _method.Fields.Clear();
			foreach (object item in fieldsListBox.Items)
			{
				_method.Fields.Add((SearcherField)item);
			}

			return true;
		}

        /// <summary>
        /// Add all available fields.
        /// </summary>
        private void addallButton_Click(object sender, EventArgs e)
        {
			List<Field> fields = new List<Field>();
			foreach (Field field in availablefieldsListBox.Items)
				fields.Add(field);

            foreach (Field field in fields)
            {
                AddField(new SearcherField(field, comparerComboBox.Text));
            }
            SetOperationButtonsStatus();
        }

        /// <summary>
        /// Add the available selected fields.
        /// </summary>
        private void addcurrentButton_Click(object sender, EventArgs e)
        {
            AddField(new SearcherField((Field)availablefieldsListBox.SelectedItem, comparerComboBox.Text));
            SetOperationButtonsStatus();
        }

        /// <summary>
        /// Add the available selected fields.
        /// </summary>
        /// 
        private void removeallButton_Click(object sender, EventArgs e)
        {
			List<SearcherField> fields = new List<SearcherField>();
			foreach (SearcherField field in fieldsListBox.Items)
				fields.Add(field);
			
			foreach (SearcherField field in fields)
            {
                RemoveField(field);
            }
            SetOperationButtonsStatus();
        }

        /// <summary>
        /// Add the available selected fields.
        /// </summary>
        /// 
        private void removecurrentButton_Click(object sender, EventArgs e)
        {
			RemoveField((SearcherField)fieldsListBox.SelectedItem);
            SetOperationButtonsStatus();
        }

        /// <summary>
        /// Enable /disable remove/add buttons.
        /// </summary>
        /// 
        private void SetOperationButtonsStatus()
        {
            removecurrentButton.Enabled = (fieldsListBox.SelectedItems.Count > 0);
            removeallButton.Enabled = (fieldsListBox.Items.Count > 0);
            addcurrentButton.Enabled = (availablefieldsListBox.SelectedItems.Count > 0);
            addallButton.Enabled = (availablefieldsListBox.Items.Count > 0);
        }

        /// <summary>
        /// Enable /disable addxxx buttons.
        /// </summary>
        /// 
        private void fieldsListBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            SetOperationButtonsStatus();
        }

        /// <summary>
        /// Enable /disable addxxx buttons.
        /// </summary>
        /// 
        private void availablefieldsListBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            SetOperationButtonsStatus();
        }

        /// <summary>
        /// Add the available selected fields.
        /// </summary>
        /// 
        private void availablefieldsListBox_DoubleClick(object sender, EventArgs e)
        {
            AddField(new SearcherField((Field)availablefieldsListBox.SelectedItem, comparerComboBox.Text));
        }
	}
}