#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Text.RegularExpressions;
using Chili.Opf3.Wizard.CSharp.Searcher.Properties;
using Chili.Opf3.Wizard.Shared.UI;

#endregion

namespace Chili.Opf3.Wizard.CSharp.Searcher.UI
{
	/// <summary>
	/// This dialog allows to edit a searcher.
	/// </summary>
    public partial class SearcherDialog : BaseDialog
	{
		private Searcher _searcher;

		/// <summary>
        /// Creates a new instance of the <see cref="InterfaceDialog">SearcherDialog Class</see>.
		/// </summary>
		/// <param name="i"></param>
		public SearcherDialog(Searcher i)
		{
			InitializeComponent();

			if (i == null)
				throw new ArgumentNullException("i", "Searcher is null");
			_searcher = i;

			this.Text = string.Format(Resources.EditSearcherDialogTitle, i.SearcherName);

			// Populates the controls.
			PopulateControls();
            PopulateMethodListView();

            UpdateControlsState();
		}

		/// <summary>
		/// Populates the UI controls with data.
		/// </summary>
		private void PopulateControls()
		{
			persistentLabel.Text = _searcher.Table.PersistentName;
			searcherNameTextBox.Text = _searcher.SearcherName;
		}

		/// <summary>
		/// Invoked each time the form is about to close.
		/// </summary>
		private void SearcherDialog_FormClosing(object sender, FormClosingEventArgs e)
		{
			if (this.DialogResult != DialogResult.OK)
				return;

			e.Cancel = (!PopulateObject());
		}

		/// <summary>
		/// Populates the object with data.
		/// </summary>
		private bool PopulateObject()
		{
			if (searcherNameTextBox.Text.Trim().Length == 0)
			{
				MessageBox.Show(Resources.MissingSearcherNameText, Resources.Name, MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}
			if (!new Regex(@"[A-Z]").IsMatch(searcherNameTextBox.Text.Trim().ToUpper().Substring(0, 1)))
			{
				MessageBox.Show(Resources.InvalidSearcherNameText, Resources.Name,  MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}

			_searcher.SearcherName = searcherNameTextBox.Text;
			return true;
		}

        /// <summary>
        /// Populates the indexes list view.
        /// </summary>
        private void PopulateMethodListView()
        {
            methodsListView.Items.Clear();

            // Suspend the layout of the grid.
            methodsListView.SuspendLayout();
            try
            {
                // Loop over all index and add them to the list view.
                foreach (SearcherMethod index in _searcher.Methods)
                {
                    ListViewItem item = new ListViewItem();
                    PopulateListViewItem(index, item);
                    methodsListView.Items.Add(item);
                }
            }
            finally
            {
                // Resume the layout of the grid.
                methodsListView.ResumeLayout();
            }
        }

        /// <summary>
        /// Creates a listview item for the given method.
        /// </summary>
        /// <param name="method">The method that is used to generate the list view item.</param>
        /// <param name="item">The listview item that is created.</param>
        private void PopulateListViewItem(SearcherMethod method, ListViewItem item)
        {
            // Remove all subitems.
            item.SubItems.Clear();
            item.Tag = method;

            item.Checked = method.Generate;
            item.ImageIndex = 0;
            item.Text = method.Name;
            item.SubItems.Add(method.ReturnType);
            item.SubItems.Add(method.FieldsName);
        }

        /// <summary>
        /// Returns the selected method or null if no method has been selected.
        /// </summary>
        private SearcherMethod GetSelected()
        {
            if (methodsListView.SelectedItems.Count == 0)
                return null;

            return methodsListView.SelectedItems[0].Tag as SearcherMethod;
        }

        /// <summary>
        /// Invoked each time the selected index changed.
        /// </summary>
        private void methodsListView_SelectedIndexChanged(object sender, EventArgs e)
        {
            UpdateControlsState();
        }

        /// <summary>
        /// Updates the state of the controls on page.
        /// </summary>
        private void UpdateControlsState()
        {
            editbutton.Enabled = (methodsListView.SelectedItems.Count > 0);
            removebutton.Enabled = (methodsListView.SelectedItems.Count > 0);
        }

        /// <summary>
        /// Invoked each time the add button is clicked.
        /// </summary>
        private void addbutton_Click(object sender, EventArgs e)
        {
            SearcherMethod method = new SearcherMethod();
            method.Generate = true;

            // Show the dialog to edit the index.
            if (new MethodItemDialog(DialogModes.New, method, _searcher.Table).ShowDialog(this) == DialogResult.OK)
            {
                // Create a new listview item.
                ListViewItem item = new ListViewItem();
                PopulateListViewItem(method, item);

                // Add the index to the list and to the listview.
                methodsListView.Items.Add(item);
                _searcher.Methods.Add(method);
            }

        }

        /// <summary>
        /// Invoked each time the edit button is clicked.
        /// </summary>
        private void editbutton_Click(object sender, EventArgs e)
        {
            SearcherMethod method = GetSelected();
            if (method == null)
                return;

            // Show the dialog to edit the index.
            if (new MethodItemDialog(DialogModes.Edit, method, _searcher.Table).ShowDialog(this) == DialogResult.OK)
            {
                PopulateListViewItem(method, methodsListView.SelectedItems[0]);
            }

        }


        /// <summary>
        /// Invoked each time the remove button is clicked.
        /// </summary>
        private void removebutton_Click(object sender, EventArgs e)
        {
            SearcherMethod method = GetSelected();
            if (method == null)
                return;

            // Asks the user if the table should be deleted.
            if (MessageBox.Show(Resources.DeleteMethodText, Resources.ProgramName, MessageBoxButtons.YesNo, MessageBoxIcon.Question) ==
                DialogResult.Yes)
            {
                // Removes the table from the listview and the list.
                _searcher.Methods.Remove(method);
                methodsListView.Items.Remove(methodsListView.SelectedItems[0]);
            }

        }

        /// <summary>
        /// Invoked each time the check all button is clicked.
        /// </summary>
        private void checkAllbutton_Click(object sender, EventArgs e)
        {
            // Loop over all items and check.
            foreach (ListViewItem item in methodsListView.Items)
            {
                item.Checked = true;
            }

        }

        /// <summary>
        /// Invoked each time the uncheck all button is clicked.
        /// </summary>
        private void unCheckAllbutton_Click(object sender, EventArgs e)
        {
            // Loop over all items and uncheck.
            foreach (ListViewItem item in methodsListView.Items)
            {
                item.Checked = false;
            }

        }
	}
}