#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using Chili.Opf3.Wizard.Shared;
using Chili.Opf3.Wizard.Shared.Storages;

using Chili.Opf3.Wizard.CSharp.Searcher;
using Chili.Opf3.Wizard.CSharp.Searcher.Properties;
using Chili.Opf3.Wizard.Shared.UI;

#endregion

namespace Chili.Opf3.Wizard.CSharp.Searcher.UI
{
	/// <summary>
	/// This class allows to customize the searchers that are going created.
	/// </summary>
	public partial class SearcherPage : Page
	{
		private SearcherProvider _provider;
		private Project _project;

		/// <summary>
		/// Creates a new instance of the <see cref="SearcherPage">SearcherPage Class</see>.
		/// </summary>
		/// <param name="provider">The provider itself.</param>
		public SearcherPage(SearcherProvider provider) : base(provider)
		{
			InitializeComponent();

			if (provider == null)
				throw new ArgumentNullException("provider", "Provider is null.");
			_provider = provider;
		}

		/// <summary>
		/// Invoked each time the page is opened.
		/// </summary>
		/// <param name="project">The current project.</param>
		public override void PageOpening(Project project)
		{
			// Set the current project.
			_project = project;

			SetHeaderTitle(Resources.SearchersHeaderTitle);
			SetHeaderDescription(Resources.SearchersHeaderDescription);

			// Bind the searcher list.
			BindSearcherListToListView();
		}

		/// <summary>
		/// Binds the searcher list to the listview.
		/// </summary>
		private void BindSearcherListToListView()
		{
			// Clear the list.
			searchersListView.Items.Clear();
			// Suspend the layout.
			searchersListView.SuspendLayout();

			try
			{
				foreach (Searcher i in _provider.Seachers)
				{
					ListViewItem item = new ListViewItem();
					// Populate the list view item.
					PopulateListViewItem(i, item);
					// Add the item to the listview.
					searchersListView.Items.Add(item);
				}
			}
			finally
			{
				// Resume the layout.
				searchersListView.ResumeLayout();
			}
		}

		/// <summary>
		/// Populate one list view item with data.
		/// </summary>
		/// <param name="s">The searcher itself. This searcher is used to populate the item.</param>
		/// <param name="item">The listview item that is populated.</param>
		private void PopulateListViewItem(Searcher s, ListViewItem item)
		{
			// Clear all subitems.
			item.SubItems.Clear();

			item.Checked = s.Generate;
			item.ImageIndex = 0;
			item.Text = s.Table.PersistentName;
			item.SubItems.Add(s.SearcherName);
		}

		
		/// <summary>
		/// Invoked each time the page is closed.
		/// </summary>
		/// <param name="next">True if the next button has been clicked. False on all other.</param>
		public override bool PageClosing(bool next)
		{
			// Loop over all searchers and set the checked state.
			for (int i = 0; i < _provider.Seachers.Count; i++)
			{
				_provider.Seachers[i].Generate = searchersListView.Items[i].Checked;
			}
			return true;
		}

		/// <summary>
		/// Invoked each time the check all button is clicked.
		/// </summary>
		private void checkAllButton_Click(object sender, EventArgs e)
		{
			// Loop over all items and uncheck.
			for (int i = 0; i < searchersListView.Items.Count; i++)
			{
				searchersListView.Items[i].Checked = true;
			}
		}

		/// <summary>
		/// Invoked each time the uncheck all button is clicked.
		/// </summary>
		private void unCheckAllButton_Click(object sender, EventArgs e)
		{
			// Loop over all items and uncheck.
			for (int i = 0; i < searchersListView.Items.Count; i++)
			{
				searchersListView.Items[i].Checked = false;
			}
		}

		/// <summary>
		/// Invoked each time the edit button is clicked.
		/// </summary>
		private void editButton_Click(object sender, EventArgs e)
		{
			Searcher s = GetSelectedSearcher();
			if (s == null)
				return;

			// Check if we got an OK.
			if (new SearcherDialog(s).ShowDialog(this) == DialogResult.OK)
			{
				// Update the current selected item in the list.
				ListViewItem item = searchersListView.SelectedItems[0];
				PopulateListViewItem(s, item);
			}
		}

		/// <summary>
		/// Returns the selected searcher or null if none is selected.
		/// </summary>
		private Searcher GetSelectedSearcher()
		{
			if (searchersListView.SelectedItems.Count == 0)
				return null;

			return _provider.Seachers[searchersListView.SelectedIndices[0]];
		}

		/// <summary>
		/// Invoked each time the selected index of the list view changed.
		/// </summary>
		private void searchersListView_SelectedIndexChanged(object sender, EventArgs e)
		{
			editButton.Enabled = (searchersListView.SelectedItems.Count > 0);
		}
	}
}
