#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using Chili.Opf3.Wizard.Shared;
using System.IO;
using System.Reflection;
using Chili.Opf3.Wizard.CSharp.InterfacesProvider.UI;
using Chili.Opf3.Wizard.Shared.UI;
using Chili.Opf3.Wizard.Shared.Storages;
using Chili.Opf3.Wizard.CSharp.Properties;
using System.Windows.Forms;

#endregion

namespace Chili.Opf3.Wizard.CSharp.InterfacesProvider
{
	/// <summary>
	/// Creates the interface representation for each persistent object.
	/// </summary>
	public sealed class InterfacesProvider : IObjectsProvider
	{
		private const string TEMPLATE_FILE_NAME = "Chili.Opf3.Wizard.CSharp.InterfaceTemplate.xml";
		private const string INTERFACE_FILE_NAME = "{0}\\{1}.cs";
		private const string PERSIST_INTERFACES_ID = "interfaces";

		private static InterfacesProvider _current;
        private Project _project;

		private List<Page> _pages = new List<Page>();
		private List<Interface> _interfaces = new List<Interface>();

		/// <summary>
		/// Creates a new instance of the <see cref="CSharpPersistentInterfacesProvider">CSharpPersistentInterfacesProvider Class.</see>
		/// </summary>
		public InterfacesProvider()
		{
			_pages.Add(new InterfacePage(this));
        }

        #region IObjectsProvider Members

        /// <summary>
		/// Generates the utility objects form the project structure.
		/// </summary>
		/// <param name="project">The project containing all selected tables and fields of the database.</param>
		public void GenerateObjects(Project project)
		{
			foreach (Interface iface in this.Interfaces)
			{
				// Skip if the interface is not to generate.
				if (!iface.Generate)
					continue;

				// Invoke the event generating the object.
				OnGeneratingObject(new GeneratingObjectEventArgs(iface.ID));

				try
				{
					string directory = Path.GetDirectoryName(Assembly.GetAssembly(this.GetType()).Location);
					// Create the template.
					Template template = new Template(string.Format("{0}\\{1}", directory, TEMPLATE_FILE_NAME));

					// Create the persistent creator that creates one single persistent class.
					InterfaceCreator creator = new InterfaceCreator(project, iface, template);
					// Get the content of the interface.
					string ifaceContent = creator.GetInterface();

					WriteInterfaceFile(ifaceContent, project.OutputDirectoy, iface.InterfaceName, iface.ID);

					// Invoke the event to signalize to the wizard that the object has been written.
					OnGeneratedObject(new GeneratedObjectEventArgs(iface.ID));
 				}
				catch (Exception ex)
				{
					// Invoke the event to tell the wizard that the object couldn't be created.
					OnGeneratedObject(new GeneratedObjectEventArgs(iface.ID, ex));
				}
			}
		}

		/// <summary>
		/// Event is fired before the object is generated from the data of the <see cref="Project">Project</see> class.
		/// </summary>
		public event EventHandler<GeneratingObjectEventArgs> GeneratingObject;

		/// <summary>
		/// Event is fired after the object is generated from the data of the <see cref="Project">Project</see> class.
		/// </summary>
		public event EventHandler<GeneratedObjectEventArgs> GeneratedObject;

		/// <summary>
		/// Returns the name for the plug-in.
		/// </summary>
		public string Name
		{
			get { return Resources.NameInterfaces; }
		}

		/// <summary>
		/// Returns the description for the plug-in.
		/// </summary>
		public string Description
		{
			get { return Resources.DescriptionInterfaces; }
		}

		/// <summary>
		/// Gets the language name supported by this tool. Or null if no explicit language
		/// is supported.
		/// </summary>
		public string LanguageName
		{
			get { return Resources.LanguageNameInterfaces; }
		}

		/// <summary>
		/// Returns a list of pages displayed by the class.
		/// </summary>
		public List<Page> Pages
		{
			get { return _pages; }
		}

		/// <summary>
		/// Returns the list of objects that are generated.
		/// </summary>
		public List<IGeneratedObject> GetObjectsList()
		{
			List<IGeneratedObject> list = new List<IGeneratedObject>();
			foreach (Interface i in this.Interfaces)
			{
				if (!i.Generate)
					continue;

				list.Add((IGeneratedObject)i);
			}

			return list;
		}

		/// <summary>
		/// Returns the version of the plug-in.
		/// </summary>
		public string Version
		{
			get { return this.GetType().Assembly.GetName().Version.ToString(); }
		}

		/// <summary>
		/// Display the about dialog for the plug-in.
		/// </summary>
		public void ShowAboutDialog()
        {
			MessageBox.Show(string.Format(Resources.InterfaceProviderAboutText, "\r\n", this.Version), Resources.InterfaceProviderAboutTitle, 
				MessageBoxButtons.OK, MessageBoxIcon.Information);
        }

        /// <summary>
        /// Allow the provider to build internal data from Storage.
        /// </summary>
        public void ProcessStorage(Project project)
        {
            _project = project;

            List<Interface> temp = new List<Interface>();

            foreach (Table table in _project.Tables)
            {
               temp.Add(GetInterface(table));
            }
            _interfaces.Clear();
            _interfaces.AddRange(temp);
        }

        /// <summary>
        /// Retrieve the interface associed with the given table or create a new.
        /// </summary>
        private Interface GetInterface(Table table)
        {
            foreach (Interface i in _interfaces)
            {
                if (i.Table.PersistentName == table.PersistentName)
                    return i;
            }
            return new Interface(table);
        }

        /// <summary>
        /// Initialize defaults in a new project.
        /// </summary>
        public void InitializeDefaults(Project project)
        {
            // append the typeconversion pair List.
            // extract from the template.
            string directory = Path.GetDirectoryName(Assembly.GetAssembly(this.GetType()).Location);
            Template template = new Template(string.Format("{0}\\{1}", directory, TEMPLATE_FILE_NAME));

            XmlNodeList types = template.Document.SelectNodes("//types/type");
            foreach (XmlNode node in types)
            {
                TypeConversionPair pair = new TypeConversionPair(node.Attributes[0].Value, node.InnerText);
                pair.Language = Resources.LanguageName;

                project.TypeConverter.AppendPair(pair);
            }
        }

        /// <summary>
        /// Deserialize the plug-in state from xml.
        /// </summary>
        public void Deserialize(Project project, XmlDocument xml, XmlNode d)
        {
            // Load the interfaces items from the project's xml.
            List<Interface> work = new InterfacesSerializer().Deserialize(xml, d, project);
            if (work != null)
            {
                _interfaces = work;
            }
        }

        /// <summary>
        /// Serialize the plug-in state in xml.
        /// </summary>
        public void Serialize(Project project, XmlDocument xml, XmlNode d)
        {
            d.AppendChild(new InterfacesSerializer().Serialize(xml,_interfaces));
        }

		#endregion

		/// <summary>
		/// Writes the persistent object to a file in the output directory.
		/// </summary>
		/// <param name="interfaceContent">The interface content that is written to the file.</param>
		/// <param name="outputDirectory">The output directory.</param>
		/// <param name="interfaceName">The name of the interface object.</param>
		/// <param name="id">The id of the interface.</param>
		private void WriteInterfaceFile(string interfaceContent, string outputDirectory, string interfaceName, string id)
		{
#if DEMO
			throw new NotSupportedException("Output not supported in Demo.");
#else
			// Get the path for the interface object.
			string path = string.Format(INTERFACE_FILE_NAME, outputDirectory, interfaceName);

			// Tell the wizard that the file couldn't be overwritten.
			if (File.Exists(path))
				throw new NotSupportedException("File already exists. Please delete the existing file frist.");

			// Write the file.
			using (TextWriter writer = File.CreateText(path))
			{
				writer.WriteLine(interfaceContent);
			}
#endif
		}

		/// <summary>
		/// Invoked to fire the <see cref="GeneratingObject">GeneratingObject</see> event.
		/// </summary>
		/// <param name="e">The event args class containing further information of the event.</param>
		private void OnGeneratingObject(GeneratingObjectEventArgs e)
		{
			if (GeneratingObject != null)
				GeneratingObject(this, e);
		}

		/// <summary>
		/// Invoked to fire the <see cref="GeneratedObject">GeneratedObject</see> event.
		/// </summary>
		/// <param name="e">The event args class containing further information of the event.</param>
		private void OnGeneratedObject(GeneratedObjectEventArgs e)
		{
			if (GeneratedObject != null)
				GeneratedObject(this, e);
		}

		/// <summary>
		/// Returns the list of interfaces manages by this instance.
		/// </summary>
		internal List<Interface> Interfaces
		{
			get 
			{
				return _interfaces; 
			}
		}

		/// <summary>
		/// Specifies the current instance of the <see cref="CSharpPersistentInterfacesProvider">CSharpPersistentInterfacesProvider</see>.
		/// The application can only have once a time.
		/// </summary>
		public static InterfacesProvider Current
		{
			get { return _current; }
			set { _current = value; }
		}
	}
}
