#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.Text;
using Chili.Opf3.Wizard.Shared;
using Chili.Opf3.Wizard.Shared.Storages;

#endregion

namespace Chili.Opf3.Wizard.CSharp.InterfacesProvider
{
	/// <summary>
	/// This class creates the relations for a persistent object.
	/// </summary>
	public sealed class RelationsCreator
	{
		private Template _template;
		private Project _project;
		private Table _table;

		/// <summary>
		/// Creates a new instance of the <see cref="RelationsCreator">RelationsCreator Class</see>.
		/// </summary>
		/// <param name="template">The template for the persistent creator.</param>
		/// <param name="project">The project of the wizard.</param>
		/// <param name="table">The current table that should be converted to a persistent.</param>
		public RelationsCreator(Project project, Table table, Template template)
		{
			if (template == null)
				throw new ArgumentNullException("template", "Template is null.");
			if (project == null)
				throw new ArgumentNullException("project", "Project is null.");
			if (table == null)
				throw new ArgumentNullException("table", "Table is null.");

			_template = template;
			_project = project;
			_table = table;
		}

		#region CreateRelations

		/// <summary>
		/// Creates the relations for the persistent object.
		/// </summary>
		/// <returns></returns>
		public string CreateRelations()
		{
			string result = string.Empty;

			// Loop over all relations.
			foreach (Relation relation in _project.Relations)
			{
				// Skip relations that should not be generated.
				if (!relation.Generate)
					continue;

				// Check whether we have to create the relation.
				if ((relation.ParentTable.TableName == _table.TableName && relation.CreateInParentPersistent) ||
					(relation.ChildTable.TableName == _table.TableName && relation.CreateInChildPersistent))
				{
					// Create the relation.
					string rel = CreateRelation(relation);
					// Skip if null.
					if (rel == null)
						continue;

					if (result.Length > 0)
						result += "\r\n\r\n";

					// Add the relation to the relations string.
					result += rel;
				}
			}

			if (result == string.Empty)
				return result;
			return string.Format(_template.GetValue("//relationRegion"), result);
		}

		#endregion

		#region CreateRelation

		/// <summary>
		/// Create one single relation for a given relation object.
		/// </summary>
		/// <param name="relation">The relation object that's relation is created in the persistent.</param>
		private string CreateRelation(Relation relation)
		{
			// Check which type of relation is given and create the appropriate relation.
			switch (relation.RelationType)
			{
				case RelationTypes.OneToMany:
					{
						if (relation.ParentTable.TableName == _table.TableName)
							return CreateObjectSetHolderRelation(relation);
						else
							return CreateObjectHolderRelation(relation);
					}
				case RelationTypes.OneToOne:
					return CreateObjectHolderRelation(relation);
				case RelationTypes.ManyToMany:
					return CreateObjectSetHolderRelation(relation);
				default:
					throw new NotSupportedException("Relationtype is not supported.");
			}
		}

		#endregion

		#region GetObjectHolderRelation

		/// <summary>
		/// Create a simple ObjectHolder relation.
		/// </summary>
		/// <param name="relation">The relation object.</param>
		private string CreateObjectHolderRelation(Relation relation)
		{
			bool parentFirst = false;

			// If the table name is the parent table we need to change the order in the relation and which table
			// is displayed as related.
			if (relation.ParentTable.TableName == _table.TableName)
				parentFirst = true;

			Table perTable = null;
			string relationProperty = null;
			if (!parentFirst)
			{
				perTable = relation.ParentTable;
				relationProperty = relation.ChildProperty;
			}
			else
			{
				perTable = relation.ChildTable;
				relationProperty = relation.ParentProperty;
			}

			// Compile the ObjectHolder property.
			return string.Format(_template.GetValue("//propertyObjectHolder"), perTable.PersistentName,
				relationProperty, InterfaceCreator.OPENING_BRACKETS, InterfaceCreator.CLOSING_BRACKETS);
		}

		#endregion

		#region CreateObjectSetHolderRelation

		/// <summary>
		/// Create a simple ObjectSetHolder relation.
		/// </summary>
		/// <param name="relation">The relation object.</param>
		private string CreateObjectSetHolderRelation(Relation relation)
		{
			// Get the table of the related object.
			Table perTable = relation.ChildTable;
			// Compile the property.
			return string.Format(_template.GetValue("//propertyObjectSetHolder"), perTable.PersistentName,
				relation.ParentProperty, InterfaceCreator.OPENING_BRACKETS, InterfaceCreator.CLOSING_BRACKETS);
		}

		#endregion
	}
}
