#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using Chili.Opf3.Wizard.Shared;
using Chili.Opf3.Wizard.Shared.Storages;
using System.Threading;
using System.IO;
using System.Reflection;
using System.Windows.Forms;
using Chili.Opf3.Wizard.Shared.UI;
using Chili.Opf3.Wizard.CSharp.PersistentProvider.UI;
using Chili.Opf3.Wizard.CSharp.Properties;

#endregion

namespace Chili.Opf3.Wizard.CSharp.PersistentProvider
{
	/// <summary>
	/// This class generates persistent classes in C#.
	/// </summary>
	public sealed class PersistentProvider : IPersistentObjectsProvider
	{
		private const string TEMPLATE_FILE_NAME = "Chili.Opf3.Wizard.CSharp.PersistentTemplate.xml";
		private const string PERSISTENT_FILE_NAME = "{0}\\{1}.cs";
		private const string PERSISTENT_GENERATED_FILE_NAME = "{0}\\{1}.Generated.cs";

        private List<Page> _pages = new List<Page>();
		private List<Persistent> _persistents;
		private StateBag _statebag = new StateBag();
		private Project _project;

		/// <summary>
		/// Creates a new instance of the <see cref="CSharpPersistentObjectsProvider">CSharpPersistentObjectsProvider Class</see>.
		/// </summary>
		public PersistentProvider()
		{
            _pages.Add(new ImplementsPage(this,_statebag));
            _pages.Add(new AdditionalDataPage(this, _statebag));
		}

		#region IPersistentObjectsProvider Members

        /// <summary>
		/// Generates the persistent objects form the project structure.
		/// </summary>
		/// <param name="project">The project containing all selected tables and fields of the database.</param>
		public void GenerateObjects(Project project)
		{
            foreach (Persistent persistent in _persistents)
			{
				// Skip if the table is not to generate.
                if (!persistent.Generate)
					continue;

				// Invoke the event generating the object.
                OnGeneratingObject(new GeneratingObjectEventArgs(persistent.ID));

				try
				{
					string directory = Path.GetDirectoryName(Assembly.GetAssembly(this.GetType()).Location);
					// Create the template.
					Template template = new Template(string.Format("{0}\\{1}", directory, TEMPLATE_FILE_NAME));

					// Create the persistent creator that creates one single persistent class.
                    PersistentCreator creator = new PersistentCreator(project, _statebag, persistent.Table, template);
					// Get the content of the persistent.
					string persistentContent = creator.GetPersistent();

					WritePersistentFile(persistentContent, project, persistent.Name);
#if !DEMO
					// Check if we have to write the empty persistent file (aka persistent user). Is written, when we have a .Generated file.
                    if (_statebag.CreateGenerated && _statebag.CreateEmptySilbing)
					{
                        string persistentUser = new PersistentUserCreator(project, persistent.Table, template).GetPersistent();
                        WritePersistentEmptyFile(persistentUser, project, persistent.Name);
					}
#endif
					// Invoke the event to signalize to the wizard that the object has been written.
                    OnGeneratedObject(new GeneratedObjectEventArgs(persistent.ID));
				}
				catch (Exception ex)
				{
					// Invoke the event to tell the wizard that the object couldn't be created.
                    OnGeneratedObject(new GeneratedObjectEventArgs(persistent.ID, ex));
				}
			}
		}

		/// <summary>
		/// Event is fired before the object is generated from the data of the <see cref="Project">Project</see> class.
		/// </summary>
		public event EventHandler<GeneratingObjectEventArgs> GeneratingObject;

		/// <summary>
		/// Event is fired after the object is generated from the data of the <see cref="Project">Project</see> class.
		/// </summary>
		public event EventHandler<GeneratedObjectEventArgs> GeneratedObject;

		/// <summary>
		/// Returns the name of the plug-in.
		/// </summary>
		public string Name
		{
			get { return Resources.Name; }
		}

		/// <summary>
		/// Returns the description for the plug-in.
		/// </summary>
		public string Description
        {
            get { return Resources.Description; }
        }

		/// <summary>
		/// Gets the language name supported by this tool. Or null if no explicit language
		/// is supported.
		/// </summary>
		public string LanguageName
		{
			get { return Resources.LanguageName; }
		}

		/// <summary>
		/// Returns a list of pages displayed by the class.
		/// </summary>
		public List<Page> Pages
        {
            get { return _pages; }
        }

		/// <summary>
		/// Returns the list of objects that are generated.
		/// </summary>
		public List<IGeneratedObject> GetObjectsList()
        {
            List<IGeneratedObject> list = new List<IGeneratedObject>();
            foreach (Persistent persistent in _persistents)
            {
				if (persistent.Generate)
					list.Add((IGeneratedObject)persistent);
            }
            return list;
        }

		/// <summary>
		/// Returns the version of the plug-in.
		/// </summary>
		public string Version
        {
			get { return this.GetType().Assembly.GetName().Version.ToString(); }
        }

		/// <summary>
		/// Display the about dialog for the plug-in.
		/// </summary>
		public void ShowAboutDialog()
        {
			MessageBox.Show(string.Format(Resources.PersistentProviderAboutText, "\r\n", this.Version), Resources.PersistentProviderAboutTitle,
				MessageBoxButtons.OK, MessageBoxIcon.Information);
		}

        /// <summary>
        /// Allow the provider to build internal data from Storage.
        /// </summary>
        public void ProcessStorage(Project project)
        {
            _project = project;

            // Creates the list of persistents.
            _persistents = new List<Persistent>();

            foreach (Table table in _project.Tables)
            {
                if (table.Generate)
                    _persistents.Add(new Persistent(table));
            }
        }

        /// <summary>
        /// Initialize defaults in a new project.
        /// </summary>
        public void InitializeDefaults(Project project)
        {
            // append the typeconversion pair List.
            // extract from the template.
            string directory = Path.GetDirectoryName(Assembly.GetAssembly(this.GetType()).Location);
            Template template = new Template(string.Format("{0}\\{1}", directory, TEMPLATE_FILE_NAME));

            XmlNodeList types = template.Document.SelectNodes("//types/type");
            foreach (XmlNode node in types)
            {
                TypeConversionPair pair = new TypeConversionPair(node.Attributes[0].Value, node.InnerText);
                pair.Initializer = template.GetValue(string.Format("//initializes/initialize[@id = '{0}']", pair.BaseType));
                pair.Language = Resources.LanguageName;

                project.TypeConverter.AppendPair(pair); 
            }
        }

        /// <summary>
        /// Deserialize the plug-in state from xml.
        /// </summary>
        public void Deserialize(Project project, XmlDocument xml, XmlNode d)
        {
            // Load the StateBag from the xml.
            _statebag.Deserialize(project, xml, d); 
        }

        /// <summary>
        /// Serialize the plug-in state in xml.
        /// </summary>
        public void Serialize(Project project, XmlDocument xml, XmlNode d)
        {
            d.AppendChild(_statebag.Serialize(xml));
        }

        #endregion


		/// <summary>
		/// Writes the persistent object to a file in the output directory.
		/// </summary>
		/// <param name="persistent">The persistent that is written to the file.</param>
		/// <param name="project">The current project.</param>
		/// <param name="persistentName">The name of the persistent object.</param>
		private void WritePersistentFile(string persistent, Project project, string persistentName)
		{
#if DEMO
			throw new NotSupportedException("Output not supported in Demo.");
#else
			// Get the path for the persistent object.
            string path = string.Format(_statebag.CreateGenerated ? PERSISTENT_GENERATED_FILE_NAME : PERSISTENT_FILE_NAME, project.OutputDirectoy, persistentName);

			// Tell the wizard that the file couldn't be overwritten.
			if (File.Exists(path))
				throw new NotSupportedException("File already exists. Please delete the existing file frist.");

			// Write the file.
			using (TextWriter writer = File.CreateText(path))
			{
				writer.WriteLine(persistent);
			}
#endif
		}

		/// <summary>
		/// Writes the partial persistent.Generated file in the output directory.
		/// </summary>
		/// <param name="persistentUser">The empty persistent object that is written to the file.</param>
		/// <param name="project">The current project.</param>
		/// <param name="persistentName">The name of the persistent object.</param>
		private void WritePersistentEmptyFile(string persistentUser, Project project, string persistentName)
		{
			// Get the path for the persistent object.
			string path = string.Format(PERSISTENT_FILE_NAME, project.OutputDirectoy, persistentName);

			if (File.Exists(path))
				return;

			// Write the file.
			using (TextWriter writer = File.CreateText(path))
			{
				writer.WriteLine(persistentUser);
			}
		}

		/// <summary>
		/// Invoked to fire the <see cref="GeneratingObject">GeneratingObject</see> event.
		/// </summary>
		/// <param name="e">The event args class containing further information of the event.</param>
		private void OnGeneratingObject(GeneratingObjectEventArgs e)
		{
			if (GeneratingObject != null)
				GeneratingObject(this, e);
		}

		/// <summary>
		/// Invoked to fire the <see cref="GeneratedObject">GeneratedObject</see> event.
		/// </summary>
		/// <param name="e">The event args class containing further information of the event.</param>
		private void OnGeneratedObject(GeneratedObjectEventArgs e)
		{
			if (GeneratedObject != null)
				GeneratedObject(this, e);
		}
    }
}
