#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using Chili.Opf3.Wizard.Shared.Storages;

#endregion

namespace Chili.Opf3.Wizard.Shared
{
    /// <summary>
    /// Manage the type conversion.
    /// </summary>
    public class TypeConverter
    {
        private List<TypeConversionPair> _list = new List<TypeConversionPair>();

        /// <summary>
        /// TypeConverter constructor.
        /// </summary>
        public TypeConverter()
        {
        }

        /// <summary>
        /// Add a pair to the internal list if not exist.
        /// </summary>
        public void AppendPair(TypeConversionPair pair)
        {
            foreach (TypeConversionPair p in _list)
            {
                if (pair.Compare(p))
                    return;
            }
            _list.Add(pair);
        }

        /// <summary>
        /// Retrieve the TypeConversionPair based on field and language. 
        /// If onlynamed=true ignore the pair with the condition for the name not specified. 
        /// </summary>
        private TypeConversionPair SearchMatchingPair(String language, Field field, Boolean onlynamed)
        {
            foreach (TypeConversionPair p in _list)
            {
                if((p.Language == language) && (p.BaseType==field.Type.ToString()))
                {
                    if (((p.NameCondition== null) || (p.NameCondition.Length == 0)) && onlynamed)
                        continue;

                    if ((p.NameCondition!=null) && (p.NameCondition.Length > 0))
                    {
                        Regex r = p.IgnoreCase ? new Regex(p.NameCondition, RegexOptions.IgnoreCase | RegexOptions.Singleline) 
                                               : new Regex(p.NameCondition,RegexOptions.Singleline);

                        if(!r.IsMatch(field.FieldName))
                            continue;
                    }
                    return p;
                }
            }
            return null;
        }

        /// <summary>
        /// Retrieve the TypeConversionPair based on field and language. First look to the pair with  
        /// the condition for the name specified. if none match search for all
        /// </summary>
        private TypeConversionPair GetPair(String language, Field field)
        {
            TypeConversionPair pair = SearchMatchingPair(language, field, true);

            if(pair==null)
                pair = SearchMatchingPair(language, field, false);

            return pair;
        }

        /// <summary>
        /// Retrieve the initilaizer based on basetype and namecondition.
        /// </summary>
        /// <param name="language">The language of the wizard.</param>
        /// <param name="field">The field  on wich initialisez is based.</param>
        public string GetTypeInitializer(String language, Field field)
        {
            TypeConversionPair pair = GetPair(language, field);
            if (pair != null)
                return pair.Initializer; 

            return "unknown";
        }

        /// <summary>
        /// Retrieve the type based on field and language.
        /// </summary>
        /// <param name="language">The language of the wizard.</param>
        /// <param name="field">The field type is based.</param>
        public string GetType(String language, Field field)
        {
            TypeConversionPair pair = GetPair(language, field);

            if (pair != null)
                return pair.TargetType;

            return "unknown";
        }


        /// <summary>
        /// Retrieve the type based on field and language and formatted for nullable supporting.
        /// </summary>
        /// <param name="language">The language of the wizard.</param>
        /// <param name="field">The field type is based.</param>
        /// <param name="format">The format string for the type (e.g. c# "{0}?"  VB "Nullable(Of {0})".</param>
        public string GetFormattedType(String language, Field field, string format)
        {
            string type = GetType(language, field);

            if (field.IsNullable)
                return string.Format(format, type);
            else
                return type;
        }



        /// <summary>
        /// Retrieve the internal TypeConversionPairs list.
        /// </summary>
        public List<TypeConversionPair> TypeConversionPairs
        {
            get { return _list; }
        }

        /// <summary>
        /// Return a list of a supported BaseTypes.
        /// </summary>
        public List<string> BaseTypes
        {
            get {
                List<String> types = new List<string>();
                foreach (TypeConversionPair type in _list)
                {
                    if (!types.Contains(type.BaseType))
                        types.Add(type.BaseType);
                }
                return types;
            }
        }

        /// <summary>
        /// Return a list of a supported Languages.
        /// </summary>
        public List<string> Languages
        {
            get
            {
                List<String> languages = new List<string>();
                foreach (TypeConversionPair type in _list)
                {
                    if (!languages.Contains(type.Language))
                        languages.Add(type.Language);
                }
                return languages;
            }
        }
    }
}
