#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.Text;
using Chili.Opf3.Wizard.Shared;
using System.Xml;
using Chili.Opf3.Wizard.Shared.Storages;

#endregion

namespace Chili.Opf3.Wizard.VB.InterfacesProvider
{
	/// <summary>
	/// This class serializes and deserializes a list of interface objects.
	/// </summary>
	internal sealed class InterfacesSerializer
	{
		private const string TABLE_REFERENCE = "tableReference";

		#region Serialize

		/// <summary>
		/// Serializes the list of interfaces given.
		/// </summary>
		/// <param name="interfaces">The interfaces that are serialized.</param>
        /// <param name="xml">The document to serialze.</param>
        public XmlNode Serialize(XmlDocument xml, List<Interface> interfaces)
		{
            if (interfaces == null)
                throw new ArgumentNullException("interfaces", "Interface is null.");
            if (xml == null)
                throw new ArgumentNullException("xml", "Xml is null.");


            XmlElement root = xml.CreateElement("interfaces");

            // Loops over all interfaces and serializes them.
            foreach (Interface i in interfaces)
            {
                root.AppendChild(SerializeInterface(xml, i));
            }

            return root;
		}

		#region SerializeInterface

		/// <summary>
		/// Serializes one single interface.
		/// </summary>
		/// <param name="iface">The interface that is serialized.</param>
		private XmlNode SerializeInterface(XmlDocument xml, Interface iface)
		{
			XmlElement i = xml.CreateElement("interface");
			i.AppendChild(CreateXmlElement(xml, "generate", iface.Generate.ToString()));
			i.AppendChild(CreateXmlElement(xml, "interfaceName", iface.InterfaceName));
			i.AppendChild(CreateTableReferenceXml(xml, iface.Table));

			return i;
		}

		#endregion

		#region Utility methods

		/// <summary>
		/// Creates the xml for a table reference.
		/// </summary>
		/// <param name="table">The table that is referenced.</param>
		private XmlNode CreateTableReferenceXml(XmlDocument xml, Table table)
		{
			XmlElement f = xml.CreateElement(TABLE_REFERENCE);
			XmlAttribute attribute = xml.CreateAttribute("table");
			attribute.Value = table.TableName;
			f.Attributes.Append(attribute);

			return f;
		}

		/// <summary>
		/// Creates an xml element with the given name and value.
		/// </summary>
		/// <param name="name">The name of the element.</param>
		/// <param name="value">The value of the element.</param>
		private XmlNode CreateXmlElement(XmlDocument xml, string name, string value)
		{
			XmlElement e = xml.CreateElement(name);
			e.AppendChild(xml.CreateTextNode(value));

			return e;
		}

		#endregion

		#endregion

		#region Deserialize

		/// <summary>
		/// Deserializes the given xml to a list of interfaces.
		/// </summary>
		/// <param name="value">The xml that is deserialized.</param>
        /// <param name="node">The node that is deserialized.</param>
		/// <param name="p">The project reference to dereference the tables.</param>
        public List<Interface> Deserialize(XmlDocument xml, XmlNode d, Project p)
		{
			List<Interface> list = new List<Interface>();

            XmlNodeList nodes = d.SelectNodes("./interfaces/interface");
            foreach (XmlNode node in nodes)
            {
                DeserializeInterface(node, p, list);
            }

            return list;
		}

		/// <summary>
		/// Deserializes the given node to an interface.
		/// </summary>
		/// <param name="node">The node that is deserialized.</param>
		/// <param name="p">The project used to dereference tables.</param>
		/// <param name="interfaces">The list of interfaces.</param>
		private void DeserializeInterface(XmlNode node, Project p, List<Interface> interfaces)
		{
			Table t = DereferenceTable(node.SelectSingleNode("./" + TABLE_REFERENCE), p);
			if (t == null)
				return;

			Interface i = new Interface(t);
			i.Generate = bool.Parse(node.SelectSingleNode("./generate").InnerText);
			i.InterfaceName = node.SelectSingleNode("./interfaceName").InnerText;

			interfaces.Add(i);
		}

		#region Utility methods

		/// <summary>
		/// Dereferences the node to a table. Or null if the table hasn't been found.
		/// </summary>
		/// <param name="refNode">The node that is dereferenced.</param>
		/// <param name="project">The project that contains the nodes to dereference.</param>
		private Table DereferenceTable(XmlNode refNode, Project project)
		{
			string t = refNode.Attributes["table"].Value;

			foreach (Table table in project.Tables)
			{
				if (table.TableName == t)
					return table;
			}

			return null;
		}

		#endregion

		#endregion
	}
}
