#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.Text;
using Chili.Opf3.Wizard.Shared;
using Chili.Opf3.Wizard.Shared.Storages;

#endregion

namespace Chili.Opf3.Wizard.VB.InterfacesProvider
{
	/// <summary>
	/// Creates the properties for the persistent object.
	/// </summary>
	public sealed class PropertiesCreator
	{
		private Template _template;
		private Project _project;
		private Table _table;

		/// <summary>
		/// Creates a new instance of the <see cref="PropertiesCreator">PropertiesCreator Class</see>.
		/// </summary>
		/// <param name="template">The template for the persistent creator.</param>
		/// <param name="project">The project of the wizard.</param>
		/// <param name="table">The current table that should be converted to a persistent.</param>
		public PropertiesCreator(Project project, Table table, Template template)
		{
			if (template == null)
				throw new ArgumentNullException("template", "Template is null.");
			if (project == null)
				throw new ArgumentNullException("project", "Project is null.");
			if (table == null)
				throw new ArgumentNullException("table", "Table is null.");

			_template = template;
			_project = project;
			_table = table;
		}

		#region CreateProperties

		/// <summary>
		/// Creates the properties of the persistent object.
		/// </summary>
		public string CreateProperties()
		{
			string result = string.Empty;

			foreach (Field field in _table.Fields)
			{
				// Skip fields that are not generated.
				if (!field.Generate)
					continue;

				string r = null;
				if (field.Identifier)
					r = CreateIdentifier(field);
				else
					r = CreateProperty(field);

				// If we got a NULL don't proceed.
				if (r == null)
					continue;

				// Add a new line if we are in the >= second property.
				if (result.Length > 0)
					result += "\r\n\r\n";
				result += r;
			}
			return string.Format(_template.GetValue("//propertiesRegion"), result);
		}

		#endregion

		#region CreateProperty

		/// <summary>
		/// Creates on single property.
		/// </summary>
		/// <param name="field">The field that is converted to a property.</param>
		private string CreateProperty(Field field)
		{
			// Get the property template.
			string property = _template.GetValue("//property");
			// Get the values for the comment.
			string mandatory = string.Empty;
			string length = string.Empty;
			if (!field.AllowDBNull)
				mandatory = _template.GetValue("//mandatoryComment");
			if (field.Type == typeof(string))
				length = string.Format(_template.GetValue("//lengthComment"), field.Length);

			// Compile the comment.
			string comment = string.Format(_template.GetValue("//propertyComment"), field.FieldName, length, mandatory);

			// Create the additional properties of the field attribute.
			string setAccessor = string.Empty;
			
			// Don't create the set accessor on for a blob or foreign key.
//			if (field.Type != typeof(byte[]) && !IsForeignKeyInRelation(field))
//				setAccessor = _template.GetValue("//setAccessor");

			return string.Format(property, comment, InterfaceCreator.GetVBType(_template, field), 
				field.PropertyName, setAccessor);
		}

		#endregion

		#region CreateIdentifier

		/// <summary>
		/// Creates the property as identifier.
		/// </summary>
		/// <param name="field">The field that is created as identifier.</param>
		private string CreateIdentifier(Field field)
		{
			// Get the property template.
			string property = _template.GetValue("//property");
			string comment = null;
			string setAccessor = null;

			// Check if the identifier is an autonumber.
			if (field.AutoNumber)
			{
				// Get the comment for the autonumber property.
				comment = _template.GetValue("//autoNumberComment");
				setAccessor = string.Empty;
			}
			else
			{
				// Get the comment for the identifier property.
				comment = _template.GetValue("//identifierComment");
				setAccessor = _template.GetValue("//setAccessor");
			}

			// Returns the compiled property.
			return string.Format(property, comment, InterfaceCreator.GetVBType(_template, field), field.PropertyName);
		}

		#endregion

		#region IsForeignKeyInRelation

		/// <summary>
		/// Returns whether the field is in a relation.
		/// </summary>
		/// <param name="field">The field that is inspected.</param>
		private bool IsForeignKeyInRelation(Field field)
		{
			// Loop over all relations.
			foreach (Relation relation in _project.Relations)
			{
				// Check if the relation's child table name is the name of the current table.
				if (relation.ChildTable.TableName == _table.TableName)
				{
					foreach (RelationItem relationItem in relation.Relations)
					{
						if (relationItem.ChildField.FieldName == field.FieldName)
							return true;
					}
				}

				// We have a many-to-many relation.
				if (relation.WeakTable != null && relation.WeakTable.TableName == _table.TableName)
				{
					// Check if in the many-to-many relation one parent field is the same as the fieldname of the field to check.
					foreach (RelationItem relationItem in relation.WeakRelations)
					{
						if (relationItem.ParentField.FieldName == field.FieldName)
							return true;
					}
				}
			}

			return false;
		}

		#endregion
	}
}
