#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using Chili.Opf3.Wizard.Shared;
using Chili.Opf3.Wizard.Shared.Storages;
using Chili.Opf3.Wizard.VB.Properties;
using Chili.Opf3.Wizard.Shared.UI;

#endregion

namespace Chili.Opf3.Wizard.VB.InterfacesProvider.UI
{
	/// <summary>
	/// This class allows to customize the interfaces created by the 
	/// <see cref="VBPersistentInterfacesProvider">VBPersistentInterfacesProvider</see>.
	/// </summary>
	public partial class InterfacePage : Page
	{
		private InterfacesProvider _provider;
		private Project _project;

		/// <summary>
		/// Creates a new instance of the <see cref="InterfacePage">InterfacePage Class</see>.
		/// </summary>
		/// <param name="provider">The provider itself.</param>
		public InterfacePage(InterfacesProvider provider) : base(provider)
		{
			InitializeComponent();

			if (provider == null)
				throw new ArgumentNullException("provider", "Provider is null.");
			_provider = provider;
		}

		/// <summary>
		/// Invoked each time the page is opened.
		/// </summary>
		/// <param name="project">The current project.</param>
		public override void PageOpening(Project project)
		{
			// Set the instance managed by this page as the current one.
			InterfacesProvider.Current = _provider;

			// Set the current project.
			_project = project;

			SetHeaderTitle(Resources.InterfacesHeaderTitle);
			SetHeaderDescription(Resources.InterfacesHeaderDescription);

			// Bind the interface list.
			BindInterfacesListToListView();
		}

		/// <summary>
		/// Binds the interfaces list to the listview.
		/// </summary>
		private void BindInterfacesListToListView()
		{
			// Clear the list.
			interfacesListView.Items.Clear();
			// Suspend the layout.
			interfacesListView.SuspendLayout();

			try
			{
				foreach (Interface i in _provider.Interfaces)
				{
					ListViewItem item = new ListViewItem();
					// Populate the list view item.
					PopulateListViewItem(i, item);
					// Add the item to the listview.
					interfacesListView.Items.Add(item);
				}
			}
			finally
			{
				// Resume the layout.
				interfacesListView.ResumeLayout();
			}
		}

		/// <summary>
		/// Populate one list view item with data.
		/// </summary>
		/// <param name="i">The interface itself. This interface is used to populate the item.</param>
		/// <param name="item">The listview item that is populated.</param>
		private void PopulateListViewItem(Interface i, ListViewItem item)
		{
			// Clear all subitems.
			item.SubItems.Clear();
            item.Tag = i;

			item.Checked = i.Generate;
			item.ImageIndex = 0;
			item.Text = i.Table.PersistentName;
			item.SubItems.Add(i.InterfaceName);
		}

		/// <summary>
		/// Invoked each time the page is closed.
		/// </summary>
		/// <param name="next">True if the next button has been clicked. False on all other.</param>
		public override bool PageClosing(bool next)
		{
			// Loop over all interfaces and set the checked state.
			for (int i = 0; i < _provider.Interfaces.Count; i++)
			{
				_provider.Interfaces[i].Generate = interfacesListView.Items[i].Checked;
			}
			return true;
		}

		/// <summary>
		/// Invoked each time the check all button is clicked.
		/// </summary>
		private void checkAllButton_Click(object sender, EventArgs e)
		{
			// Loop over all items and uncheck.
			for (int i = 0; i < interfacesListView.Items.Count; i++)
			{
				interfacesListView.Items[i].Checked = true;
			}
		}

		/// <summary>
		/// Invoked each time the uncheck all button is clicked.
		/// </summary>
		private void unCheckAllButton_Click(object sender, EventArgs e)
		{
			// Loop over all items and uncheck.
			for (int i = 0; i < interfacesListView.Items.Count; i++)
			{
				interfacesListView.Items[i].Checked = false;
			}
		}

		/// <summary>
		/// Invoked each time the edit button is clicked.
		/// </summary>
		private void editButton_Click(object sender, EventArgs e)
		{
			Interface i = GetSelectedInterface();
			if (i == null)
				return;

			// Check if we got an OK.
			if (new InterfaceDialog(i).ShowDialog(this) == DialogResult.OK)
			{
				// Update the current selected item in the list.
				ListViewItem item = interfacesListView.SelectedItems[0];
				PopulateListViewItem(i, item);
			}
		}

		/// <summary>
		/// Returns the selected interface or null if none is selected.
		/// </summary>
		private Interface GetSelectedInterface()
		{
			if (interfacesListView.SelectedItems.Count == 0)
				return null;

			return _provider.Interfaces[interfacesListView.SelectedIndices[0]];
		}

		/// <summary>
		/// Invoked each time the selected index of the list view changed.
		/// </summary>
		private void interfacesListView_SelectedIndexChanged(object sender, EventArgs e)
		{
			editButton.Enabled = (interfacesListView.SelectedItems.Count > 0);
		}
	}
}
