#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.Text;
using Chili.Opf3.Wizard.Shared.Storages;
using Chili.Opf3.Wizard.Shared;
using Chili.Opf3.Wizard.VB.Properties;

#endregion

namespace Chili.Opf3.Wizard.VB.PersistentProvider
{
	/// <summary>
	/// Creates the interfaces implementation for a persistent object.
	/// </summary>
	public sealed class InterfacesImplementationGenerator
	{
		private Template _template;
		private Table _table;
		private Project _project;
        private StateBag _statebag;

		/// <summary>
		/// Creates a new instance of the <see cref="InterfacesImplementationGenerator">InterfacesImplementationGenerator Class</see>.
		/// </summary>
		/// <param name="project">The project of the wizard.</param>
		/// <param name="template">The template for the persistent creator.</param>
		/// <param name="table">The current table that should be converted to a persistent.</param>
        /// <param name="statebag">The plugin's statebag.</param>
        public InterfacesImplementationGenerator(Project project, StateBag statebag, Table table, Template template)
		{
			if (template == null)
				throw new ArgumentNullException("template", "Template is null.");
			if (project == null)
				throw new ArgumentNullException("project", "Project is null.");
			if (table == null)
				throw new ArgumentNullException("table", "Table is null.");
            if (statebag == null)
                throw new ArgumentNullException("statebag", "StateBag is null.");

			_template = template;
			_table = table;
            _statebag = statebag;
			_project = project;
		}

		#region CreateInterfacesImplementation

		/// <summary>
		/// Create the implementation for the interfaces.
		/// </summary>
		public string CreateInterfacesImplementation()
		{
			string result = string.Empty;

            if (_statebag.IDynamicExtensible)
				result += CreateIDynamicExtensible();
            if (_statebag.IObjectNotification)
				result += CreateIObjectNotification();
            if (_statebag.IPopulateHelper)
				result += CreateIPopulateHelper();
            if (_statebag.ISelfContainingObject)
				result += CreateISelfContainingObject();
            if (_statebag.INotifyPropertyChanged)
				result += CreateINotifyPropertyChanged();

			return result;
		}

		#endregion

		#region INotifyPropertyChanged

		/// <summary>
		/// Creates the INotifyPropertyChanged interface for the persistent object.
		/// </summary>
		private string CreateINotifyPropertyChanged()
		{
			return _template.GetValue("//iNotifyPropertyChanged");
		}

		#endregion

		#region CreateIDynamicExtensible

		/// <summary>
		/// Creates the IDynamicExtensible interface for the persistent object.
		/// </summary>
		private string CreateIDynamicExtensible()
		{
			return _template.GetValue("//iDynamicExtensible");
		}

		#endregion

		#region CreateIObjectNotification

		/// <summary>
		/// Creates the IObjectNotification interface for the persistent object.
		/// </summary>
		private string CreateIObjectNotification()
		{
			return _template.GetValue("//iObjectNotification");
		}

		#endregion

		#region CreateIPopulateHelper

		/// <summary>
		/// Creates the IPopulateHelper interface for the persistent object.
		/// </summary>
		private string CreateIPopulateHelper()
		{
			string setterPart = CreateSetter();
			string getterPart = CreateGetter();

			return string.Format(_template.GetValue("//iPopulateHelper"), getterPart, setterPart);
		}

		/// <summary>
		/// Creates the getter part of the interface.
		/// </summary>
		private string CreateGetter()
		{
			string result = string.Empty;

			foreach (Field field in _table.Fields)
			{
				if (result.Length > 0)
					result += "\r\n";

				result += string.Format(_template.GetValue("//getterCase"), field.PropertyName);
			}

			return result;
		}

		/// <summary>
		/// Creates the setter part of the interface.
		/// </summary>
		/// <returns></returns>
		private string CreateSetter()
		{
			string result = string.Empty;

			foreach (Field field in _table.Fields)
			{
				if (result.Length > 0)
					result += "\r\n";

				result += string.Format(_template.GetValue("//setterCase"), field.PropertyName,
                    _project.TypeConverter.GetFormattedType(Resources.LanguageName, field, "Nullable(Of {0})"));
			}

			return result;
		}

		#endregion

		#region CreateISelfContainingObject

		/// <summary>
		/// Creates the ISelfContainingObject interface for the persistent object.
		/// </summary>
		/// <returns></returns>
		private string CreateISelfContainingObject()
		{
			return _template.GetValue("//iSelfContainingObject");
		}

		#endregion
	}
}
