#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.Text;
using Chili.Opf3.Wizard.Shared;
using Chili.Opf3.Wizard.Shared.Storages;
using VBPersistentInterfacesProvider = Chili.Opf3.Wizard.VB.InterfacesProvider.InterfacesProvider;
using Interface = Chili.Opf3.Wizard.VB.InterfacesProvider.Interface;

#endregion

namespace Chili.Opf3.Wizard.VB.PersistentProvider
{
	/// <summary>
	/// Creates a persistent object.
	/// </summary>
	public sealed class PersistentCreator
	{
		public const string OPENING_BRACKETS = "{";
		public const string CLOSING_BRACKETS = "}";
		public const string TRUE = "True";
		public const string FALSE = "False";

		private Template _template;
		private Project _project;
		private Table _table;
        private StateBag _statebag;

		/// <summary>
		/// Creates a new instance of the <see cref="PersistentCreator">PersistentCreator Class</see>.
		/// </summary>
		/// <param name="template">The template for the persistent creator.</param>
		/// <param name="project">The project of the wizard.</param>
		/// <param name="table">The current table that should be converted to a persistent.</param>
        /// <param name="statebag">The plugin's statebag.</param>
        public PersistentCreator(Project project, StateBag statebag, Table table, Template template)
		{
			if (template == null)
				throw new ArgumentNullException("template", "Template is null.");
			if (project == null)
				throw new ArgumentNullException("project", "Project is null.");
			if (table == null)
				throw new ArgumentNullException("table", "Table is null.");
            if (statebag == null)
                throw new ArgumentNullException("statebag", "StateBag is null.");


			_template = template;
			_project = project;
            _statebag = statebag;
			_table = table;
		}

		/// <summary>
		/// Returns the persistent object's file content.
		/// </summary>
		public string GetPersistent()
		{
            string fields = new FieldsCreator(_project, _table, _template).CreateFields();
			string properties = new PropertiesCreator(_project, _statebag, _table, _template).CreateProperties();
            string relations = new RelationsCreator(_project, _statebag, _table, _template).CreateRelations();

			// Get the persistent template.
			string persistent = _template.GetValue("//persistent");

			// Get the size of the pool for the persistent.
			string poolSize = string.Empty;
			if (_table.PoolSize > 0 && !_table.NoPool)
				poolSize = string.Format(", {0}:={1}", _template.GetValue("//poolSize"), _table.PoolSize);
			// Get the rights of the persistent.
			string rights = GetRights();

			// Compiled the attributes.
			string attributes = string.Format("{0}{1}", poolSize, rights);

			// Create the header for the interfaces.
			string interfacesHeader = CreateInterfacesHeader();
			// Create the body of the interfaces.
			string interfacesBody = new InterfacesImplementationGenerator(_project, _statebag, _table, _template).CreateInterfacesImplementation();

			// Create the body of the persistent object.
			string body = string.Format("{0}{1}{2}{3}", fields, relations, properties, interfacesBody);

			string seal = string.Empty;
            if (_statebag.PersistentSealed)
				seal = _template.GetValue("//sealed");
			string partial = string.Empty;
            if (_statebag.CreateGenerated)
				partial = _template.GetValue("//partial");

			// Return the compiled template.
			return string.Format(persistent, _table.PersistentName, _project.AutorInfo.Name, _project.AutorInfo.Copyright,
				DateTime.Now.ToShortDateString(), _table.TableName, attributes, seal, partial,
                _table.PersistentName, interfacesHeader, body, _statebag.AdditionalUsingDirective);
		}

		#region CreateInterfacesHeader

		/// <summary>
		/// Create the headers for the interfaces.
		/// </summary>
		private string CreateInterfacesHeader()
		{
			string result = string.Empty;

            if ((_table.BaseClass != null) && (_table.BaseClass.Length > 0))
                result += CreatePersistentBaseClassHeader(_table.BaseClass, result);
            else if ((_statebag.BaseClass != null) && (_statebag.BaseClass.Length > 0))
                result += CreatePersistentBaseClassHeader(_statebag.BaseClass, result);

            if (_statebag.IDynamicExtensible)
				result = CreateInterfaceHeader("//iDynamicExtensibleName", result);
            if (_statebag.IObjectNotification)
				result = CreateInterfaceHeader("//iObjectNotificationName", result);
            if (_statebag.IPopulateHelper)
				result = CreateInterfaceHeader("//iPopulateHelperName", result);
            if (_statebag.ISelfContainingObject)
				result = CreateInterfaceHeader("//iSelfContainingObjectName", result);
            if (_statebag.INotifyPropertyChanged)
				result = CreateInterfaceHeader("//iNotifyPropertyChangedName", result);

			result = CreatePersistentInterfaceHeader(result);

			return result;
		}

        //// <summary>
        /// Adds the persistent base class to the header.
        /// </summary>
        /// <param name="baseclass">The baseclass to build.</param>
        /// <param name="header">The currently build header.</param>
        private string CreatePersistentBaseClassHeader(string baseclass, string header)
		{
            if ((baseclass == null) || baseclass.Length == 0)
                return "";

			if (header.Length > 0)
                header += string.Format(", {0}", baseclass);
			else
                header += string.Format("\r\n\timplements {0}", baseclass);

			return header;
		}

        /// <summary>
        /// Adds the persistent interface to the header.
        /// </summary>
        /// <param name="header">The currently build header.</param>
        private string CreatePersistentInterfaceHeader(string header)
        {
            string value = GetPersistentInterfaceName();
            if (value == null)
                return header;

            if (header.Length > 0)
                header += string.Format(", {0}", value);
            else
                header += string.Format("\r\n\timplements {0}", value);

            return header;
        }

		/// <summary>
		/// Returns the persistent interface name or NULL if none has been found.
		/// </summary>
		/// <returns></returns>
		private string GetPersistentInterfaceName()
		{
			// Return null if no provider is selected.
			if (VBPersistentInterfacesProvider.Current == null)
				return null;

			foreach (Interface i in VBPersistentInterfacesProvider.Current.Interfaces)
			{
				if (i.Table == _table)
					return i.InterfaceName;
			}
			return null;
		}

		/// <summary>
		/// Create the interface header for the given xpath.
		/// </summary>
		/// <param name="xpath">The xpath to get the name of the interface.</param>
		/// <param name="header">The currently build header.</param>
		private string CreateInterfaceHeader(string xpath, string header)
		{
			string value = _template.GetValue(xpath);
			if (header.Length > 0)
				header += string.Format(", {0}", value);
			else
				header += string.Format("\r\n\timplements {0}", value);

			return header;
		}

		#endregion

		#region GetRights

		/// <summary>
		/// Returns the string for the rights on the persistent.
		/// </summary>
		private string GetRights()
		{
			// Return an empty string if all rights are given.
			if (_table.Rights == PersistentRights.All)
				return string.Empty;

			string result = string.Format(", {0}:=", _template.GetValue("//rights"));
			bool second = false;

			if ((_table.Rights & PersistentRights.Delete) != 0)
			{
				result += _template.GetValue("//persistentRights/delete");
				second = true;
			}
			if ((_table.Rights & PersistentRights.Insert) != 0)
			{
				if (second)
                    result += " Or ";
				result += _template.GetValue("//persistentRights/insert");
				second = true;
			}
			if ((_table.Rights & PersistentRights.Load) != 0)
			{
				if (second)
                    result += " Or ";
				result += _template.GetValue("//persistentRights/load");
				second = true;
			}
			if ((_table.Rights & PersistentRights.Update) != 0)
			{
				if (second)
					result += " Or ";
				result += _template.GetValue("//persistentRights/update");
				second = true;
			}

			// Set as no rights if we have not had any right until now.
			if (!second)
			{
				result += _template.GetValue("//persistentRights/none");
			}

			return result;
		}

		#endregion

	}
}
