#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.Text;
using Chili.Opf3.Wizard.Shared;
using Chili.Opf3.Wizard.Shared.Storages;
using Chili.Opf3.Wizard.VB.Properties;

#endregion

namespace Chili.Opf3.Wizard.VB.PersistentProvider
{
	/// <summary>
	/// Creates the properties for the persistent object.
	/// </summary>
	public sealed class PropertiesCreator
	{
		private Template _template;
		private Project _project;
		private Table _table;
        private StateBag _statebag;

		/// <summary>
		/// Creates a new instance of the <see cref="PropertiesCreator">PropertiesCreator Class</see>.
		/// </summary>
		/// <param name="template">The template for the persistent creator.</param>
		/// <param name="project">The project of the wizard.</param>
		/// <param name="table">The current table that should be converted to a persistent.</param>
        /// <param name="statebag">The plugin's statebag.</param>
        public PropertiesCreator(Project project, StateBag statebag, Table table, Template template)
		{
			if (template == null)
				throw new ArgumentNullException("template", "Template is null.");
			if (project == null)
				throw new ArgumentNullException("project", "Project is null.");
			if (table == null)
				throw new ArgumentNullException("table", "Table is null.");
            if (statebag == null)
                throw new ArgumentNullException("statebag", "StateBag is null.");

			_template = template;
			_project = project;
            _statebag = statebag;
			_table = table;
		}

		#region CreateProperties

		/// <summary>
		/// Creates the properties of the persistent object.
		/// </summary>
		public string CreateProperties()
		{
			string result = string.Empty;

			foreach (Field field in _table.Fields)
			{
				// Skip fields that are not generated.
				if (!field.Generate)
					continue;

				// Add a new line if we are in the >= second property.
				if (result.Length > 0)
					result += "\r\n\r\n";

				if (field.Identifier)
					result += CreateIdentifier(field);
				else
					result += CreateProperty(field);

			}
			return string.Format(_template.GetValue("//propertiesRegion"), result);
		}

		#endregion

		#region CreateProperty

		/// <summary>
		/// Creates on single property.
		/// </summary>
		/// <param name="field">The field that is converted to a property.</param>
		private string CreateProperty(Field field)
		{
			// Get the property template.
			string property = _template.GetValue("//property");
			string comment = null;
			string virt = null;

			string visibility = null;
			// Check if the property is a foreign key. Set that property private.
			if (IsForeignKeyInRelation(field))
			{
				// Foreign keys are private.
				visibility = _template.GetValue("//private") + " ";
				comment = _template.GetValue("//foreignKeyComment");

				// Foreign keys are private. They can't be virtual.
				virt = string.Empty;
			}
			else
			{
				visibility = string.Empty;

				// Get the values for the comment.
				string mandatory = string.Empty;
				string length = string.Empty;
				if (!field.AllowDBNull)
					mandatory = _template.GetValue("//mandatoryComment");
				if (field.Type == typeof(string))
					length = string.Format(_template.GetValue("//lengthComment"), field.Length);

				// Compile the comment.
				comment = string.Format(_template.GetValue("//propertyComment"), field.FieldName, length, mandatory);

				if (_statebag.PersistentPropertiesVirtual)
					virt = _template.GetValue("//virtual");
			}

			// Create the additional properties of the field attribute.
			string attributes = CreateFieldAttributeProperties(field);
			string setAccessor = string.Empty;
			string convertedType = _project.TypeConverter.GetFormattedType(Resources.LanguageName, field, "{0}?");
			string initializer = _project.TypeConverter.GetTypeInitializer(Resources.LanguageName, field);

			// Don't create the set accessor for fields with initilizer.
			//if ((initializer == null) || (initializer == ""))
			setAccessor = CreateSetAccessor(field, visibility);

			return string.Format(property, comment, field.FieldName, attributes, _template.GetValue("//public"), virt,
                _project.TypeConverter.GetFormattedType(Resources.LanguageName, field, "Nullable(Of {0})"), field.PropertyName, CreateGetAccessor(field), setAccessor);
		}

		#endregion

		#region CreateFieldAttributeProperties

		/// <summary>
		/// Creates the field attribute properties.
		/// </summary>
		/// <param name="field">The current field.</param>
		private string CreateFieldAttributeProperties(Field field)
		{
			string result = string.Empty;

			// Check if the property for AllowDBNull is to set.
			if (!field.AllowDBNull)
				result = string.Format(", {0}:={1}", _template.GetValue("//allowDBNull"), PersistentCreator.FALSE);

			bool second = false;

			// Check the persist behaviour of the property.
			if ((field.PersistBehaviour & PersistBehaviours.DontInsert) != 0)
			{
				if (!second)
					result += string.Format(", {0}:=", _template.GetValue("//persistBehaviours"));

				result += _template.GetValue("//fieldAttribute/dontInsert");
				second = true;
			}
			if ((field.PersistBehaviour & PersistBehaviours.DontUpdate) != 0)
			{
				if (second)
					result += " | ";
				else
					result += string.Format(", {0}:=", _template.GetValue("//persistBehaviours"));

				result += _template.GetValue("//fieldAttribute/dontUpdate");
			}

			return result;
		}

		#endregion

		#region CreateIdentifier

		/// <summary>
		/// Creates the property as identifier.
		/// </summary>
		/// <param name="field">The field that is created as identifier.</param>
		private string CreateIdentifier(Field field)
		{
			// Get the property template.
			string property = _template.GetValue("//property");
			string comment = null;
			// Compiles the attributes.
			string attributes = string.Format(", {0}:={1}, {2}:={3}", _template.GetValue("//allowDBNull"), PersistentCreator.FALSE,
				_template.GetValue("//identifier"), PersistentCreator.TRUE);
			string setAccessor = null;
			string pub = _template.GetValue("//visibility/public");
			string priv = _template.GetValue("//visibility/private") + " ";

			// Check if the identifier is an autonumber.
			if (field.AutoNumber)
			{
				// Get the comment for the autonumber property.
				comment = _template.GetValue("//autoNumberComment");
				attributes += string.Format(", {0}:={1}", _template.GetValue("//autoNumber"), PersistentCreator.TRUE);
				setAccessor = CreateSetAccessor(field, priv);
			}
			else
			{
				// Get the comment for the identifier property.
				comment = _template.GetValue("//identifierComment");
				setAccessor = CreateSetAccessor(field, string.Empty);
			}

			string virt = string.Empty;
			if (_statebag.PersistentPropertiesVirtual)
				virt = _template.GetValue("//virtual");

			// Returns the compiled property.
            return string.Format(property, comment, field.FieldName, attributes, pub, virt, _project.TypeConverter.GetFormattedType(Resources.LanguageName, field, "Nullable(Of {0})"),
				field.PropertyName, CreateGetAccessor(field), setAccessor);
		}

		#endregion

		#region CreateGetAccessor

		/// <summary>
		/// Creates the get accessor for the property.
		/// </summary>
		/// <param name="field">The field for which the get accessor is created.</param>
		private string CreateGetAccessor(Field field)
		{
			string getAccessor = _template.GetValue("//getAccessor");
			return string.Format(getAccessor, field.ObjectFieldName);
		}

		#endregion

		#region CreateSetAccessor

		/// <summary>
		/// Creates the set accessor for the property.
		/// </summary>
		/// <param name="field">The field for which the set accessor is created.</param>
		/// <param name="modifier">The modifier of the field.</param>
		private string CreateSetAccessor(Field field, string modifier)
		{
			string setAccessor = _template.GetValue("//setAccessor");
			string argumentNullException = string.Empty;
			string argumentException = string.Empty;
			string onRowChanged = string.Empty;

			if (field.Type == typeof(string) && !field.AllowDBNull)
				argumentNullException = _template.GetValue("//argumentNullException");
			if (field.Type == typeof(string))
				argumentException = string.Format(_template.GetValue("//argumentException"), field.Length);

            if (_statebag.INotifyPropertyChanged)
			{
				onRowChanged = string.Format(_template.GetValue("//onRowChanged"), field.PropertyName);
			}

			return string.Format(setAccessor, modifier, argumentNullException,
                argumentException, field.ObjectFieldName, onRowChanged, _project.TypeConverter.GetFormattedType(Resources.LanguageName, field, "Nullable(Of {0})"));
		}

		#endregion

		#region IsForeignKeyInRelation

		/// <summary>
		/// Returns whether the field is in a relation.
		/// </summary>
		/// <param name="field">The field that is inspected.</param>
		private bool IsForeignKeyInRelation(Field field)
		{
			// Loop over all relations.
			foreach (Relation relation in _project.Relations)
			{
				// Check if the relation's child table name is the name of the current table.
				if (relation.ChildTable.TableName == _table.TableName)
				{
					foreach (RelationItem relationItem in relation.Relations)
					{
						if (relationItem.ChildField.FieldName == field.FieldName)
							return true;
					}
				}

				// We have a many-to-many relation.
				if (relation.WeakTable != null && relation.WeakTable.TableName == _table.TableName)
				{
					// Check if in the many-to-many relation one parent field is the same as the fieldname of the field to check.
					foreach (RelationItem relationItem in relation.WeakRelations)
					{
						if (relationItem.ParentField.FieldName == field.FieldName)
							return true;
					}
				}
			}

			return false;
		}

		#endregion
	}
}
