#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.Text;
using Chili.Opf3.Wizard.Shared;
using Chili.Opf3.Wizard.Shared.Storages;

#endregion

namespace Chili.Opf3.Wizard.VB.PersistentProvider
{
	/// <summary>
	/// This class creates the relations for a persistent object.
	/// </summary>
	public sealed class RelationsCreator
	{
		private Template _template;
		private Project _project;
		private Table _table;
        private StateBag _statebag;

		/// <summary>
		/// Creates a new instance of the <see cref="RelationsCreator">RelationsCreator Class</see>.
		/// </summary>
		/// <param name="template">The template for the persistent creator.</param>
		/// <param name="project">The project of the wizard.</param>
		/// <param name="table">The current table that should be converted to a persistent.</param>
        /// <param name="statebag">The plugin's statebag.</param>
        public RelationsCreator(Project project, StateBag statebag, Table table, Template template)
		{
			if (template == null)
				throw new ArgumentNullException("template", "Template is null.");
			if (project == null)
				throw new ArgumentNullException("project", "Project is null.");
			if (table == null)
				throw new ArgumentNullException("table", "Table is null.");
            if (statebag == null)
                throw new ArgumentNullException("statebag", "StateBag is null.");

			_template = template;
			_project = project;
            _statebag = statebag;
			_table = table;
		}

		#region CreateRelations

		/// <summary>
		/// Creates the relations for the persistent object.
		/// </summary>
		/// <returns></returns>
		public string CreateRelations()
		{
			string result = string.Empty;

			// Loop over all relations.
			foreach (Relation relation in _project.Relations)
			{
				// Skip relations that should not be generated.
				if (!relation.Generate)
					continue;

				// Check whether we have to create the relation.
				if ((relation.ParentTable.TableName == _table.TableName && relation.CreateInParentPersistent) ||
					(relation.ChildTable.TableName == _table.TableName && relation.CreateInChildPersistent))
				{
					// Create the relation.
					string rel = CreateRelation(relation);
					// Skip if null.
					if (rel == null)
						continue;

					if (result.Length > 0)
						result += "\r\n\r\n";

					// Add the relation to the relations string.
					result += rel;
				}
			}

			if (result == string.Empty)
				return result;
			return string.Format(_template.GetValue("//relationRegion"), result);
		}

		#endregion

		#region CreateRelation

		/// <summary>
		/// Create one single relation for a given relation object.
		/// </summary>
		/// <param name="relation">The relation object that's relation is created in the persistent.</param>
		private string CreateRelation(Relation relation)
		{
			// Check which type of relation is given and create the appropriate relation.
			switch (relation.RelationType)
			{
				case RelationTypes.OneToMany:
					{
						if (relation.ParentTable.TableName == _table.TableName)
							return CreateObjectSetHolderRelation(relation);
						else
							return CreateObjectHolderRelation(relation);
					}
				case RelationTypes.OneToOne:
					return CreateObjectHolderRelation(relation);
				case RelationTypes.ManyToMany:
					return CreateObjectSetHolderRelation(relation);
				default:
					throw new NotSupportedException("Relationtype is not supported.");
			}
		}

		#endregion

		#region GetObjectHolderRelation

		/// <summary>
		/// Create a simple ObjectHolder relation.
		/// </summary>
		/// <param name="relation">The relation object.</param>
		private string CreateObjectHolderRelation(Relation relation)
		{
			bool parentFirst = false;

			// If the table name is the parent table we need to change the order in the relation and which table
			// is displayed as related.
			if (relation.ParentTable.TableName == _table.TableName)
				parentFirst = true;

			string rel = CreateAttributeRelation(relation.Relations, parentFirst);
			// Return null if the relationattribute is null.
			if (rel == null)
				return null;

			string persistRelationShip = string.Empty;
			// Get the table of the first table.
			Table perTable = null;
			string relationProperty = null;
			if (!parentFirst)
			{
				// In this case we have to create a relationship property.
				persistRelationShip = string.Format(", {0}", _template.GetValue("//childFirst"));
				perTable = relation.ParentTable;
				relationProperty = relation.ChildProperty;
			}
			else
			{
				perTable = relation.ChildTable;
				relationProperty = relation.ParentProperty;
			}

			// Compile the whole relation attribute.
            string relationAttribute = string.Format(_template.GetValue("//simpleRelation"), rel, persistRelationShip, CreatePersistBehaviourProperty(relation));

			string fieldName = GetSmartFieldName(relationProperty);

			string virt = string.Empty;
            if (_statebag.PersistentPropertiesVirtual)
				virt = _template.GetValue("//virtual");

			// Compile the holder.
			string holder = string.Format(_template.GetValue("//objectHolder"), relationAttribute,
				perTable.PersistentName, fieldName);
			// ... and the property.
			string propertyObjectHolder = string.Format(_template.GetValue("//propertyObjectHolder"), virt, perTable.PersistentName,
				relationProperty, fieldName, perTable.PersistentName);

			return string.Format("{0}\r\n\r\n{1}", holder, propertyObjectHolder);
		}

		#endregion

        #region CreatePersistBehaviourProperty

        /// <summary>
        /// Creates the PersistBehaviour part of the Relation Attribute.
        /// </summary>
        /// <param name="relation">The relation object.</param>
        private string CreatePersistBehaviourProperty(Relation relation)
        {
            string persistBehaviour = string.Empty;
            bool second = false;

            // Check the persist behaviour of the property.
            if ((relation.PersistBehaviour & PersistBehaviours.DontInsert) != 0)
            {
                if (!second)
                    persistBehaviour += string.Format(", {0} = ", _template.GetValue("//persistBehaviours"));

                persistBehaviour += _template.GetValue("//fieldAttribute/dontInsert");
                second = true;
            }
            if ((relation.PersistBehaviour & PersistBehaviours.DontUpdate) != 0)
            {
                if (second)
                    persistBehaviour += " or ";
                else
                    persistBehaviour += string.Format(", {0} = ", _template.GetValue("//persistBehaviours"));

                persistBehaviour += _template.GetValue("//fieldAttribute/dontUpdate");
            }

            return persistBehaviour;
        }

        #endregion

		#region CreateAttributeRelation

		/// <summary>
		/// Creates the relation of the attribute and returns it.
		/// </summary>
		/// <param name="list">The list of which the relation is created.</param>
		/// <param name="parentFirst">Specifies if the parent field is the first in the relation.</param>
		private string CreateAttributeRelation(List<RelationItem> list, bool parentFirst)
		{
			string result = string.Empty;
			foreach (RelationItem item in list)
			{
				if (result.Length > 0)
					result += "; ";

				if (parentFirst)
					result += string.Format("{0} = {1}", item.ParentField.PropertyName, item.ChildField.PropertyName);
				else
					result += string.Format("{0} = {1}", item.ChildField.PropertyName, item.ParentField.PropertyName);
			}
			return result;
		}

		#endregion

		#region CreateObjectSetHolderRelation

		/// <summary>
		/// Create a simple ObjectSetHolder relation.
		/// </summary>
		/// <param name="relation">The relation object.</param>
		private string CreateObjectSetHolderRelation(Relation relation)
		{
			// Compile the whole relation attribute.
			if (relation.WeakTable == null)
				return CreateOneToManyObjectSetHolderRelation(relation);
			else
				return CreateManyToManyObjectSetHolderRelation(relation);
		}

		#endregion

		#region CreateManyToManyObjectSetHolderRelation

		/// <summary>
		/// Create a Many-To-Many ObjectSetHolder relation.
		/// </summary>
		/// <param name="relation">The relation object.</param>
		private string CreateManyToManyObjectSetHolderRelation(Relation relation)
		{
			string rel = null;
			string relationAttribute = null;
			Table perTable = null;
			string fieldName = null;
			string propertyName = null;

			if (relation.ParentTable.TableName == _table.TableName)
			{
				// Get the relation from the parent table to the weak table.
				rel = CreateAttributeRelation(relation.Relations, true);
				// Return null if the relationattribute is null.
				if (rel == null)
					return null;
				// Get the weak relation (from the weak to the child table).
				string weakRel = CreateAttributeRelation(relation.WeakRelations, true);
				if (weakRel == null)
					return null;
                relationAttribute = string.Format(_template.GetValue("//manyToManyRelation"), rel, relation.WeakTable.PersistentName, weakRel,
                    CreatePersistBehaviourProperty(relation));

				// Get the table of the related object.
				perTable = relation.ChildTable;
				// Get the field name.
				fieldName = GetSmartFieldName(relation.ParentProperty);
				propertyName = relation.ParentProperty;
			}
			else
			{
				// Get the relation from the child table to the weak table.
				rel = CreateAttributeRelation(relation.Relations, false);
				// Return null if the relationattribute is null.
				if (rel == null)
					return null;
				// Get the weak relation (from the weak to the parent table).
				string weakRel = CreateAttributeRelation(relation.WeakRelations, false);
				if (weakRel == null)
					return null;
                relationAttribute = string.Format(_template.GetValue("//manyToManyRelation"), weakRel, relation.WeakTable.PersistentName, rel,
                    CreatePersistBehaviourProperty(relation));

				// Get the table of the related object.
				perTable = relation.ParentTable;
				// Get the field name.
				fieldName = GetSmartFieldName(relation.ChildProperty);
				propertyName = relation.ChildProperty;
			}

			string virt = string.Empty;
            if (_statebag.PersistentPropertiesVirtual)
				virt = _template.GetValue("//virtual");

			// Compile the holder.
			string holder = string.Format(_template.GetValue("//objectSetHolder"), relationAttribute, perTable.PersistentName, fieldName);
			// ... and the property.
			string propertyObjectSetHolder = string.Format(_template.GetValue("//propertyObjectSetHolder"), virt, perTable.PersistentName,
				propertyName, fieldName);

			return string.Format("{0}\r\n\r\n{1}", holder, propertyObjectSetHolder);
		}

		#endregion

		#region CreateOneToManyObjectSetHolderRelation

		/// <summary>
		/// Create a One-To-Many ObjectSetHolder relation.
		/// </summary>
		/// <param name="relation">The relation object.</param>
		private string CreateOneToManyObjectSetHolderRelation(Relation relation)
		{
			// Get the relation from the parent to the child table.
			string rel = CreateAttributeRelation(relation.Relations, true);
			// Return null if the relationattribute is null.
			if (rel == null)
				return null;

            string relationAttribute = string.Format(_template.GetValue("//simpleRelation"), rel, string.Empty, CreatePersistBehaviourProperty(relation));

			// Get the table of the related object.
			Table perTable = relation.ChildTable;
			// Get the field name.
			string fieldName = GetSmartFieldName(relation.ParentProperty);

			string virt = string.Empty;
            if (_statebag.PersistentPropertiesVirtual)
				virt = _template.GetValue("//virtual");

			// Compile the holder.
			string holder = string.Format(_template.GetValue("//objectSetHolder"), relationAttribute,
				perTable.PersistentName, fieldName);
			// ... and the property.
			string propertyObjectSetHolder = string.Format(_template.GetValue("//propertyObjectSetHolder"), virt, perTable.PersistentName,
				relation.ParentProperty, fieldName);

			return string.Format("{0}\r\n\r\n{1}", holder, propertyObjectSetHolder);
		}

		#endregion

		/// <summary>
		/// Returns a field name for the persistent name.
		/// </summary>
		/// <param name="persistentName">The name of the persistent object.</param>
		private string GetSmartFieldName(string persistentName)
		{
			return string.Format("{0}{1}", persistentName[0].ToString().ToLower(), persistentName.Substring(1));
		}
	}
}
