#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.IO;
using System.Xml;
using System.Xml.Serialization;
using System.Data.OleDb;
using System.Windows.Forms;  

using Chili.Opf3.Wizard.Shared;
using Chili.Opf3.Wizard.Properties;
using System.Text.RegularExpressions;

#endregion

namespace Chili.Opf3.Wizard
{
    /// <summary>
    /// This class loads /save the project from the corresponding xml file.
    /// </summary>
    public class ProjectManager
    {
        private Project _project = new Project();
        private string _filename;
        private XmlDocument _xml;

        /// <summary>
        /// Returns the Filename.
        /// </summary>
        public string Filename
        {
            get { return _filename; }
        }

		/// <summary>
		/// Returns the current open project. Or null if none is open.
		/// </summary>
		public Project Project
		{
			get { return _project; }
		}

		/// <summary>
		/// Creates a new project in the project manager. The current open project is removed.
		/// </summary>
		public void NewProject()
		{
			_project = new Project();
			_filename = null;
			// Don't change this here: Avoids the user to pop up with a save dialog if nothing has changed!
            _xml = SerializeProject();
		}

		#region Open

		/// <summary>
        /// Open project file and load it.
        /// </summary>
        public bool Open()
        {
            OpenFileDialog openfile = new OpenFileDialog();
			openfile.Filter = Resources.WizardProjectFilter;
            openfile.RestoreDirectory = true;

            if (openfile.ShowDialog() == DialogResult.OK)
            {
                _filename = openfile.FileName;
				return LoadFile();
            }
            return false;
		}

		/// <summary>
		/// Opens the given project file and loads it.
		/// </summary>
		/// <param name="fileName">The file name of the project file.</param>
		public bool Open(string fileName)
		{
			_filename = fileName;
			return LoadFile();
		}

		#endregion

		#region LoadFile

		/// <summary>
        /// Load project from file.
        /// </summary>
		private bool LoadFile()
        {
            try
			{
				ProjectSerializer serializer = new ProjectSerializer();

				// Store the current project as XMLdocument.
                XmlDocument xml = new XmlDocument();
                xml.Load(_filename);

				// Deserializes the project.
                _project = serializer.Deserialize(xml);

                // save the base xml: is old plugin's informations.
                _xml = xml;

                return true;
            }
            catch (Exception ex)
            {
				MessageBox.Show(string.Format(Resources.CouldNotLoadProjectDescription, ex.Message), Resources.ProgramName, 
					MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }
		}

		#endregion

		#region Save

		/// <summary>
        /// Save project to file.
        /// </summary>
        public bool Save()
        {
            // Serialize the project to a XmlDocument.
            XmlDocument xml = SerializeProject();
            if (xml == null)
				return false;

            // Check if the project is changed.
			if ((_xml != null) && (Regex.Replace(_xml.OuterXml, @"<\?.+?\?>", "") == xml.OuterXml))
				return true;

			DialogResult result = MessageBox.Show(Resources.SaveProjectDescription, Resources.ProgramName, MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question);
			// Check if we need to leave the current method.
			switch (result)
			{
				case DialogResult.Cancel:
					return false;
				case DialogResult.No:
					return true;
			}

			if (_filename == null)
			{
				SaveFileDialog d = new SaveFileDialog();
				d.RestoreDirectory = true;
				d.Filter = Resources.WizardProjectFilter;

				if (DialogResult.OK == d.ShowDialog())
					_filename = d.FileName;
				else
					return false;
			}

			// Persist the changes.
            if (PersistFile(xml))
			{
                // old the document.
                _xml = xml;
				return true;
			}
			return false;
		}

		#endregion

		#region PersistFile

		/// <summary>
		/// Persists the changes on the project.
		/// </summary>
		/// <param name="serializedProject">The project that is saved.</param>
        private bool PersistFile(XmlDocument xml)
		{
			try
			{
                XmlWriterSettings settings = new XmlWriterSettings();
                settings.Indent = true;
                settings.NewLineOnAttributes = true;

                using (XmlWriter writer = XmlWriter.Create(_filename,settings))
				{
                    xml.WriteContentTo(writer);
				}

				return true;
			}
			catch (Exception ex)
			{
				MessageBox.Show(string.Format(Resources.CouldNotSaveProjectDescription, ex.Message), Resources.ProgramName,
					MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}
		}

		#endregion

		#region SerializeProject

		/// <summary>
        /// Serializes the project and returns a XmlDocument that is the serialize project.
		/// Null is returned if not possible to serialize.
		/// </summary>
		/// <returns></returns>
        private XmlDocument SerializeProject()
		{
			try
			{
				ProjectSerializer serializer = new ProjectSerializer();
				return serializer.Serialize(_project);
			}
			catch (Exception ex)
			{
				MessageBox.Show(string.Format(Resources.CouldNotSaveProjectDescription, ex.Message), Resources.ProgramName,
					MessageBoxButtons.OK, MessageBoxIcon.Error);
				return null;
			}
		}

		#endregion

        /// <summary>
        /// Specifies the XmlDocument the project is based to, null if a new project.
        /// </summary>
        public XmlDocument Xml
        {
            get { return _xml; }
        }

        /// <summary>
        /// Retrieve the XmlNode for the specifierd specified Specifies the XmlDocument the project is based to, null if a new project.
        /// </summary>
        public XmlNode GetObjectProviderNode(IObjectsProvider provider)
        {
            if (_xml == null) return null;

            if (provider == null)
                throw new ArgumentNullException("provider", "ObjectsProvider is null.");

            ProjectSerializer serializer = new ProjectSerializer();

            return serializer.GetObjectProviderNode(_xml,provider);
        }
	}
}
