#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.Text;
using Chili.Opf3.Wizard.Shared;
using System.Xml;
using Chili.Opf3.Wizard.Shared.Storages;
using System.Reflection;
using System.Xml.Serialization;
using System.IO;

#endregion

namespace Chili.Opf3.Wizard
{
	/// <summary>
	/// Serializes and deserializes a given project.
	/// </summary>
	internal sealed class ProjectSerializer
	{
        private const string FIELD_REFERENCE = "fieldReference";
		private const string TABLE_REFERENCE = "tableReference";

		#region Serialize

		/// <summary>
		/// Serializes the given project.
		/// </summary>
		/// <param name="project">The project that is serialized.</param>
		public XmlDocument Serialize(Project project)
		{
			if (project == null)
				throw new ArgumentNullException("project", "Project is null.");

			XmlDocument xml = new XmlDocument();

			XmlElement root = xml.CreateElement("opf3WizProject");
			XmlAttribute version = xml.CreateAttribute("version");
			version.Value = Assembly.GetExecutingAssembly().GetName().Version.ToString();
			root.Attributes.Append(version);
			xml.AppendChild(root);

			root.AppendChild(SerializeGeneralProject(xml, project));
			root.AppendChild(SerializeStorage(xml, project.Storage));
			root.AppendChild(SerializeAutorInfo(xml, project.AutorInfo));
			root.AppendChild(SerializeTables(xml, project.Tables));
			root.AppendChild(SerializeRelations(xml, project.Relations));
            root.AppendChild(SerializeTypeConversionPairs(xml, project.TypeConverter.TypeConversionPairs));

            root.AppendChild(SerializeObjectProviders(xml, project, Session.Current.ObjectProvider));

			return xml;
		}

		#region SerializeObjectProviders

		/// <summary>
        /// Creates the xml for each ObjectProvider.
        /// </summary>
        /// <param name="ObjectProvider">The collection that is serialized.</param>
		private XmlNode SerializeObjectProviders(XmlDocument xml, Project project, List<IObjectsProvider> objectProviders)
        {
            XmlElement d = xml.CreateElement("objectproviders");

			foreach (IObjectsProvider objectprovider in objectProviders)
            {
                // Creates the xml for the dynamic property.
                XmlElement o = xml.CreateElement("objectprovider");
                XmlAttribute attribute = xml.CreateAttribute("type");
                attribute.Value = objectprovider.GetType().FullName;
                o.Attributes.Append(attribute);

                XmlElement v = xml.CreateElement("value");
                o.AppendChild(v);

                // tell the objectprovider to append their state to the 'v' node.
                objectprovider.Serialize(project, xml, v);

                d.AppendChild(o); 
            }

            return d;
        } 
        #endregion

		#region SerializeRelations

		/// <summary>
		/// Creates the xml for the relations.
		/// </summary>
		/// <param name="list">The list containing the relations.</param>
		private XmlNode SerializeRelations(XmlDocument xml, List<Relation> list)
		{
			XmlElement r = xml.CreateElement("relations");

			if (list != null)
			{
				foreach (Relation relation in list)
				{
					// Creates the xml for the relation.
					r.AppendChild(SerializeRelation(xml, relation));
				}
			}

			return r;
		}

		/// <summary>
		/// Creates the xml for one single relation.
		/// </summary>
		/// <param name="relation">The relation that is transformed to xml.</param>
		private XmlNode SerializeRelation(XmlDocument xml, Relation relation)
		{
			XmlElement r = xml.CreateElement("relation");

            r.AppendChild(CreateXmlElement(xml, "name", relation.Name));
			r.AppendChild(CreateXmlElement(xml, "createInChildPersistent", relation.CreateInChildPersistent.ToString()));
			r.AppendChild(CreateXmlElement(xml, "createInParentPersistent", relation.CreateInParentPersistent.ToString()));
			r.AppendChild(CreateXmlElement(xml, "generate", relation.Generate.ToString()));
			r.AppendChild(CreateXmlElement(xml, "persistBehaviour", relation.PersistBehaviour.ToString()));
			r.AppendChild(CreateXmlElement(xml, "relationType", relation.RelationType.ToString()));

			XmlElement parent = xml.CreateElement("parent");
			parent.AppendChild(CreateXmlElement(xml, "parentProperty", relation.ParentProperty));
			parent.AppendChild(CreateTableReferenceXml(xml, relation.ParentTable));
			r.AppendChild(parent);

			XmlElement child = xml.CreateElement("child");
			child.AppendChild(CreateXmlElement(xml, "childProperty", relation.ChildProperty));
			child.AppendChild(CreateTableReferenceXml(xml, relation.ChildTable));
			r.AppendChild(child);

			XmlElement weak = xml.CreateElement("weak");
			if (relation.WeakTable != null)
			{
				weak.AppendChild(CreateTableReferenceXml(xml, relation.WeakTable));
			}
			r.AppendChild(weak);

			// Creates the xml for the relation items.
			r.AppendChild(SerializeSubRelations(xml, relation.Relations, relation));
			// Creates the xml for the weak relation items.
			r.AppendChild(SerializeWeakSubRelations(xml, relation.WeakRelations, relation));

			return r;
		}

		/// <summary>
		/// Creates the xml for the weak relation items.
		/// </summary>
		/// <param name="list">The relation items that are transformed to xml.</param>
		/// <param name="relation">The relation.</param>
		private XmlNode SerializeWeakSubRelations(XmlDocument xml, List<RelationItem> list, Relation relation)
		{
			XmlElement r = xml.CreateElement("weakRelationItems");

			foreach (RelationItem item in list)
			{
				// Creates the xml for the relation items.
				r.AppendChild(SerializeRelationItem(xml, item, relation.WeakTable, relation.ChildTable));
			}

			return r;
		}

		/// <summary>
		/// Creates the xml for the relation items.
		/// </summary>
		/// <param name="list">The relation items that are transformed to xml.</param>
		/// <param name="relation">The relation.</param>
		private XmlNode SerializeSubRelations(XmlDocument xml, List<RelationItem> list, Relation relation)
		{
			XmlElement r = xml.CreateElement("relationItems");

			foreach (RelationItem item in list)
			{
				Table t = (relation.RelationType == RelationTypes.ManyToMany ? relation.WeakTable : relation.ChildTable);

				// Creates the xml for the relation items.
				r.AppendChild(SerializeRelationItem(xml, item, relation.ParentTable, t));
			}

			return r;
		}

		/// <summary>
		/// Creates the xml for the relation item.
		/// </summary>
		/// <param name="list">The relation item that are transformed to xml.</param>
		private XmlNode SerializeRelationItem(XmlDocument xml, RelationItem item, Table parentTable, Table childTable)
		{
			XmlElement r = xml.CreateElement("relationItem");

			r.AppendChild(CreateFieldReferenceXml(xml, parentTable, item.ParentField));
			r.AppendChild(CreateFieldReferenceXml(xml, childTable, item.ChildField));

			return r;
		}

		#endregion

		#region SerializeTables

		/// <summary>
		/// Creates the xml for the tables.
		/// </summary>
		/// <param name="list">The list containing the tables.</param>
		private XmlNode SerializeTables(XmlDocument xml, List<Table> list)
		{
			XmlElement t = xml.CreateElement("tables");

			if (list != null)
			{
				foreach (Table table in list)
				{
					// Creates the xml for the table.
					t.AppendChild(SerializeTable(xml, table));
				}
			}

			return t;
		}

		/// <summary>
		/// Creates the xml for one single table.
		/// </summary>
		/// <param name="table">The table that is transformed to xml.</param>
		private XmlNode SerializeTable(XmlDocument xml, Table table)
		{
			XmlElement t = xml.CreateElement("table");

			t.AppendChild(CreateXmlElement(xml, "generate", table.Generate.ToString()));
			t.AppendChild(CreateXmlElement(xml, "isView", table.IsView.ToString()));
			t.AppendChild(CreateXmlElement(xml, "noPool", table.NoPool.ToString()));
			t.AppendChild(CreateXmlElement(xml, "persistentName", table.PersistentName));
			t.AppendChild(CreateXmlElement(xml, "poolSize", table.PoolSize.ToString()));
			t.AppendChild(CreateXmlElement(xml, "rights", table.Rights.ToString()));
			t.AppendChild(CreateXmlElement(xml, "tableName", table.TableName));
            t.AppendChild(CreateXmlElement(xml, "BaseClass", table.BaseClass));

			// Creates the xml for the fields.
			t.AppendChild(SerializeFields(xml, table.Fields));
			// Create the xml for the indexes.
			t.AppendChild(SerializeIndexes(xml, table.Indexes, table));

			return t;
		}

		#endregion

        #region SerializeTypeConversionPairs

        /// <summary>
        /// Creates the xml for the ypeConversionPairs.
        /// </summary>
        /// <param name="list">The list containing the TypeConversionPair.</param>
        private XmlNode SerializeTypeConversionPairs(XmlDocument xml, List<TypeConversionPair> list)
        {
            XmlElement i = xml.CreateElement("typesconversion");

            // Loops over all searchers and serializes them.
            foreach (TypeConversionPair t in list)
            {
                XmlElement s = xml.CreateElement("typeconversion");

                s.AppendChild(CreateXmlElement(xml, "language", t.Language));
                s.AppendChild(CreateXmlElement(xml, "basetype", t.BaseType));
                s.AppendChild(CreateXmlElement(xml, "targettype", t.TargetType));
                s.AppendChild(CreateXmlElement(xml, "initializer", t.Initializer));
                s.AppendChild(CreateXmlElement(xml, "namecondition", t.NameCondition));
                s.AppendChild(CreateXmlElement(xml, "ignorecase", t.IgnoreCase.ToString()));
                i.AppendChild(s);
            }
            return i;
        }
        #endregion

		#region SerializeIndexes

		/// <summary>
		/// Creates the xml for the indexes.
		/// </summary>
		/// <param name="list">The list containing the indexes.</param>
		/// <param name="table">The table containing the fields for the indexes.</param>
		private XmlNode SerializeIndexes(XmlDocument xml, List<Index> list, Table table)
		{
			XmlElement i = xml.CreateElement("indexes");

			foreach (Index index in list)
			{
				// Creates the xml for the index.
				i.AppendChild(SerializeIndex(xml, index, table));
			}

			return i;
		}

		/// <summary>
		/// Creates the xml for the index.
		/// </summary>
		/// <param name="index">The index that is transformed to xml.</param>
		/// <param name="table">The table containing the fields for the indexes.</param>
		private XmlNode SerializeIndex(XmlDocument xml, Index index, Table table)
		{
			XmlElement i = xml.CreateElement("index");

			i.AppendChild(CreateXmlElement(xml, "name", index.Name));
			i.AppendChild(CreateXmlElement(xml, "generate", index.Generate.ToString()));
			i.AppendChild(CreateXmlElement(xml, "primaryKey", index.PrimaryKey.ToString()));

			// Creates the xml for the fields of the indexes.
			i.AppendChild(SerializeIndexFields(xml, index.Fields, table));

			return i;
		}

		/// <summary>
		/// Creates the xml for the field's index.
		/// </summary>
		/// <param name="list">The list containing the fields for the index.</param>
		/// <param name="table">The table containing the fields for the index.</param>
		private XmlNode SerializeIndexFields(XmlDocument xml, List<Field> list, Table table)
		{
			XmlElement fields = xml.CreateElement("fields");

			foreach (Field field in list)
			{
				fields.AppendChild(CreateFieldReferenceXml(xml, table, field));
			}

			return fields;
		}

		#endregion

		#region SerializeFields

		/// <summary>
		/// Creates the xml for the fields.
		/// </summary>
		/// <param name="list">The list containing the fields.</param>
		private XmlNode SerializeFields(XmlDocument xml, List<Field> list)
		{
			XmlElement f = xml.CreateElement("fields");

			foreach (Field field in list)
			{
				// Creates the xml for the fields.
				f.AppendChild(SerializeField(xml, field));
			}

			return f;
		}

		/// <summary>
		/// Creats the xml for a single field.
		/// </summary>
		/// <param name="field">The field that is transformed to xml.</param>
		private XmlNode SerializeField(XmlDocument xml, Field field)
		{
			XmlElement f = xml.CreateElement("field");

			f.AppendChild(CreateXmlElement(xml, "allowDBNull", field.AllowDBNull.ToString()));
			f.AppendChild(CreateXmlElement(xml, "autoNumber", field.AllowDBNull.ToString()));
			f.AppendChild(CreateXmlElement(xml, "fieldName", field.FieldName));
			f.AppendChild(CreateXmlElement(xml, "description", field.Description));
			f.AppendChild(CreateXmlElement(xml, "generate", field.Generate.ToString()));
			f.AppendChild(CreateXmlElement(xml, "identifier", field.Identifier.ToString()));
			f.AppendChild(CreateXmlElement(xml, "length", field.Length.ToString()));
			f.AppendChild(CreateXmlElement(xml, "objectFieldName", field.ObjectFieldName));
			f.AppendChild(CreateXmlElement(xml, "persistBehaviour", field.PersistBehaviour.ToString()));
			f.AppendChild(CreateXmlElement(xml, "propertyName", field.PropertyName));
			f.AppendChild(CreateXmlElement(xml, "type", field.Type.ToString()));

			return f;
		}

		#endregion

		#region SerializeGeneralProject

		/// <summary>
		/// Create the general xml for the project.
		/// </summary>
		/// <param name="project">The general options that are converted to xml.</param>
		private XmlNode SerializeGeneralProject(XmlDocument xml, Project project)
		{
			XmlElement g = xml.CreateElement("general");

			g.AppendChild(CreateXmlElement(xml, "namespace", project.Namespace));
			g.AppendChild(CreateXmlElement(xml, "outputDirectory", project.OutputDirectoy));
			g.AppendChild(CreateXmlElement(xml, "renameItemsToSingular", project.RenameItemsToSingular.ToString()));
            g.AppendChild(CreateXmlElement(xml, "removefirstchars", project.RemoveFirstChars.ToString()));
			g.AppendChild(SerializeLastRetrievedStorageSchema(xml, project.LastRetrievedStorageSchema));

			return g;
		}

		/// <summary>
		/// Creates the xml for the last retrieved storage schema.
		/// </summary>
		/// <param name="storageSchema">The storage schema that is transformed to xml.</param>
		private XmlNode SerializeLastRetrievedStorageSchema(XmlDocument xml, StorageSchema storageSchema)
		{
			XmlElement l = xml.CreateElement("lastRetrievedStorageSchema");

			if (storageSchema != null)
			{
				string x = string.Format("<schema>{0}{1}</schema>", SerializeTables(xml, storageSchema.Tables).OuterXml, SerializeRelations(xml, storageSchema.Relations).OuterXml);
				l.AppendChild(xml.CreateCDataSection(x));
			}
			
			return l;
		}

		#endregion

		#region SerializeAutorInfo

		/// <summary>
		/// Creates the xml for the autorinfo.
		/// </summary>
		/// <param name="autorInfo">The autorinfo object that is transformed to xml.</param>
		private XmlNode SerializeAutorInfo(XmlDocument xml, AutorInfo autorInfo)
		{
			XmlElement a = xml.CreateElement("autorInfo");

			a.AppendChild(CreateXmlElement(xml, "copyright", autorInfo.Copyright));
			a.AppendChild(CreateXmlElement(xml, "name", autorInfo.Name));

			return a;
		}

		#endregion

		#region SerializeStorage

		/// <summary>
		/// Creates the xml for the storage.
		/// </summary>
		/// <param name="storage">The storage that is transformed to xml.</param>
		private XmlNode SerializeStorage(XmlDocument xml, Storage storage)
		{
			XmlElement s = xml.CreateElement("storage");

			if (storage != null)
			{
				s.AppendChild(CreateXmlElement(xml, "alias", storage.Alias));
				s.AppendChild(CreateXmlElement(xml, "connectionString", storage.ConnectionString));
			}

			return s;
		}

		#endregion

		#region Utility methods

		/// <summary>
		/// Creates the xml for a field reference.
		/// </summary>
		/// <param name="table">The table of the field.</param>
		/// <param name="field">The field that is referenced.</param>
		private XmlNode CreateFieldReferenceXml(XmlDocument xml, Table table, Field field)
		{
			XmlElement f = xml.CreateElement(FIELD_REFERENCE);
			XmlAttribute attribute = xml.CreateAttribute("table");
			attribute.Value = table.TableName;
			f.Attributes.Append(attribute);

			attribute = xml.CreateAttribute("field");
			attribute.Value = field.FieldName;
			f.Attributes.Append(attribute);

			return f;
		}

		/// <summary>
		/// Creates the xml for a table reference.
		/// </summary>
		/// <param name="table">The table that is referenced.</param>
		private XmlNode CreateTableReferenceXml(XmlDocument xml, Table table)
		{
			XmlElement f = xml.CreateElement(TABLE_REFERENCE);
			XmlAttribute attribute = xml.CreateAttribute("table");
			attribute.Value = table.TableName;
			f.Attributes.Append(attribute);

			return f;
		}

		/// <summary>
		/// Creates an xml element with the given name and value.
		/// </summary>
		/// <param name="name">The name of the element.</param>
		/// <param name="value">The value of the element.</param>
		private XmlNode CreateXmlElement(XmlDocument xml, string name, string value)
		{
			XmlElement e = xml.CreateElement(name);
			e.AppendChild(xml.CreateTextNode(value));

			return e;
		}

        /// <summary>
        /// Returns the inner text for the given node. Prevents an exception to be raised if the node does not exist.
        /// </summary>
        private string GetNodeInnerText(XmlNode n, string path, string value)
        {
            XmlNode node = n.SelectSingleNode(path);
            return (node == null) ? value : node.InnerText;
        }

		#endregion

		#endregion

		#region Deserialize

		/// <summary>
        /// Deserializes the given XmlDocument and returns the deserialized project.
		/// </summary>
        /// <param name="value">The XmlDocument that is deserialized.</param>
        public Project Deserialize(XmlDocument xml)
		{
            if (xml == null)
                throw new ArgumentNullException("xml", "xml is null.");

			// Creates a new project.
			Project p = new Project();

			XmlNode root = xml.SelectSingleNode("//opf3WizProject");

			DeserializeGeneralProject(root, p);
			DeserializeStorage(root, p);
			DeserializeAutorInfo(root, p);
			DeserializeTables(root, p);
			DeserializeRelations(root, p);
            DeserializeTypeConversionPairs(root, p.TypeConverter);

			return p;
		}

        #region DeserializeIObjectsProvider

        /// <summary>
        /// Deserializes the IObjectsProvider  xml.
        /// </summary>
        public XmlNode GetObjectProviderNode(XmlDocument xml, IObjectsProvider provider)
        {
			XmlNode root = xml.SelectSingleNode("//opf3WizProject");

			XmlNodeList nodes = root.SelectNodes("./objectproviders/objectprovider");
            foreach (XmlNode node in nodes)
            {
                string type = node.Attributes["type"].Value;
                if (type==provider.GetType().FullName)
                    return node.SelectSingleNode("./value");
            }

            return null;
        }
        #endregion

        #region DeserializeRelations

		/// <summary>
		/// Deserializes the relations xml.
		/// </summary>
		private void DeserializeRelations(XmlNode root, Project p)
		{
			XmlNodeList nodes = root.SelectNodes("./relations/relation");

			if (nodes.Count > 0)
			{
				// Add an empty list that holds the relations.
				p.Relations = new List<Relation>();

				foreach (XmlNode n in nodes)
				{
					DeserializeRelation(n, p);
				}
			}
		}

		/// <summary>
		/// Deserialize the xml of a single relation.
		/// </summary>
		private void DeserializeRelation(XmlNode n, Project p)
		{
			Relation relation = new Relation();
			p.Relations.Add(relation);

            relation.Name = GetNodeInnerText(n, "./name", null);
            relation.CreateInChildPersistent = bool.Parse(GetNodeInnerText(n,"./createInChildPersistent","False"));
            relation.CreateInParentPersistent = bool.Parse(GetNodeInnerText(n, "./createInParentPersistent", "False"));
            relation.Generate = bool.Parse(GetNodeInnerText(n, "./generate","True"));
            relation.PersistBehaviour = (PersistBehaviours)Enum.Parse(typeof(PersistBehaviours), GetNodeInnerText(n, "./persistBehaviour",null));
            relation.RelationType = (RelationTypes)Enum.Parse(typeof(RelationTypes), GetNodeInnerText(n, "./relationType",null));

            relation.ParentProperty = GetNodeInnerText(n, "./parent/parentProperty",null);
            relation.ParentTable = DereferenceTable(n.SelectSingleNode("./parent/" + TABLE_REFERENCE), p);

            relation.ChildProperty = GetNodeInnerText(n, "./child/childProperty",null);
            relation.ChildTable = DereferenceTable(n.SelectSingleNode("./child/" + TABLE_REFERENCE), p);

			XmlNode weak = n.SelectSingleNode("./weak/" + TABLE_REFERENCE);
			if (weak != null)
			{
				relation.WeakTable = DereferenceTable(weak, p);
			}

			// Deserializes the relation items.
			DeserializeSubRelations(n, relation, p);
			// Deserializes the weak relation items.
			DeserializeWeakSubRelations(n, relation, p);
		}

		/// <summary>
		/// Deserializes the weak sub relation items.
		/// </summary>
		private void DeserializeWeakSubRelations(XmlNode n, Relation relation, Project p)
		{
			XmlNodeList nodes = n.SelectNodes("./weakRelationItems/relationItem");
			foreach (XmlNode node in nodes)
			{
				RelationItem item = new RelationItem();

				item.ParentField = DereferenceTableField(node.SelectNodes("./" + FIELD_REFERENCE)[0], p);
				item.ChildField = DereferenceTableField(node.SelectNodes("./" + FIELD_REFERENCE)[1], p);
				
				relation.WeakRelations.Add(item);
			}
		}

		/// <summary>
		/// Deserializes the sub relation items.
		/// </summary>
		private void DeserializeSubRelations(XmlNode n, Relation relation, Project p)
		{
			XmlNodeList nodes = n.SelectNodes("./relationItems/relationItem");
			foreach (XmlNode node in nodes)
			{
				RelationItem item = new RelationItem();

				item.ParentField = DereferenceTableField(node.SelectNodes("./" + FIELD_REFERENCE)[0], p);
				item.ChildField = DereferenceTableField(node.SelectNodes("./" + FIELD_REFERENCE)[1], p);

				relation.Relations.Add(item);
			}
		}

		#endregion

		#region DeserializeTables

		/// <summary>
		/// Deserializes the tables xml.
		/// </summary>
		private void DeserializeTables(XmlNode root, Project p)
		{
			XmlNodeList nodes = root.SelectNodes("./tables/table");

			if (nodes.Count > 0)
			{
				// Add an empty list that holds the tables.
				p.Tables = new List<Table>();

				foreach (XmlNode n in nodes)
				{
					DeserializeTable(n, p);
				}
			}
		}

		/// <summary>
		/// Deserializes the single tables xml.
		/// </summary>
		private void DeserializeTable(XmlNode n, Project p)
		{
			Table table = new Table();
			p.Tables.Add(table);

			table.Generate = bool.Parse(GetNodeInnerText(n,"./generate","False"));
            table.IsView = bool.Parse(GetNodeInnerText(n, "./isView", "False"));
            table.NoPool = bool.Parse(GetNodeInnerText(n, "./noPool", "False"));
			table.PersistentName = GetNodeInnerText(n,"./persistentName",null);
			table.PoolSize = int.Parse(GetNodeInnerText(n,"./poolSize","0"));
            table.Rights = (PersistentRights)Enum.Parse(typeof(PersistentRights), GetNodeInnerText(n, "./rights", null));
            table.TableName = GetNodeInnerText(n, "./tableName", null);
            table.BaseClass = GetNodeInnerText(n, "./BaseClass", null);

			// Deserializes all fields of the table.
			DeserializeFields(n, table);
			// Deserializes all indexes of the table.
			DeserializeIndexes(n, table, p);
		}

		#endregion

        #region DeserializeTypeConversionPairs

        /// <summary>
        /// Deserializes the indexes xml.
        /// </summary>
        private void DeserializeTypeConversionPairs(XmlNode node, TypeConverter typeconverter)
        {
            XmlNodeList nodes = node.SelectNodes("./typesconversion/typeconversion");
            foreach (XmlNode n in nodes)
            {
                TypeConversionPair typeconversion = new TypeConversionPair();
                typeconversion.Language = GetNodeInnerText(n, "./language", null);
                typeconversion.BaseType = GetNodeInnerText(n, "./basetype", null);
                typeconversion.NameCondition = GetNodeInnerText(n, "./namecondition", null);
                typeconversion.TargetType = GetNodeInnerText(n, "./targettype", null);
                typeconversion.Initializer = GetNodeInnerText(n, "./initializer", null);
                typeconversion.IgnoreCase = bool.Parse(GetNodeInnerText(n,"./ignorecase","False"));
                typeconverter.AppendPair(typeconversion);

            }
        }

        #endregion

        #region DeserializeIndexes

        /// <summary>
		/// Deserializes the indexes xml.
		/// </summary>
		private void DeserializeIndexes(XmlNode tableNode, Table table, Project p)
		{
			XmlNodeList nodes = tableNode.SelectNodes("./indexes/index");
			foreach (XmlNode n in nodes)
			{
				DeserializeIndex(n, table, p);
			}
		}

		/// <summary>
		/// Deserializes the index xml.
		/// </summary>
		private void DeserializeIndex(XmlNode n, Table table, Project p)
		{
			Index index = new Index(GetNodeInnerText(n,"./name",null));
            index.Generate = bool.Parse(GetNodeInnerText(n, "./generate", "False"));
            index.PrimaryKey = bool.Parse(GetNodeInnerText(n, "./primaryKey", "False"));

			// Deserializes the fields of the index.
			DeserializeIndexFields(n, index, p);

			table.Indexes.Add(index);
		}

		/// <summary>
		/// Deserializes the index fields.
		/// </summary>
		private void DeserializeIndexFields(XmlNode n, Index index, Project p)
		{
			XmlNodeList nodes = n.SelectNodes("./fields/" + FIELD_REFERENCE);

			foreach (XmlNode node in nodes)
			{
				index.Fields.Add(DereferenceTableField(node, p));
			}
		}

		#endregion

		#region DeserializeFields

		/// <summary>
		/// Deserializes the fields xml.
		/// </summary>
		private void DeserializeFields(XmlNode tableNode, Table table)
		{
			XmlNodeList nodes = tableNode.SelectNodes("./fields/field");
			foreach (XmlNode n in nodes)
			{
				DeserializeField(n, table);
			}
		}

		/// <summary>
		/// Deserializes a single fields xml.
		/// </summary>
		private void DeserializeField(XmlNode n, Table table)
		{
			Field field = new Field();
			table.Fields.Add(field);

            field.AllowDBNull = bool.Parse(GetNodeInnerText(n, "./allowDBNull", "False"));
            field.AutoNumber = bool.Parse(GetNodeInnerText(n, "./autoNumber", "False"));
            field.FieldName = GetNodeInnerText(n, "./fieldName", null);
            field.Description = GetNodeInnerText(n, "./description", null);
            field.Generate = bool.Parse(GetNodeInnerText(n, "./generate", "False"));
            field.Identifier = bool.Parse(GetNodeInnerText(n, "./identifier", "False"));
            field.Length = int.Parse(GetNodeInnerText(n, "./length", "0"));
            field.ObjectFieldName = GetNodeInnerText(n, "./objectFieldName", "False");
            field.PersistBehaviour = (PersistBehaviours)Enum.Parse(typeof(PersistBehaviours), GetNodeInnerText(n, "./persistBehaviour", null));
            field.PropertyName = GetNodeInnerText(n, "./propertyName", null);
			field.Type = Type.GetType(GetNodeInnerText(n,"./type",null));
		}

		#endregion

		#region DeserializeStorage

		/// <summary>
		/// Deserializes the storage xml.
		/// </summary>
		private void DeserializeStorage(XmlNode root, Project p)
		{
			if (root.SelectSingleNode("./storage/alias") != null)
			{
				p.Storage = new Storage(root.SelectSingleNode("./storage/alias").InnerText,
					root.SelectSingleNode("./storage/connectionString").InnerText);
			}
		}

		#endregion

		#region DeserializeGeneralProject

		/// <summary>
		/// Deserializes the general project xml.
		/// </summary>
		private void DeserializeGeneralProject(XmlNode root, Project p)
		{
			string n = root.SelectSingleNode("./general/namespace").InnerText;
			if (n != null && n.Length > 0)
				p.Namespace = n;

            p.OutputDirectoy = GetNodeInnerText(root, "./general/outputDirectory", null);
            p.RenameItemsToSingular = bool.Parse(GetNodeInnerText(root, "./general/renameItemsToSingular", null));
            p.RemoveFirstChars = int.Parse(GetNodeInnerText(root, "./general/removefirstchars","0"));

            string lastSchema = GetNodeInnerText(root, "./general/lastRetrievedStorageSchema", null);
			if (lastSchema != null && lastSchema.Length > 0)
			{
				XmlDocument doc = new XmlDocument();
				doc.LoadXml(lastSchema);
				// Get the root node of the document.
				XmlNode docRoot = doc.SelectSingleNode("//schema");

				Project project = new Project();

				// Deserializes the tables and relations.
				DeserializeTables(docRoot, project);
				DeserializeRelations(docRoot, project);

				p.LastRetrievedStorageSchema = new StorageSchema(project.Tables, project.Relations);
			}
		}

		#endregion

		#region DeserializeAutorInfo

		/// <summary>
		///	Deserializes the autorinfo xml.
		/// </summary>
		private void DeserializeAutorInfo(XmlNode root, Project p)
		{
			p.AutorInfo.Copyright = root.SelectSingleNode("./autorInfo/copyright").InnerText;
			p.AutorInfo.Name = root.SelectSingleNode("./autorInfo/name").InnerText;
		}

		#endregion

		#region Utility methods

		/// <summary>
		/// Dereferences the node to a table field.
		/// </summary>
		/// <param name="refNode">The node that is dereferenced.</param>
		/// <param name="project">The project that contains the nodes to dereference.</param>
		private Field DereferenceTableField(XmlNode refNode, Project project)
		{
			string t = refNode.Attributes["table"].Value;
			string f = refNode.Attributes["field"].Value;

            Field field = project.GetFieldByName(t, f);

            if(field!=null)
                return field;

			// Throw an exception if not possible to dereference.
			throw new ArgumentException(string.Format("Could not dereference table field: {0}.{1}", t, f));
		}

		/// <summary>
		/// Dereferences the node to a table.
		/// </summary>
		/// <param name="refNode">The node that is dereferenced.</param>
		/// <param name="project">The project that contains the nodes to dereference.</param>
		private Table DereferenceTable(XmlNode refNode, Project project)
		{
			string t = refNode.Attributes["table"].Value;

            Table table = project.GetTableByName(t);

            if (table != null)
                return table;

			// Throw an exception if not possible to dereference.
			throw new ArgumentException(string.Format("Could not dereference table field: {0}", t));
		}

		#endregion

		#endregion
	}
}
