#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.Text;
using Chili.Opf3.Wizard.Shared;
using Chili.Opf3.Wizard.Shared.Storages;

#endregion

namespace Chili.Opf3.Wizard
{
	/// <summary>
	/// This class analyzes the differences between two schemas and adds them to the project.
	/// </summary>
	internal sealed class SchemaDifferencer
	{
		/// <summary>
		/// Synchronizes the project with the differences between the old and the new schema.
		/// </summary>
		/// <param name="newSchema">The new schema.</param>
		/// <param name="oldSchema">The old schema.</param>
		/// <param name="p">The project where the differences are set.</param>
		public void Synchronize(StorageSchema newSchema, StorageSchema oldSchema, Project p)
		{
			if (newSchema == null)
				throw new ArgumentNullException("newSchema", "NewSchema is null.");
			if (p == null)
				throw new ArgumentNullException("p", "Project is null.");

			// Create a new schema for the old one if none is given.
			if (oldSchema == null)
				oldSchema = new StorageSchema();

			SynchronizeTables(newSchema, oldSchema, p);
			SynchronizeRelations(newSchema, oldSchema, p);
			
			// Sets the current schema as last retrieved storage schema.
			p.LastRetrievedStorageSchema = newSchema;
		}

		#region SynchronizeRelations

		/// <summary>
		/// Synchronizes the relations of the schema.
		/// </summary>
		private void SynchronizeRelations(StorageSchema newSchema, StorageSchema oldSchema, Project p)
		{
			// Loop over the new schema's relations to see if there are changes.
			foreach (Relation relation in newSchema.Relations)
			{
				// Get the relation from the old schema.
				Relation oldRelation = GetRelationFromList(relation, oldSchema.Relations);

				if (oldRelation != null)
				{
					// Get the relation from the project.
					Relation projectRelation = GetRelationFromList(relation, p.Relations);
					if (projectRelation != null)
					{
						SynchronizeRelationItems(relation, oldRelation, projectRelation);
						// Synchronizes the references of the relation.
						SynchronizeRelationReferences(projectRelation, p.Tables);
					}
				}
				else
				{
					// Check if a custom relation has been added by the user. If the current retrieval
					// returns a new relation that matches the relation that has been added by the user,
					// the one of the user is removed and replaced with the one returned by the schema.
					Relation r = GetRelationFromList(relation, p.Relations);
					if (r != null)
						RemoveRelationFromList(r, p.Relations);

					// Synchronize the references in the relation.
					SynchronizeRelationReferences(relation, p.Tables);
					// Add the relation to the list of relations.
					p.Relations.Add(relation);
				}
			}

			// Loop over all relations in the old schema. 
			foreach (Relation relation in oldSchema.Relations)
			{
				// ... If they are no longer in the new schema, we have to remove them from the list.
				if (GetRelationFromList(relation, newSchema.Relations) == null)
					RemoveRelationFromList(relation, p.Relations);
			}
		}

		/// <summary>
		/// Synchronizes the relation.
		/// </summary>
		private void SynchronizeRelationItems(Relation relation, Relation oldRelation, Relation projectRelation)
		{
			// Go over all relation items.
			foreach (RelationItem item in relation.Relations)
			{
				// Check if the relation item was not in the old schema. If not add it. If it was there, just leave it.
				RelationItem oldItem = GetRelationItemFromList(item, oldRelation.Relations);
				if (oldItem == null)
				{
					// Remove the custom relation item from the list, if the schema returns a new relations
					// item that matches the item that has been added by the user.
					RelationItem r = GetRelationItemFromList(item, projectRelation.Relations);
					if (r != null)
						RemoveRelationItemFromList(r, projectRelation.Relations);

					projectRelation.Relations.Add(item);
				}
			}
			foreach (RelationItem item in oldRelation.Relations)
			{
				if (GetRelationItemFromList(item, relation.Relations) == null)
					RemoveRelationItemFromList(item, projectRelation.Relations);
			}

			// Return if no weak table is present.
			if (relation.WeakTable == null)
				return;

			// Go over all weak relation items.
			foreach (RelationItem item in relation.WeakRelations)
			{
				// Check if the relation item was not in the old schema. If not add it. If it was there, just leave it.
				RelationItem oldItem = GetRelationItemFromList(item, oldRelation.WeakRelations);
				if (oldItem == null)
				{
					// Remove the custom relation item from the list, if the schema returns a new relations
					// item that matches the item that has been added by the user.
					RelationItem r = GetRelationItemFromList(item, projectRelation.WeakRelations);
					if (r != null)
						RemoveRelationItemFromList(r, projectRelation.WeakRelations);
                    
                    projectRelation.WeakRelations.Add(item);
				}
			}
			foreach (RelationItem item in oldRelation.WeakRelations)
			{
				if (GetRelationItemFromList(item, relation.WeakRelations) == null)
					RemoveRelationItemFromList(item, projectRelation.WeakRelations);
			}
		}

		/// <summary>
		/// Synchronizes the relation references.
		/// </summary>
		private void SynchronizeRelationReferences(Relation relation, List<Table> list)
		{
			relation.ParentTable = GetTableFromList(relation.ParentTable, list);
			relation.ChildTable = GetTableFromList(relation.ChildTable, list);
			if (relation.WeakTable != null)
				relation.WeakTable = GetTableFromList(relation.WeakTable, list);

			foreach (RelationItem r in relation.Relations)
			{
				// Check which list of fields has to be given to the child field.
				List<Field> fields = (relation.WeakTable != null ? relation.WeakTable.Fields : relation.ChildTable.Fields);

				r.ParentField = GetTableFieldFromList(r.ParentField, relation.ParentTable.Fields);
				r.ChildField = GetTableFieldFromList(r.ChildField, fields);
			}

			foreach (RelationItem r in relation.WeakRelations)
			{
				// Check which list of fields has to be given to the parent field.
				List<Field> fields = (relation.WeakTable != null ? relation.WeakTable.Fields : relation.ChildTable.Fields);

				r.ParentField = GetTableFieldFromList(r.ChildField, fields);
				r.ChildField = GetTableFieldFromList(r.ParentField, relation.ChildTable.Fields);
			}
		}

		#endregion

		#region SynchronizeTables

		/// <summary>
		/// Synchronizes the tables of the schemas.
		/// </summary>
		private void SynchronizeTables(StorageSchema newSchema, StorageSchema oldSchema, Project p)
		{
			// Loop over the new schema's tables to see if there are changes.
			foreach (Table table in newSchema.Tables)
			{
				// Check if the old schema contained the table.
				Table oldTable = GetTableFromList(table, oldSchema.Tables);

				if (oldTable != null)
				{
					// Get the table from the list.
					Table projectTable = GetTableFromList(table, p.Tables);
					if (projectTable != null)
					{
						// Update the properties of the table.
						if (table.IsView != oldTable.IsView)
							projectTable.IsView = table.IsView;

						// Synchronize the fields.
						SynchronizeFields(table, oldTable, projectTable);
						// Synchronize the indexes.
						SynchronizeIndexes(table, oldTable, projectTable);
					}
				}
				else
				{
					// Remove the table if already in the list... This could be a custom table. We should it
					// remove therefore and add the one that is returned by the schema.
					Table t = GetTableFromList(table, p.Tables);
					if (t != null)
						RemoveTableFromList(t, p.Tables);
                    
                    // Add the table to the project. It has not been in the old schema.
                    p.Tables.Add(table);
				}
			}

			// Loop over all tables in the old schema. 
			foreach (Table table in oldSchema.Tables)
			{
				// ... If they are no longer in the new schema, we have to remove them from the list.
				if (GetTableFromList(table, newSchema.Tables) == null)
					RemoveTableFromList(table, p.Tables);
			}
		}

		#endregion

		#region SynchronizeIndexes

		/// <summary>
		/// Synchronizes the indexes of the table.
		/// </summary>
		private void SynchronizeIndexes(Table newTable, Table oldTable, Table projectTable)
		{
			foreach (Index index in newTable.Indexes)
			{
				// Check if the old table contains the index.
				Index oldIndex = GetTableIndexFromList(index, oldTable.Indexes);

				if (oldIndex != null)
				{
					// Get the index from the project table.
					Index tableIndex = GetTableIndexFromList(index, projectTable.Indexes);
					if (tableIndex != null)
					{
						// Update the properties of the index.
						if (index.Name != oldIndex.Name)
							tableIndex.Name = index.Name;
						if (index.PrimaryKey != oldIndex.PrimaryKey)
							tableIndex.PrimaryKey = index.PrimaryKey;

						SynchronizeIndexFields(index, oldIndex, tableIndex, projectTable.Fields);
					}
				}
				else
				{
					// Remove the old index if one has been added by hand.
					Index i = GetTableIndexFromList(index, projectTable.Indexes);
					if (i != null)
						RemoveIndexFromList(i, projectTable.Indexes);

					SynchronizeIndexFieldReferences(index, projectTable.Fields);
					// Add the index to the table.
					projectTable.Indexes.Add(index);
				}
			}

			// Loop over all index in the old table.
			foreach (Index index in oldTable.Indexes)
			{
				// ... If no longer present remove it.
				if (GetTableIndexFromList(index, newTable.Indexes) == null)
					RemoveIndexFromList(index, projectTable.Indexes);
			}
		}

		/// <summary>
		/// Synchronizes the fields of the index with the one in the list.
		/// </summary>
		private void SynchronizeIndexFields(Index newIndex, Index oldIndex, Index tableIndex, List<Field> list)
		{
			foreach (Field field in newIndex.Fields)
			{
				// Check if the field was not in the old schema. If not add it. If it was there, just leave it.
				Field oldField = GetTableFieldFromList(field, oldIndex.Fields);
				if (oldField == null)
				{
					tableIndex.Fields.Add(field);
				}
			}

			foreach (Field field in oldIndex.Fields)
			{
				if (GetTableFieldFromList(field, newIndex.Fields) == null)
					RemoveFieldFromList(field, tableIndex.Fields);
			}

			// Synchronize the fields of the index.
			SynchronizeIndexFieldReferences(tableIndex, list);
		}

		/// <summary>
		/// Synchronizes the fields references of the index with the one in the list.
		/// </summary>
		private void SynchronizeIndexFieldReferences(Index index, List<Field> list)
		{
			for(int i = index.Fields.Count -1; i >= 0; i--)
			{
				// Loop over all fields in the index and adjust them to the fields from the project.
				index.Fields[i] = GetTableFieldFromList(index.Fields[i], list);
			}
		}

		#endregion

		#region SynchronizeFields

		/// <summary>
		/// Synchronizes the fields of the table.
		/// </summary>
		private void SynchronizeFields(Table newtable, Table oldTable, Table projectTable)
		{
			foreach (Field field in newtable.Fields)
			{
				// Check if the old table contains the field.
				Field oldField = GetTableFieldFromList(field, oldTable.Fields);

				if (oldField != null)
				{
					// Get the field from the project table.
					Field tableField = GetTableFieldFromList(field, projectTable.Fields);
					if (tableField != null)
					{
						// Update the properties of the field.
						if (field.AllowDBNull != oldField.AllowDBNull)
							tableField.AllowDBNull = field.AllowDBNull;
						if (field.AutoNumber != oldField.AutoNumber)
							tableField.AutoNumber = field.AutoNumber;
						if (field.Identifier != oldField.Identifier)
							tableField.Identifier = field.Identifier;
						if (field.Length != oldField.Length)
							tableField.Length = field.Length;
						if (field.Type != oldField.Type)
							tableField.Type = field.Type;
					}
				}
				else
				{
					// Remove a custom field that has been added to the list, if the schema returns
					// the same field added since the last schema retrieval.
					Field f = GetTableFieldFromList(field, projectTable.Fields);
					if (f != null)
						RemoveFieldFromList(f, projectTable.Fields);

					// Add the field to the table. It has not been in the old table or project table.
					projectTable.Fields.Add(field);
				}
			}

			// Loop over all fields in the old table.
			foreach (Field field in oldTable.Fields)
			{
				// ... If no longer present remove it.
				if (GetTableFieldFromList(field, newtable.Fields) == null)
					RemoveFieldFromList(field, projectTable.Fields);
			}
		}

		#endregion

		#region Utility methods

		/// <summary>
		/// Removes the index from the given list.
		/// </summary>
		private bool RemoveIndexFromList(Index index, List<Index> list)
		{
			foreach (Index i in list)
			{
				if (i.Name == index.Name)
				{
					list.Remove(i);
					return true;
				}
			}
			return false;
		}

		/// <summary>
		/// Removes the field from the given list.
		/// </summary>
		private bool RemoveFieldFromList(Field field, List<Field> list)
		{
			foreach (Field f in list)
			{
				if (f.FieldName == field.FieldName)
				{
					list.Remove(f);
					return true;
				}
			}
			return false;
		}

		/// <summary>
		/// Removes the table from the given list.
		/// </summary>
		private bool RemoveTableFromList(Table table, List<Table> list)
		{
			foreach (Table t in list)
			{
				if (t.TableName == table.TableName)
				{
					list.Remove(t);
					return true;
				}
			}
			return false;
		}

		/// <summary>
		/// Remove relation from the given list.
		/// </summary>
		private bool RemoveRelationFromList(Relation relation, List<Relation> list)
		{
			foreach (Relation r in list)
			{
				if (r.ChildTable.PersistentName == relation.ChildTable.PersistentName &&
					r.ParentTable.PersistentName == relation.ParentTable.PersistentName)
				{
					if (relation.WeakTable != null)
					{
						if (r.WeakTable != null && relation.WeakTable.PersistentName == r.WeakTable.PersistentName)
						{
							list.Remove(r);
							return true;
						}
					}
					else
					{
						list.Remove(r);
						return true;
					}
				}
			}
			return false;
		}

		/// <summary>
		/// Remove relation items from the given list.
		/// </summary>
		private bool RemoveRelationItemFromList(RelationItem item, List<RelationItem> list)
		{
			foreach (RelationItem i in list)
			{
				if (i.ChildField.FieldName == item.ChildField.FieldName &&
					i.ParentField.FieldName == item.ParentField.FieldName)
				{
					list.Remove(i);
					return true;
				}
			}
			return false;
		}

		/// <summary>
		/// Gets the relation item from the given list. Null is returned if the list does not contain the given item.
		/// </summary>
		private RelationItem GetRelationItemFromList(RelationItem item, List<RelationItem> list)
		{
			foreach (RelationItem i in list)
			{
				if (i.ChildField.FieldName == item.ChildField.FieldName &&
					i.ParentField.FieldName == item.ParentField.FieldName)
				{
					return i;
				}
			}
			return null;
		}

		/// <summary>
		/// Gets the relation from the given list. Null is returned if the list does not contain the given relation.
		/// </summary>
		private Relation GetRelationFromList(Relation relation, List<Relation> list)
		{
			foreach (Relation r in list)
			{
				// Check if the tables of the references match.
				if (r.ChildTable.PersistentName == relation.ChildTable.PersistentName &&
					r.ParentTable.PersistentName == relation.ParentTable.PersistentName)
				{
                    if (!EqualRelationItems(r.Relations, relation.Relations))
                        continue;

					// Check if we have a weak table.
					if (relation.WeakTable != null)
					{
						// If the relation in the list has none, just continue.
						if (r.WeakTable == null)
							continue;
						// ... if the relation has one, check the persistent name of the weak items.
						if (relation.WeakTable.PersistentName == r.WeakTable.PersistentName)
							return r;
					}
					else
						return r;
				}
			}
			return null;
		}

        /// <summary>
        /// Returns true if both lists of relation items equal.
        /// </summary>
        /// <param name="items1"></param>
        /// <param name="items2"></param>
        private bool EqualRelationItems(List<RelationItem> items1, List<RelationItem> items2)
        {
            if (items1.Count != items2.Count)
                return false;

            for (int i = 0; i < items1.Count; i++)
            {
                if (items1[i].ChildField.FieldName != items2[i].ChildField.FieldName ||
                items1[i].ParentField.FieldName != items2[i].ParentField.FieldName)
                    return false;
            }

            return true;
        }
 

		/// <summary>
		/// Gets the table's index from the given list. Null is returned if the list does not contain the given index.
		/// </summary>
		private Index GetTableIndexFromList(Index index, List<Index> list)
		{
			foreach (Index i in list)
			{
				if (i.Name == index.Name)
					return i;
			}
			return null;
		}

		/// <summary>
		/// Gets the table's field from the given list. Null is returned if the list does not contain the given field.
		/// </summary>
		/// <param name="field">The field that is searched in the list.</param>
		/// <param name="list">The list holding the fields.</param>
		private Field GetTableFieldFromList(Field field, List<Field> list)
		{
			foreach (Field f in list)
			{
				if (f.FieldName == field.FieldName)
					return f;
			}
			return null;
		}

		/// <summary>
		/// Gets the given table from the list. Null is returned if the list does not contain the given table.
		/// </summary>
		/// <param name="table">The table that is searched in the list.</param>
		/// <param name="list">The list holding tables.</param>
		private Table GetTableFromList(Table table, List<Table> list)
		{
			foreach (Table t in list)
			{
				if (t.TableName == table.TableName)
					return t;
			}
			return null;
		}

		#endregion
	}
}
