#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.Text;
using System.Data.OleDb;
using Chili.Opf3.Wizard.Shared.Storages;
using System.Data;
using Chili.Opf3.Wizard.Shared;

#endregion

namespace Chili.Opf3.Wizard
{
	/// <summary>
	/// Retrieves the schema information from the storage.
	/// </summary>
	public sealed class StorageSchemaReader
	{
		private OleDbConnection _connection;
        private Project _project;

		/// <summary>
		/// Creates a new instance of the <see cref="StorageSchemaReader">StorageSchemaReader Class</see>.
		/// </summary>
		/// <param name="connection">The connection used to get the storage information.</param>
        /// <param name="project">The project.</param>
		public StorageSchemaReader(OleDbConnection connection,Project project)
		{
			if (connection == null)
				throw new ArgumentNullException("connection", "Connection is null.");
            if (project == null)
                throw new ArgumentNullException("project", "Project is null.");

			_connection = connection;
            _project = project;
		}

		/// <summary>
		/// Event fired after a table has been fetched.
		/// </summary>
		public event EventHandler<FetchedTableEventArgs> FetchedTable;

		/// <summary>
		/// Event fired after a relation has been fecthed.
		/// </summary>
		public event EventHandler<FetchedRelationEventArgs> FetchedRelation;

		/// <summary>
		/// Returns the schema of the storage.
		/// </summary>
		public StorageSchema GetSchema()
		{
			StorageSchema schema = new StorageSchema();
			// Get first the tables then the relations. The tables are required for the relations.
            GetTablesAndColumns(schema.Tables);
            GetRelations(schema.Tables, schema.Relations);

			return schema;
		}

		#region GetTablesAndColumns

		/// <summary>
		/// Returns the tables and the columns from the storage.
		/// </summary>
		/// <param name="tables">The list that is filled with the tables.</param>
        private void GetTablesAndColumns(List<Table> tables)
		{
			// Opens the connection if not open yet.
			if (_connection.State == ConnectionState.Closed)
				_connection.Open();

			try
			{
				// Get the tables of the database.
				DataTable schema = _connection.GetOleDbSchemaTable(OleDbSchemaGuid.Tables, new object[] { null, null, null, "TABLE" });
				foreach (DataRow row in schema.Rows)
				{
					try
					{
						// the name of the table.
						string tableName = row["TABLE_NAME"].ToString();

						// Get a table.
						Table table = GetTable(tableName);
						if (table != null)
						{
							tables.Add(table);
							// Fire the event.
							OnFetchedTable(new FetchedTableEventArgs(table));
						}
					}
					catch
					{
						// Ignore all exceptions happening here.
					}
				}

				// Check if we get the views from the list of tables, but with type "View".
				schema = _connection.GetOleDbSchemaTable(OleDbSchemaGuid.Tables, new object[] { null, null, null, "VIEW" });
				if (schema.Rows.Count > 0)
				{
					foreach (DataRow row in schema.Rows)
					{
						try
						{
							GetView(tables, row, false);
						}
						catch
						{
							// Ignore all exceptions happening here.
						}
					}
				}
				else
				{
					// Use first table data to create view filter.
					object[] restrictions = null;
					if (schema.Rows.Count > 0)
						restrictions = new object[] { schema.Rows[0]["TABLE_CATALOG"], schema.Rows[0]["TABLE_SCHEMA"], null };

					// Get the view of the database.
					schema = _connection.GetOleDbSchemaTable(OleDbSchemaGuid.Views, restrictions);
					foreach (DataRow row in schema.Rows)
					{
						try
						{
							GetView(tables, row, true);
						}
						catch
						{
							// Ignore all exceptions happening here.
						}
					}
				}
			}
			catch
			{
				// Ignore all exceptions happening here.
			}
		}

		/// <summary>
		/// Gets a view from the given data row.
		/// </summary>
		/// <param name="tables">The list of tables where the view is added.</param>
		/// <param name="row">The current row, where the data of the view is fetched.</param>
		/// <param name="checkIfIsUpdatable">True to check if the view is updatable.</param>
		private void GetView(List<Table> tables, DataRow row, bool checkIfIsUpdatable)
		{
			// the name of the view.
			string tableName = row["TABLE_NAME"].ToString();

			// Get a table.
			Table table = GetTable(tableName);
			if (table != null)
			{
				table.IsView = true;
				table.Rights = PersistentRights.Load;

				if (checkIfIsUpdatable)
				{
					bool isUpdatable = false;
					Boolean.TryParse(row["IS_UPDATABLE"].ToString(), out isUpdatable);
					table.Rights = isUpdatable ? PersistentRights.All : PersistentRights.Load;
				}

				tables.Add(table);
				// Fire the event.
				OnFetchedTable(new FetchedTableEventArgs(table));
			}
		}

		/// <summary>
		/// Returns the schema information for one table.
		/// </summary>
		/// <param name="tableName">The table that's schema information is returned.</param>
		private Table GetTable(string tableName)
		{
			// Create a new table object.
			Table table = new Table();
			table.TableName = tableName;
			table.PersistentName = NameUtility.GetPersistentName(tableName, _project.RenameItemsToSingular);

            // Create the command used to return the schemainformation for the table.
            OleDbCommandBuilder commandBuilder = new OleDbCommandBuilder();
            string query = string.Format("select * from {0}", commandBuilder.QuoteIdentifier(tableName, _connection));
            OleDbCommand command = new OleDbCommand(query);
            command.Connection = _connection;

			try
			{
				// Get the schema information.
				using (IDataReader dataReader = command.ExecuteReader(CommandBehavior.KeyInfo | CommandBehavior.SchemaOnly))
				{
					DataTable schema = dataReader.GetSchemaTable();

					// Create the field objects.
					foreach (DataRow row in schema.Rows)
					{
						Field field = GetField(row);
						if (field != null)
							table.Fields.Add(field);
					}

					// Create the index objects.
					System.Data.DataTable indexes = _connection.GetSchema("Indexes");
					foreach (DataRow row in indexes.Select(String.Format("TABLE_NAME='{0}'", tableName)))
					{
						AddIndex(table.Indexes, table.Fields, row);
					}
				}
			}
			catch
			{
				return null;
			}

			return table;
		}

        /// <summary>
        /// Returns the schema information for one index.
        /// </summary>
        /// <param name="indexes">The table's indexes collection.</param>
        /// <param name="fields">The table's fields collection.</param>
        /// <param name="row">The current row that contains the schema information for the current row.</param>
        private void AddIndex(List<Index> indexes, List<Field> fields, DataRow row)
        {
            Index index = null;
            string indexname = (string)row["INDEX_NAME"];

            // Search if the index exist (composite index).
			foreach (Index current in indexes)
			{
				if (current.Name == indexname)
				{
					index = current;
					break;
				}
			}

            // Create a new index object if not exist and add to the collection.
            if (index == null)
            {
                index = new Index(indexname);
                indexes.Add(index);
            }

            bool isPrimaryKey = false;
            Boolean.TryParse(row["PRIMARY_KEY"].ToString(), out isPrimaryKey);

            index.PrimaryKey = isPrimaryKey;
            string fieldName = row["COLUMN_NAME"].ToString();

            // Search for field in the fields collection.
            foreach(Field field in  fields)
			{
				if (field.FieldName == fieldName)
				{
					index.Fields.Add(field);
					break;
				}
			}
        }

        /// <summary>
        /// Returns the schema information for one field.
        /// </summary>
        /// <param name="row">The current row that contains the schema information for the current row.</param>
        private Field GetField(DataRow row)
        {
            string fieldName = row["ColumnName"].ToString();

            // Create a new field object and fill it with data.
            Field field = new Field();
            field.FieldName = fieldName;
            field.ObjectFieldName = NameUtility.GetObjectFieldName(fieldName.Substring(_project.RemoveFirstChars));
            field.PropertyName = NameUtility.GetPropertyName(fieldName.Substring(_project.RemoveFirstChars));
            field.AllowDBNull = Convert.ToBoolean(row["AllowDBNull"]);
            field.AutoNumber = Convert.ToBoolean(row["IsAutoIncrement"]);
            field.Identifier = Convert.ToBoolean(row["IsKey"]);
			try
			{
				field.Type = (Type)row["DataType"];
			}
			catch (Exception)
			{
				throw new InvalidCastException("Can't cast: " + row["DataType"] + " in column '" + fieldName + "' to a .NET type.");
			}

            field.Length = Convert.ToInt32(row["ColumnSize"]);

			// Get the base table name.
			object baseTableName = row["BaseTableName"];

			if (baseTableName != DBNull.Value)
			{
				// retrieve additional data
				string[] restrictions = new string[] { null, null, (string)baseTableName, fieldName };
				System.Data.DataTable schema = _connection.GetSchema("Columns", restrictions);
				if (schema.Rows.Count > 0)
				{
					field.AllowDBNull = Convert.ToBoolean(schema.Rows[0]["IS_NULLABLE"]);
					field.Description = schema.Rows[0]["DESCRIPTION"].ToString();
				}
			}

            return field;
        }

		#endregion

		#region GetRelations

		/// <summary>
		/// Returns the relations for the database.
		/// </summary>
		/// <param name="tables">List filled with the tables from the storage.</param>
		/// <param name="relations">List that is filled with the tables from the storage.</param>
		private void GetRelations(List<Table> tables, List<Relation> relations)
		{
			// Opens the connection if not open yet.
			if (_connection.State == ConnectionState.Closed)
				_connection.Open();

			// Returns the relations of the database.
			DataTable schema = _connection.GetOleDbSchemaTable(OleDbSchemaGuid.Foreign_Keys, null);

			Relation relation = null;
			// Loop over all relations.
			for (int i = 0; i < schema.Rows.Count; i++)
			{
                try
				{
				    string parentTableName = schema.Rows[i]["PK_TABLE_NAME"].ToString();
				    string childTableName = schema.Rows[i]["FK_TABLE_NAME"].ToString();
    				
				    Table parentTable = GetRelationTable(tables, parentTableName);
				    Table childTable = GetRelationTable(tables, childTableName);

					// skip if the parent or child table are null.
                    if (parentTable == null || childTable == null)
						continue;
                    
				    Field parentField = GetRelationField(parentTable, schema.Rows[i]["PK_COLUMN_NAME"].ToString());
				    Field childField = GetRelationField(childTable, schema.Rows[i]["FK_COLUMN_NAME"].ToString());

					// skip if the length of the parent or child field is zero.
					if (parentField.Length == 0 || childField.Length == 0)
						continue;

//					long ordinal = Convert.ToInt64(schema.Rows[i]["ORDINAL"]);

					// check whether we have to create a new relation. The two cases are if
					// no relation has been created yet or one of the table names has altered.
					// we assume that the foreign keys of a relation with multiple foreign keys
					// follow directly after each other.
                    if (relation == null || 
						(relation != null && (relation.ParentTable.TableName != parentTable.TableName || 
						relation.ChildTable.TableName != childTable.TableName)))
				    {
					    // Create a new relation.
					    relation = new Relation();
                        relation.Name = schema.Rows[i]["FK_NAME"].ToString();
					    // Switch because those properties are the one for the Object(Set)Holder.
					    relation.ParentProperty = NameUtility.GetPersistentName(childTableName, false);
					    relation.ChildProperty = NameUtility.GetPersistentName(parentTableName, _project.RenameItemsToSingular);
					    relation.ParentTable = parentTable;
					    relation.ChildTable = childTable;
                        // Add the new relation.
                        relations.Add(relation);
                        // Fire the event.
                        OnFetchedRelation(new FetchedRelationEventArgs(relation));
				    }

				    // Add the relation item.
				    RelationItem relationItem = new RelationItem();
				    relationItem.ParentField = parentField;
				    relationItem.ChildField = childField;
				    relation.Relations.Add(relationItem);
                }
                catch
                {
                    // Ignore all exceptions happening here.
                }
			}
		}

		#region GetTable, GetField

		/// <summary>
		/// Returns the table associated with the given table name or null.
		/// </summary>
		/// <param name="tables">The list of tables. This list is searched.</param>
		/// <param name="tableName">The name of the table used to get the table object.</param>
		private Table GetRelationTable(List<Table> tables, string tableName)
		{
			// Loop over all tables and return the right one.
			foreach (Table table in tables)
			{
				if (table.TableName == tableName)
					return table;
			}
			// No table has been found. Null is returned.
			return null;
		}

		/// <summary>
		/// Returns the field associated with the given table name.
		/// </summary>
		/// <param name="table">The table used to get the field object.</param>
		/// <param name="fieldName">The name of the field used to get the field object.</param>
		private Field GetRelationField(Table table, string fieldName)
		{
			// If no table has been found return null.
			if (table == null)
				return null;

			// Loop over all fields and return the appropriate.
			foreach (Field field in table.Fields)
			{
				if (field.FieldName == fieldName)
					return field;
			}
			// No field has been found. Null is returned.
			return null;
		}

		#endregion

		#endregion

		/// <summary>
		/// Fires the FetchedTable event.
		/// </summary>
		/// <param name="e"></param>
		private void OnFetchedTable(FetchedTableEventArgs e)
		{
			if (FetchedTable != null)
				FetchedTable(this, e);
		}

		/// <summary>
		/// Fires the FetchedRelation event.
		/// </summary>
		/// <param name="e"></param>
		private void OnFetchedRelation(FetchedRelationEventArgs e)
		{
			if (FetchedRelation != null)
				FetchedRelation(this, e);
		}
	}
}
