#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Chili.Opf3.Wizard.Shared.Storages;
using Chili.Opf3.Wizard.Shared.UI;
using Chili.Opf3.Wizard.Properties;

#endregion

namespace Chili.Opf3.Wizard.UI.Forms
{
	/// <summary>
	/// This dialog allows to edit one relation item.
	/// </summary>
	public partial class IndexItemDialog : BaseDialog
	{
        private Index _index;
        private Table _table;

		/// <summary>
		/// Creates a new instance of the <see cref="RelationsItemDialog">RelationsItemDialog Class</see>.
		/// </summary>
		/// <param name="mode">The mode of the dialog.</param>
		/// <param name="relationItem">The item that is edited.</param>
		/// <param name="childTable">The child table.</param>
		/// <param name="parentTable">The parent table.</param>
        public IndexItemDialog(DialogModes mode, Index index, Table table)
		{
			InitializeComponent();

            if (index == null)
                throw new ArgumentNullException("index", "Index is null.");
            if (table == null)
                throw new ArgumentNullException("table", "Table is null.");

            _index = index;
            _table = table;

            PopulateAvailableFieldsList(_table, availablefieldsListBox);

			// Set the labels with the persistent names.
            indexNameTextBox.Text = _index.Name;

			if (mode == DialogModes.New)
				this.Text = Resources.NewIndexItemDialogTitle;
			else
			{
				this.Text = Resources.EditIndexItemDialogTitle;
				PopulateControls();
			}
		}

		/// <summary>
        /// Populates the given listbox with the fields of the given table.
		/// </summary>
		/// <param name="table">The table that's fields are populated.</param>
		/// <param name="combobox">The combobox.</param>
        private void PopulateAvailableFieldsList(Table table, ListBox listbox)
		{
            // Clear all items in the listbox.
            listbox.Items.Clear();

			foreach (Field field in table.Fields)
			{
                listbox.Items.Add(field);
			}
            SetOperationButtonsStatus();
		}

		/// <summary>
		/// Populates the UI controls.
		/// </summary>
		private void PopulateControls()
		{
            foreach (Field field in _index.Fields)
            {
                AddField(field);
            }
            SetOperationButtonsStatus();
		}

		
		/// <summary>
		/// Invoked each time the form is about to close.
		/// </summary>
		private void IndexItemDialog_FormClosing(object sender, FormClosingEventArgs e)
		{
			if (this.DialogResult != DialogResult.OK)
				return;

			e.Cancel = (!PopulateObject());
		}


        /// <summary>
        /// Add a field to the fields list and remove it from availables.
        /// </summary>
        private void AddField(Field field)
        {
            if (field == null) 
				return;
            
			if (fieldsListBox.Items.Contains(field)) 
				return;
            
			fieldsListBox.Items.Add(field);
            availablefieldsListBox.Items.Remove(field);  
        }

        /// <summary>
        /// Remove a field from the fields list and add it from availables.
        /// </summary>
        private void RemoveField(Field field)
        {
            if (field == null) 
				return;

            if (!fieldsListBox.Items.Contains(field)) 
				return;
            
			fieldsListBox.Items.Remove(field);
            availablefieldsListBox.Items.Add(field); 
        }


		/// <summary>
		/// Populates the object with data.
		/// </summary>
		private bool PopulateObject()
		{
			if (indexNameTextBox.Text.Length == 0)
			{
				MessageBox.Show(Resources.MissingIndexNameDescription, Resources.ProgramName, MessageBoxButtons.OK, MessageBoxIcon.Error);
				indexNameTextBox.Focus();
				return false;
			}

            if (fieldsListBox.Items.Count == 0)
			{
				MessageBox.Show(Resources.MissingIndexFieldsText, Resources.ProgramName, MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}

            _index.Name=indexNameTextBox.Text;

            // fill the fields list
            _index.Fields.Clear();
            foreach (object item in fieldsListBox.Items)
                _index.Fields.Add((Field)item);

			return true;
		}

        /// <summary>
        /// Add all available fields.
        /// </summary>
        private void addallButton_Click(object sender, EventArgs e)
        {
            foreach (Field field in _table.Fields)
            {
                AddField(field);
            }
            SetOperationButtonsStatus();
        }

        /// <summary>
        /// Add the available selected fields.
        /// </summary>
        private void addcurrentButton_Click(object sender, EventArgs e)
        {
            AddField((Field)availablefieldsListBox.SelectedItem);
            SetOperationButtonsStatus();
        }

        /// <summary>
        /// Add the available selected fields.
        /// </summary>
        /// 
        private void removeallButton_Click(object sender, EventArgs e)
        {
            foreach (Field field in _table.Fields)
            {
                RemoveField(field);
            }
            SetOperationButtonsStatus();
        }

        /// <summary>
        /// Add the available selected fields.
        /// </summary>
        /// 
        private void removecurrentButton_Click(object sender, EventArgs e)
        {
            RemoveField((Field)fieldsListBox.SelectedItem);
            SetOperationButtonsStatus();
        }

        /// <summary>
        /// Enable /disable remove/add buttons.
        /// </summary>
        /// 
        private void SetOperationButtonsStatus()
        {
            removecurrentButton.Enabled = (fieldsListBox.SelectedItems.Count > 0);
            removeallButton.Enabled = (fieldsListBox.Items.Count > 0);
            addcurrentButton.Enabled = (availablefieldsListBox.SelectedItems.Count > 0);
            addallButton.Enabled = (availablefieldsListBox.Items.Count > 0);
        }

        /// <summary>
        /// Enable /disable addxxx buttons.
        /// </summary>
        /// 
        private void fieldsListBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            SetOperationButtonsStatus();
        }

        /// <summary>
        /// Enable /disable addxxx buttons.
        /// </summary>
        /// 
        private void availablefieldsListBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            SetOperationButtonsStatus();
        }

        /// <summary>
        /// Add the available selected fields.
        /// </summary>
        /// 
        private void availablefieldsListBox_DoubleClick(object sender, EventArgs e)
        {
            AddField((Field)availablefieldsListBox.SelectedItem);
        }
	}
}