#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Chili.Opf3.Wizard.Shared.Storages;
using Chili.Opf3.Wizard.Shared.UI;
using Chili.Opf3.Wizard.Properties;
using System.Text.RegularExpressions;

#endregion

namespace Chili.Opf3.Wizard.UI.Forms
{
	/// <summary>
	/// This form allows to modify the properties a persistent object.
	/// </summary>
	public partial class PersistentDialog : BaseDialog
	{
		private Table _table;
        private List<Relation> _relations;

		/// <summary>
		/// Creates a new instance of the <see cref="PropertiesForm">PropertiesForm Class</see>.
		/// </summary>
		/// <param name="table">The table representing the persistent object.</param>
		/// <param name="mode">The mode of the dialog.</param>
		public PersistentDialog(DialogModes mode, Table table,List<Relation> relations)
		{
			InitializeComponent();

			if (table == null)
				throw new ArgumentNullException("table", "Table is null.");
			_table = table;

            if (relations == null)
                throw new ArgumentNullException("relations", "Relations is null.");
            _relations = relations;

			if (mode == DialogModes.New)
			{
				// Set the title of the form.
				this.Text = Resources.NewPersistentDialogTitle;
			}
			else
			{
				// Set the title of the form.
				this.Text = string.Format(Resources.EditPersistentDialogTitle, table.PersistentName);
				// Populate the controls with data.
				PopulateControls();
			}
		}

		/// <summary>
		/// Invoked each time the no pool checkbox is clicked.
		/// </summary>
		private void noPoolCheckBox_CheckedChanged(object sender, EventArgs e)
		{
			poolSizeNumericUpDown.Enabled = (!noPoolCheckBox.Checked);
		}

		/// <summary>
		/// Populate the UI controls with data.
		/// </summary>
		private void PopulateControls()
		{
			tableNameTextBox.Text = _table.TableName;
			persistentNameTextBox.Text = _table.PersistentName;
            baseclassTextBox.Text = _table.BaseClass;
			noPoolCheckBox.Checked = _table.NoPool;
			if (!_table.NoPool)
				poolSizeNumericUpDown.Value = (decimal)_table.PoolSize;

            pictureBox1.Image = _table.IsView ? global::Chili.Opf3.Wizard.Properties.Resources.table_view : global::Chili.Opf3.Wizard.Properties.Resources.table; 
			updateCheckBox.Checked = ((_table.Rights & PersistentRights.Update) != 0);
			insertCheckBox.Checked = ((_table.Rights & PersistentRights.Insert) != 0);
			deleteCheckBox.Checked = ((_table.Rights & PersistentRights.Delete) != 0);
			loadCheckBox.Checked = ((_table.Rights & PersistentRights.Load) != 0);

			// Populates the fields list view.
			PopulateFieldsListView();

            // Populates the index list view.
            PopulateIndexesListView();

            // Populates the childs list view.
            PopulateChildrenListView();

            // Populates the childs list view.
            PopulateParentsListView();
        }

        /// <summary>
        /// Invoked when the form is about to close.
        /// </summary>
        private void PropertiesForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (this.DialogResult != DialogResult.OK)
                return;

            e.Cancel = (!PopulateObject());
        }

        /// <summary>
        /// Populates the object with data.
        /// </summary>
        private bool PopulateObject()
        {
            if (!CheckControls())
                return false;

            // Set the status of all properties.
            foreach (ListViewItem item in fieldsListView.Items)
            {
                 ((Field)item.Tag).Generate = item.Checked;
            }

            // Set the status of all index.
            foreach (ListViewItem item in indexesListView.Items)
            {
                ((Index)item.Tag).Generate = item.Checked;
            }

            // Set the status of all children.
            foreach (ListViewItem item in childrenListView.Items)
            {
                ((Relation)item.Tag).Generate = item.Checked;
            }

            // Set the status of all parents.
            foreach (ListViewItem item in parentsListView.Items)
            {
                ((Relation)item.Tag).Generate = item.Checked;
            }


            _table.TableName = tableNameTextBox.Text.Trim();
            _table.PersistentName = persistentNameTextBox.Text.Trim();
            _table.BaseClass = baseclassTextBox.Text; 
            _table.NoPool = noPoolCheckBox.Checked;
            _table.PoolSize = (int)poolSizeNumericUpDown.Value;

            _table.Rights = PersistentRights.None;
            if (loadCheckBox.Checked)
                _table.Rights |= PersistentRights.Load;
            if (insertCheckBox.Checked)
                _table.Rights |= PersistentRights.Insert;
            if (updateCheckBox.Checked)
                _table.Rights |= PersistentRights.Update;
            if (deleteCheckBox.Checked)
                _table.Rights |= PersistentRights.Delete;

            return true;
        }

        #region CheckControls

        /// <summary>
        /// Checks the UI controls.
        /// </summary>
        private bool CheckControls()
        {
            // Check if a valid table name has been given.
            if (tableNameTextBox.Text.Trim().Length == 0)
            {
                // Switch to the first tab page.
                tabControl1.SelectedTab = generalTabPage;
				MessageBox.Show(Resources.MissingTableNameDescription, Resources.ProgramName, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }

            // Check if a valid persistent name has been given.
            if (persistentNameTextBox.Text.Trim().Length == 0)
            {
                // Switch to the first tab page.
                tabControl1.SelectedTab = generalTabPage;
				MessageBox.Show(Resources.MissingPersistentNameDescription, Resources.ProgramName, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }

            // Check if there have been given fields.
            if (_table.Fields.Count == 0)
            {
                // Switch to the second tab page.
                tabControl1.SelectedTab = propertiesTabPage;
                MessageBox.Show(Resources.MissingPropertiesText, Resources.ProgramName, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }

            return true;
        }

        #endregion

        #region propertiesTabPage

        /// <summary>
		/// Populates the fields list view.
		/// </summary>
		private void PopulateFieldsListView()
		{
			fieldsListView.Items.Clear();

			// Suspend the layout of the grid.
			fieldsListView.SuspendLayout();
			try
			{
				// Loop over all fields and add them to the list view.
				foreach (Field field in _table.Fields)
				{
					ListViewItem item = new ListViewItem();
					PopulateListViewItem(field, item);
					fieldsListView.Items.Add(item);
				}
			}
			finally
			{
				// Resume the layout of the grid.
				fieldsListView.ResumeLayout();
			}
		}

		/// <summary>
		/// Creates a listview item for the given field.
		/// </summary>
		/// <param name="field">The field that is used to generate the list view item.</param>
		/// <param name="item">The listview item that is created.</param>
		private void PopulateListViewItem(Field field, ListViewItem item)
		{
			// Remove all subitems.
			item.SubItems.Clear();
            item.Tag = field;

			item.Checked = field.Generate;
			item.ImageIndex = 0;
			item.Text = field.FieldName;
			item.SubItems.Add(field.PropertyName);
			item.SubItems.Add(field.Identifier ? Resources.Yes : Resources.No);
			item.SubItems.Add(field.AllowDBNull ? Resources.Yes : Resources.No);
            item.SubItems.Add(field.Description);
		}

        

		/// <summary>
		/// Invoked each time the selected item changed.
		/// </summary>
		private void fieldsListView_SelectedIndexChanged(object sender, EventArgs e)
		{
			UpdatePropertiesControlsState();
		}

		/// <summary>
		/// Updates the state of the controls on Properties page.
		/// </summary>
		private void UpdatePropertiesControlsState()
		{
			editButton.Enabled = (fieldsListView.SelectedItems.Count > 0);
			deleteButton.Enabled = (fieldsListView.SelectedItems.Count > 0);
		}

		/// <summary>
		/// Returns the selected field or null if no field has been selected.
		/// </summary>
		private Field GetSelectedField()
		{
			if (fieldsListView.SelectedItems.Count == 0)
				return null;

            return fieldsListView.SelectedItems[0].Tag as Field;
		}

		/// <summary>
		/// Invoked each time the check all button is clicked.
		/// </summary>
		private void checkAllButton_Click(object sender, EventArgs e)
		{
            // Loop over all items and check.
            foreach (ListViewItem item in fieldsListView.Items)
            {
                item.Checked = true;
            }
		}

		/// <summary>
		/// Invoked each time the uncheck all button is clicked.
		/// </summary>
		private void unCheckAllButton_Click(object sender, EventArgs e)
		{
            // Loop over all items and uncheck.
            foreach (ListViewItem item in fieldsListView.Items)
            {
                item.Checked = false;
            }
        }

        /// <summary>
        /// Invoked each time the edit button is clicked.
        /// </summary>
        private void editButton_Click(object sender, EventArgs e)
        {
            Field field = GetSelectedField();
            if (field == null)
                return;

            // Show the dialog to edit the property.
            if (new PropertyDialog(DialogModes.Edit, field).ShowDialog(this) == DialogResult.OK)
            {
                PopulateListViewItem(field, fieldsListView.SelectedItems[0]);
            }
        }

        /// <summary>
        /// Invoked each time the add button is clicked.
        /// </summary>
        private void addButton_Click(object sender, EventArgs e)
        {
            Field field = new Field();
            field.Generate = true;

            // Show the dialog to edit the property.
            if (new PropertyDialog(DialogModes.New, field).ShowDialog(this) == DialogResult.OK)
            {
                // Create a new listview item.
                ListViewItem item = new ListViewItem();
                PopulateListViewItem(field, item);

                // Add the field to the list and to the listview.
                fieldsListView.Items.Add(item);
                _table.Fields.Add(field);
            }
        }

        /// <summary>
        /// Invoked each time the remove button is clicked.
        /// </summary>
        private void deleteButton_Click(object sender, EventArgs e)
        {
            Field field = GetSelectedField();
            if (field == null)
                return;

            // Asks the user if the table should be deleted.
            if (MessageBox.Show(Resources.DeletePropertyText, Resources.ProgramName, MessageBoxButtons.YesNo, MessageBoxIcon.Question) ==
                DialogResult.Yes)
            {
                // Removes the table from the listview and the list.
                _table.Fields.Remove(field);
                fieldsListView.Items.Remove(fieldsListView.SelectedItems[0]);
            }
        }

        #endregion propertiesTabPage

        #region indexesTabPage

        /// <summary>
        /// Populates the indexes list view.
        /// </summary>
        private void PopulateIndexesListView()
        {
            indexesListView.Items.Clear();

            // Suspend the layout of the grid.
            indexesListView.SuspendLayout();
            try
            {
                // Loop over all index and add them to the list view.
                foreach (Index index in _table.Indexes)
                {
                    ListViewItem item = new ListViewItem();
                    PopulateListViewItem(index, item);
                    indexesListView.Items.Add(item);
                }
            }
            finally
            {
                // Resume the layout of the grid.
                indexesListView.ResumeLayout();
            }
        }

        /// <summary>
        /// Creates a listview item for the given index.
        /// </summary>
        /// <param name="field">The index that is used to generate the list view item.</param>
        /// <param name="item">The listview item that is created.</param>
        private void PopulateListViewItem(Index index, ListViewItem item)
        {
            // Remove all subitems.
            item.SubItems.Clear();
            item.Tag = index;

            item.Checked = index.Generate;
            item.ImageIndex = index.PrimaryKey ? 1 : 2;
            item.Text = index.Name;
            item.SubItems.Add(index.FieldsName);
        }

        /// <summary>
        /// Returns the selected index or null if no index has been selected.
        /// </summary>
        private Index GetSelectedIndex()
        {
            if (indexesListView.SelectedItems.Count == 0)
                return null;

            return indexesListView.SelectedItems[0].Tag as Index;
        }


        /// <summary>
        /// Invoked each time the selected index changed.
        /// </summary>
        private void indexesListView_SelectedIndexChanged(object sender, EventArgs e)
        {
            UpdateIndexesControlsState();
        }

        /// <summary>
        /// Updates the state of the controls on Properties page.
        /// </summary>
        private void UpdateIndexesControlsState()
        {
            editIndexbutton.Enabled = (indexesListView.SelectedItems.Count > 0);
            removeIndexbutton.Enabled = (indexesListView.SelectedItems.Count > 0);
        }

        /// <summary>
        /// Invoked each time the check all button is clicked.
        /// </summary>
        private void checkAllIndexbutton_Click(object sender, EventArgs e)
        {
            // Loop over all items and check.
            foreach (ListViewItem item in indexesListView.Items)
            {
                item.Checked = true;
            }
        }

        /// <summary>
        /// Invoked each time the uncheck all button is clicked.
        /// </summary>
        private void unCheckAllIndexbutton_Click(object sender, EventArgs e)
        {
            // Loop over all items and uncheck.
            foreach (ListViewItem item in indexesListView.Items)
            {
                item.Checked = false;
            }
        }

        /// <summary>
        /// Invoked each time the editindex button is clicked.
        /// </summary>
        private void editIndexbutton_Click(object sender, EventArgs e)
        {
            Index index = GetSelectedIndex();
            if (index == null)
                return;

            // Show the dialog to edit the index.
            if (new IndexItemDialog(DialogModes.Edit, index, _table).ShowDialog(this) == DialogResult.OK)
            {
                PopulateListViewItem(index, indexesListView.SelectedItems[0]);
            }

        }

        /// <summary>
        /// Invoked each time the addindex button is clicked.
        /// </summary>
        private void addIndexbutton_Click(object sender, EventArgs e)
        {
            Index index = new Index();
            index.Generate = true;

            // Show the dialog to edit the index.
            if (new IndexItemDialog(DialogModes.New, index, _table).ShowDialog(this) == DialogResult.OK)
            {
                // Create a new listview item.
                ListViewItem item = new ListViewItem();
                PopulateListViewItem(index, item);

                // Add the index to the list and to the listview.
                indexesListView.Items.Add(item);
                _table.Indexes.Add(index);
            }
        }

        /// <summary>
        /// Invoked each time the removeindex button is clicked.
        /// </summary>
        private void removeIndexbutton_Click(object sender, EventArgs e)
        {
            Index index = GetSelectedIndex();
            if (index == null)
                return;

            // Asks the user if the table should be deleted.
            if (MessageBox.Show(Resources.DeletePropertyText, Resources.ProgramName, MessageBoxButtons.YesNo, MessageBoxIcon.Question) ==
                DialogResult.Yes)
            {
                // Removes the table from the listview and the list.
                _table.Indexes.Remove(index);
                indexesListView.Items.Remove(indexesListView.SelectedItems[0]);
            }
        }

        #endregion indexesTabPage

        #region childrenTabPage

        /// <summary>
        /// Populates the indexes list view.
        /// </summary>
        private void PopulateChildrenListView()
        {
            childrenListView.Items.Clear();

            // Suspend the layout of the grid.
            childrenListView.SuspendLayout();
            try
            {
                // Loop over all relations and add them to the list view.
                foreach (Relation relation in _relations)
                {
                    if (relation.ParentTable == _table)
                    {
                        ListViewItem item = new ListViewItem();
                        PopulateChildrenListViewItem(relation, item);

                        // Add the weak entity if required.
                        if (relation.WeakTable != null)
                            item.SubItems.Add(relation.WeakTable.PersistentName);

                        // Add the item to the listview.
                        childrenListView.Items.Add(item);
                    }
                }
            }
            finally
            {
                // Resume the layout of the grid.
                childrenListView.ResumeLayout();
            }
            UpdateChildrenControlsState();
        }

        /// <summary>
        /// Creates a listview item for the given relation.
        /// </summary>
        /// <param name="field">The index that is used to generate the list view item.</param>
        /// <param name="item">The listview item that is created.</param>
        private void PopulateChildrenListViewItem(Relation relation, ListViewItem item)
        {
            // Remove all subitems.
            item.SubItems.Clear();
            item.Tag = relation;

            item.Text = CreateMember(relation, true);
            item.ImageIndex = 3;
            item.Checked = relation.Generate;
            item.SubItems.Add(relation.ChildTable.PersistentName);
            item.SubItems.Add(CreateMember(relation, false));
        }


        /// <summary>
        /// Creates the parent member of the relations.
        /// </summary>
        /// <param name="parent">True to create the parent member.</param>
        /// <param name="relation">The relation that's member are created.</param>
        private string CreateMember(Relation relation, bool parent)
        {
            string result = string.Empty; ;

            foreach (RelationItem item in relation.Relations)
            {
                if (result.Length > 0)
                    result += ";";

                if (parent)
                    result += item.ParentField.PropertyName;
                else
                    result += item.ChildField.PropertyName;
            }

            return result;
        }

        /// <summary>
        /// Updates the state of the controls.
        /// </summary>
        private void UpdateChildrenControlsState()
        {
            editChildButton.Enabled = (childrenListView.SelectedItems.Count > 0);
            removeChildButton.Enabled = (childrenListView.SelectedItems.Count > 0);
        }


        /// <summary>
        /// Invoked each time the selected index changed.
        /// </summary>
        private void childrenListView_SelectedIndexChanged(object sender, EventArgs e)
        {
            UpdateChildrenControlsState();
        }

        /// <summary>
        /// Invoked each time the button to select (check) all children is clicked.
        /// </summary>
        private void checkallChildrenButton_Click(object sender, EventArgs e)
        {
            // Loop over all items and check.
            foreach (ListViewItem item in childrenListView.Items)
            {
                item.Checked = true;
            }
        }

        /// <summary>
        /// Invoked each time the button to select (check) all children is clicked.
        /// </summary>
        private void uncheckallChildrenButton_Click(object sender, EventArgs e)
        {
            // Loop over all items and uncheck.
            foreach (ListViewItem item in childrenListView.Items)
            {
                item.Checked = false;
            }
        }

        /// <summary>
        /// Returns the selected field or null if no field has been selected.
        /// </summary>
        private Relation GetSelectedChild()
        {
            if (childrenListView.SelectedItems.Count == 0)
                return null;

            return childrenListView.SelectedItems[0].Tag as Relation;
        }

        /// <summary>
        /// Invoked each time the edit relation button is clicked.
        /// </summary>
        private void editChildButton_Click(object sender, EventArgs e)
        {
            Relation relation = GetSelectedChild();
            if (relation == null)
                return;

            // Check if we got an OK.
            if (new RelationDialog(DialogModes.Edit, relation).ShowDialog(this) == DialogResult.OK)
            {
                // Update the list view item.
                PopulateChildrenListViewItem(relation, childrenListView.SelectedItems[0]);
            }
        }

        /// <summary>
        /// Invoked each time the add button is clicked.
        /// </summary>
        private void addChildButton_Click(object sender, EventArgs e)
        {
            Relation relation = new Relation();
            relation.Generate = true;
            relation.ParentTable = _table;


            // Check if we got an OK.
            if (new RelationDialog(DialogModes.New, relation).ShowDialog(this) == DialogResult.OK)
            {
                // Create a new listview item and add it to the list view.
                ListViewItem item = new ListViewItem();
                PopulateChildrenListViewItem(relation, item);
                childrenListView.Items.Add(item);

                // Add the relation to the list of relations.
                _relations.Add(relation);
            }
        }

        /// <summary>
        /// Invoked each time the remove button is clicked.
        /// </summary>
        private void removeChildButton_Click(object sender, EventArgs e)
        {
            Relation relation = GetSelectedChild();
            if (relation == null)
                return;

            // Asks the user if the table should be deleted.
            if (MessageBox.Show(Resources.DeleteRelationText, Resources.ProgramName, MessageBoxButtons.YesNo, MessageBoxIcon.Question) ==
                DialogResult.Yes)
            {
                // Removes the relation from the listview and the list.
                _relations.Remove(relation);
                childrenListView.Items.Remove(childrenListView.SelectedItems[0]);
            }
        }


        #endregion childrenTabPage

        #region parentsTabPage

        /// <summary>
        /// Populates the indexes list view.
        /// </summary>
        private void PopulateParentsListView()
        {
            parentsListView.Items.Clear();

            // Suspend the layout of the grid.
            parentsListView.SuspendLayout();
            try
            {
                // Loop over all relations and add them to the list view.
                foreach (Relation relation in _relations)
                {
                    if (relation.ChildTable  == _table)
                    {
                        ListViewItem item = new ListViewItem();
                        PopulateParentsListViewItem(relation, item);

                        // Add the weak entity if required.
                        if (relation.WeakTable != null)
                            item.SubItems.Add(relation.WeakTable.PersistentName);

                        // Add the item to the listview.
                        parentsListView.Items.Add(item);
                    }
                }
            }
            finally
            {
                // Resume the layout of the grid.
                parentsListView.ResumeLayout();
            }
            UpdateParentsControlsState();
        }

        /// <summary>
        /// Creates a listview item for the given relation.
        /// </summary>
        /// <param name="field">The index that is used to generate the list view item.</param>
        /// <param name="item">The listview item that is created.</param>
        private void PopulateParentsListViewItem(Relation relation, ListViewItem item)
        {
            // Remove all subitems.
            item.SubItems.Clear();
            item.Tag = relation;

            item.Text = CreateMember(relation, false);
            item.ImageIndex = 5;
            item.Checked = relation.Generate;
            item.SubItems.Add(relation.ParentTable.PersistentName);
            item.SubItems.Add(CreateMember(relation, true));
        }

        /// <summary>
        /// Updates the state of the controls.
        /// </summary>
        private void UpdateParentsControlsState()
        {
            editParentButton.Enabled = (parentsListView.SelectedItems.Count > 0);
            removeParentButton.Enabled = (parentsListView.SelectedItems.Count > 0);
        }

        /// <summary>
        /// Invoked each time the selected parents changed.
        /// </summary>
        private void parentsListView_SelectedIndexChanged(object sender, EventArgs e)
        {
            UpdateParentsControlsState();
        }

        /// <summary>
        /// Invoked each time the button to unselect (uncheck) all parent is clicked.
        /// </summary>
        private void checkallParentsButton_Click(object sender, EventArgs e)
        {
            // Loop over all items and check.
            foreach (ListViewItem item in parentsListView.Items)
            {
                item.Checked = false;
            }
        }

        /// <summary>
        /// Invoked each time the button to unselect (uncheck) all parent is clicked.
        /// </summary>
        private void uncheckallParentsButton_Click(object sender, EventArgs e)
        {
            // Loop over all items and uncheck.
            foreach (ListViewItem item in parentsListView.Items)
            {
                item.Checked = false;
            }
        }

        /// <summary>
        /// Returns the selected field or null if no field has been selected.
        /// </summary>
        private Relation GetSelectedParent()
        {
            if (parentsListView.SelectedItems.Count == 0)
                return null;

            return parentsListView.SelectedItems[0].Tag as Relation;
        }

        /// <summary>
        /// Invoked each time the add button is clicked.
        /// </summary>
        private void addParentButton_Click(object sender, EventArgs e)
        {
            Relation relation = new Relation();
            relation.Generate = true;
            relation.ChildTable = _table;

            // Check if we got an OK.
            if (new RelationDialog(DialogModes.New, relation).ShowDialog(this) == DialogResult.OK)
            {
                // Create a new listview item and add it to the list view.
                ListViewItem item = new ListViewItem();
                PopulateParentsListViewItem(relation, item);
                parentsListView.Items.Add(item);

                // Add the relation to the list of relations.
                _relations.Add(relation);
            }
        }

        /// <summary>
        /// Invoked each time the edit relation button is clicked.
        /// </summary>
        private void editParentButton_Click(object sender, EventArgs e)
        {
            Relation relation = GetSelectedParent();
            if (relation == null)
                return;

            // Check if we got an OK.
            if (new RelationDialog(DialogModes.Edit, relation).ShowDialog(this) == DialogResult.OK)
            {
                // Update the list view item.
                PopulateChildrenListViewItem(relation, parentsListView.SelectedItems[0]);
            }
        }


        /// <summary>
        /// Invoked each time the remove button is clicked.
        /// </summary>
        private void removeParentButton_Click(object sender, EventArgs e)
        {
            Relation relation = GetSelectedParent();
            if (relation == null)
                return;

            // Asks the user if the table should be deleted.
            if (MessageBox.Show(Resources.DeleteRelationText, Resources.ProgramName, MessageBoxButtons.YesNo, MessageBoxIcon.Question) ==
                DialogResult.Yes)
            {
                // Removes the relation from the listview and the list.
                _relations.Remove(relation);
                parentsListView.Items.Remove(parentsListView.SelectedItems[0]);
            }
        }
        #endregion parentsTabPage
    }
}