#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Chili.Opf3.Wizard.Shared.Storages;
using Chili.Opf3.Wizard.Shared.UI;
using Chili.Opf3.Wizard.Properties;
using System.Text.RegularExpressions;

#endregion

namespace Chili.Opf3.Wizard.UI.Forms
{
	/// <summary>
	/// This dialog allows to edit a single property of the persistent.
	/// </summary>
	public partial class PropertyDialog : BaseDialog
	{
		#region Constants

		private const string STRING = "String";
		private const string SHORT = "Int16";
		private const string INT = "Int32";
		private const string LONG = "Int64";
		private const string DOUBLE = "Double";
		private const string FLOAT = "Single";
		private const string DECIMAL = "Decimal";
		private const string BOOLEAN = "Boolean";
		private const string DATETIME = "DateTime";
		private const string GUID = "Guid";
		private const string BLOB = "Byte[] (= Chili.Opf3.Blob)";
		private const string BYTE = "Byte";

		#endregion

		private Field _field;

		/// <summary>
		/// Creates a new instance of the <see cref="PropertyDialog">PropertyDialog Class</see>.
		/// </summary>
		/// <param name="field">The field that is converted to a property in the persistent.</param>
		/// <param name="mode">The mode of the dialog.</param>
		public PropertyDialog(DialogModes mode, Field field)
		{
			InitializeComponent();

			if (field == null)
				throw new ArgumentNullException("field", "Field is null.");
			_field = field;

			// Populate the type combobox.
			PopulateTypeComboBox();
			// Enable only if the type of the field is a string.
			lengthNumericUpDown.Enabled = (_field.Type == typeof(string));

			if (mode == DialogModes.New)
			{
				// Set the title of the form.
				this.Text = Resources.NewPropertyDialogTitle;
			}
			else
			{
				// Set the title of the form.
				this.Text = string.Format(Resources.EditPropertyDialogTitle, _field.PropertyName);
				// Populate the controls with data.
				PopulateControls();
			}
		}

		/// <summary>
		/// Populates the types combobox.
		/// </summary>
		private void PopulateTypeComboBox()
		{
			typeComboBox.Items.Add(STRING);
			typeComboBox.Items.Add(SHORT);
			typeComboBox.Items.Add(INT);
			typeComboBox.Items.Add(LONG);
			typeComboBox.Items.Add(DOUBLE);
			typeComboBox.Items.Add(FLOAT);
			typeComboBox.Items.Add(DECIMAL);
			typeComboBox.Items.Add(BOOLEAN);
			typeComboBox.Items.Add(DATETIME);
			typeComboBox.Items.Add(BYTE);
			typeComboBox.Items.Add(GUID);
			typeComboBox.Items.Add(BLOB);

			// Select the first item in the combobox.
			typeComboBox.SelectedIndex = 0;
		}

		/// <summary>
		/// Populates the UI controls with data.
		/// </summary>
		private void PopulateControls()
		{
			fieldNameTextBox.Text = _field.FieldName;
			propertyNameTextBox.Text = _field.PropertyName;
			objectFieldNameTextBox.Text = _field.ObjectFieldName;
			
			allowDBNullCheckBox.Checked = _field.AllowDBNull;
			identifierCheckBox.Checked = _field.Identifier;
			autoNumberCheckBox.Checked = _field.AutoNumber;

			SelectTypeInComboBox();

			lengthNumericUpDown.Value = _field.Length;

			dontInsertCheckBox.Checked = ((_field.PersistBehaviour & PersistBehaviours.DontInsert) != 0);
			dontUpdateCheckBox.Checked = ((_field.PersistBehaviour & PersistBehaviours.DontUpdate) != 0);
		}

		/// <summary>
		/// Selects the type in the combobox.
		/// </summary>
		private void SelectTypeInComboBox()
		{
			for (int i = 0; i < typeComboBox.Items.Count; i++)
			{
				string type = (string)typeComboBox.Items[i];
				if (_field.Type == GetTypeFromString(type))
				{
					typeComboBox.SelectedIndex = i;
					return;
				}
			}
		}

		/// <summary>
		/// Converts the type string to a real type.
		/// </summary>
		public Type GetTypeFromString(string type)
		{
			if (type == STRING)
				return typeof(string);
			else if (type == SHORT)
				return typeof(short);
			else if (type == INT)
				return typeof(int);
			else if (type == LONG)
				return typeof(long);
			else if (type == DOUBLE)
				return typeof(double);
			else if (type == FLOAT)
				return typeof(float);
			else if (type == DECIMAL)
				return typeof(decimal);
			else if (type == BOOLEAN)
				return typeof(bool);
			else if (type == DATETIME)
				return typeof(DateTime);
			else if (type == BLOB)
				return typeof(byte[]);
			else if (type == GUID)
				return typeof(Guid);
			else if (type == BYTE)
				return typeof(byte);

			return null;
		}

		/// <summary>
		/// Invoked each time the form is about to close.
		/// </summary>
		private void PropertyDialog_FormClosing(object sender, FormClosingEventArgs e)
		{
			if (this.DialogResult != DialogResult.OK)
				return;

			e.Cancel = (!PopulateObject());
		}

		/// <summary>
		/// Populates the object from the data of the form.
		/// </summary>
		private bool PopulateObject()
		{
			// Check if a valid property name has been given.
			if (propertyNameTextBox.Text.Trim().Length == 0)
			{
				MessageBox.Show(Resources.MissingPropertyNameDescription, Resources.ProgramName, MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}
			// Check if a valid object field name has been given.
			if (objectFieldNameTextBox.Text.Trim().Length == 0)
			{
				MessageBox.Show(Resources.MissingObjectFieldNameDescription, Resources.ProgramName, MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}

			// Check if the storage field name has been given.
			if (fieldNameTextBox.Text.Trim().Length == 0)
			{
				MessageBox.Show(Resources.MissingStorageFieldNameDesription, Resources.ProgramName, MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}

			_field.FieldName = fieldNameTextBox.Text;
			_field.AllowDBNull = allowDBNullCheckBox.Checked;
			_field.AutoNumber = autoNumberCheckBox.Checked;
			_field.Identifier = identifierCheckBox.Checked;
			_field.Length = (int)lengthNumericUpDown.Value;
			_field.ObjectFieldName = objectFieldNameTextBox.Text.Trim();
			_field.PropertyName = propertyNameTextBox.Text.Trim();
			_field.Type = GetTypeFromString((string)typeComboBox.SelectedItem);

			_field.PersistBehaviour = PersistBehaviours.None;
			if (dontInsertCheckBox.Checked)
				_field.PersistBehaviour |= PersistBehaviours.DontInsert;
			if (dontUpdateCheckBox.Checked)
				_field.PersistBehaviour |= PersistBehaviours.DontUpdate;

			return true;
		}

		/// <summary>
		/// Invoked each time the selected index changed.
		/// </summary>
		private void typeComboBox_SelectedIndexChanged(object sender, EventArgs e)
		{
			// Enable only for strings.
			lengthNumericUpDown.Enabled = (GetTypeFromString((string)typeComboBox.SelectedItem) == typeof(string));
		}
	}
}