#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Chili.Opf3.Wizard.Shared.Storages;
using Chili.Opf3.Wizard.Shared.UI;
using Chili.Opf3.Wizard.Properties;
using Chili.Opf3.Wizard.Shared;
using System.Text.RegularExpressions;

#endregion

namespace Chili.Opf3.Wizard.UI.Forms
{
	/// <summary>
	/// This dialog allows to edit a single relation between persistent objects.
	/// </summary>
	public partial class RelationDialog : BaseDialog
	{
		#region Constants

		private const string ONE_TO_ONE = "One-To-One";
		private const string ONE_TO_MANY = "One-To-Many";
		private const string MANY_TO_MANY = "Many-To-Many";

		#endregion

		private Relation _relation;

		/// <summary>
		/// Creates a new instance of the <see cref="RelationDialog">RelationDialog Class</see>.
		/// </summary>
		/// <param name="mode">The mode of the dialog.</param>
		/// <param name="relation">The relation that is edited by the dialog.</param>
		public RelationDialog(DialogModes mode, Relation relation)
		{
			InitializeComponent();

			if (relation == null)
				throw new ArgumentNullException("relation", "Relation is null.");
			_relation = relation;

			PopulateRelationTypeComboBox();
			PopulatePersistentComboBoxes();

			if (mode == DialogModes.New)
			{
				// Set the title of the form.
				this.Text = Resources.NewRelationDialogTitle;
			}
			else
			{
				// Set the title of the form.
				this.Text = Resources.EditRelationDialogTitle;
				// Populate the controls with data.
				PopulateControls();
			}
		}

		#region PopulatePersistentComboBoxes

		/// <summary>
		/// Populates the comboboxes showing the parent persistents, child persistents and weak persistents.
		/// </summary>
		private void PopulatePersistentComboBoxes()
		{
			Project project = Session.Current.Project;

			// Clear all items in the comboboxes.
			weakObjectsComboBox.Items.Clear();
			parentObjectsComboBox.Items.Clear();
			childObjectsComboBox.Items.Clear();

			bool foundChildTable = false;
			bool foundParentTable = false;
			bool foundWeakTable = false;

			// Loop over all tables and add them to the list.
			foreach (Table table in project.Tables)
			{
				if (_relation.ParentTable != null && table.PersistentName == _relation.ParentTable.PersistentName)
					foundParentTable = true;
				if (_relation.ChildTable != null && table.PersistentName == _relation.ChildTable.PersistentName)
					foundChildTable = true;
				if (_relation.WeakTable != null && table.PersistentName == _relation.WeakTable.PersistentName)
					foundWeakTable = true;

				weakObjectsComboBox.Items.Add(table);
				parentObjectsComboBox.Items.Add(table);
				childObjectsComboBox.Items.Add(table);
			}

			// Add the child, parent or weak table if not found in the list of tables.
			if (_relation.ParentTable != null && !foundParentTable)
				childObjectsComboBox.Items.Add(_relation.ParentTable);
			if (_relation.ChildTable != null && !foundChildTable)
				childObjectsComboBox.Items.Add(_relation.ChildTable);
			if (_relation.WeakTable != null && !foundWeakTable)
				childObjectsComboBox.Items.Add(_relation.WeakTable);

			if (weakObjectsComboBox.Items.Count > 0)
				weakObjectsComboBox.SelectedIndex = 0;
			if (parentObjectsComboBox.Items.Count > 0)
				parentObjectsComboBox.SelectedIndex = 0;
			if (childObjectsComboBox.Items.Count > 0)
				childObjectsComboBox.SelectedIndex = 0;
		}

		#endregion

		/// <summary>
		/// Populates the relation type combobox.
		/// </summary>
		private void PopulateRelationTypeComboBox()
		{
			relationTypeComboBox.Items.Clear();

			relationTypeComboBox.Items.Add(ONE_TO_MANY);
			relationTypeComboBox.Items.Add(ONE_TO_ONE);
			relationTypeComboBox.Items.Add(MANY_TO_MANY);

			// Select the first item.
			relationTypeComboBox.SelectedIndex = 0;
		}

		/// <summary>
		/// Populates the UI controls.
		/// </summary>
		private void PopulateControls()
		{
            nameTextBox.Text = _relation.Name;
			createInChildCheckBox.Checked = _relation.CreateInChildPersistent;
			createInParentCheckBox.Checked = _relation.CreateInParentPersistent;
			parentPropertyTextBox.Text = _relation.ParentProperty;
			childPropertyTextBox.Text = _relation.ChildProperty;

			dontInsertCheckBox.Checked = ((_relation.PersistBehaviour & PersistBehaviours.DontInsert) != 0);
			dontUpdateCheckBox.Checked = ((_relation.PersistBehaviour & PersistBehaviours.DontUpdate) != 0);

			// Select the type of relation.
			SelectRelationType();

			// Select the property in the combobox.
			SelectInObjectComboBox(parentObjectsComboBox, _relation.ParentTable);
			SelectInObjectComboBox(childObjectsComboBox, _relation.ChildTable);
			if (_relation.RelationType == RelationTypes.ManyToMany)
			{
				SelectInObjectComboBox(weakObjectsComboBox, _relation.WeakTable);
				// Populate the list view with the weak relations.
				PopulateRelationsListView(_relation.WeakRelations, weakRelationsListView);
			}

			// Populate the list view with the relations.
			PopulateRelationsListView(_relation.Relations, relationsListView);
		}

		/// <summary>
		/// Populates the list view with the relations.
		/// </summary>
		/// <param name="relations">List of items that is populated.</param>
		/// <param name="listView">The list view that is populated.</param>
		private void PopulateRelationsListView(List<RelationItem> relations, ListView listView)
		{
			listView.Items.Clear();

			// Add all items in the relations list.
			foreach (RelationItem ri in relations)
			{
				ListViewItem item = new ListViewItem();
				PopulateRelationListViewItem(ri, item);

				listView.Items.Add(item);
			}
		}

		/// <summary>
		/// Populates a relation list view item.
		/// </summary>
		/// <param name="ri">The relation item.</param>
		/// <param name="item">The list view item.</param>
		private void PopulateRelationListViewItem(RelationItem ri, ListViewItem item)
		{
			// Remove all subitems.
			item.SubItems.Clear();

			item.ImageIndex = 0;
			item.Text = ri.ParentField.PropertyName;
			item.SubItems.Add(ri.ChildField.PropertyName);
		}

		/// <summary>
		/// Selects the given table in the given combobox.
		/// </summary>
		/// <param name="combobox">The combobox where the table is selected.</param>
		/// <param name="table">The table that is selected.</param>
		private void SelectInObjectComboBox(ComboBox combobox, Table table)
		{
			for (int i = 0; i < combobox.Items.Count; i++)
			{
				Table t = (Table)combobox.Items[i];
				if (t.PersistentName == table.PersistentName)
				{
					combobox.SelectedIndex = i;
					return;
				}
			}
		}

		#region SelectRelationType

		/// <summary>
		/// Select the relation type.
		/// </summary>
		private void SelectRelationType()
		{
			switch (_relation.RelationType)
			{
				case RelationTypes.OneToMany:
					relationTypeComboBox.SelectedIndex = 0;
					break;
				case RelationTypes.OneToOne:
					relationTypeComboBox.SelectedIndex = 1;
					break;
				case RelationTypes.ManyToMany:
					relationTypeComboBox.SelectedIndex = 2;
					break;
			}
		}

		#endregion

		#region GetSelectedRelationType

		/// <summary>
		/// Returns the selected relation type.
		/// </summary>
		private RelationTypes GetSelectedRelationType()
		{
			if (relationTypeComboBox.SelectedIndex == 2)
				return RelationTypes.ManyToMany;
			if (relationTypeComboBox.SelectedIndex == 1)
				return RelationTypes.OneToOne;
			if (relationTypeComboBox.SelectedIndex == 0)
				return RelationTypes.OneToMany;
			return RelationTypes.OneToMany;
		}

		#endregion

		/// <summary>
		/// Invoked each time the relation type combobox's selected index changed.
		/// </summary>
		private void relationTypeComboBox_SelectedIndexChanged(object sender, EventArgs e)
		{
			RelationTypes relationType = GetSelectedRelationType();

			weakObjectsComboBox.Enabled = (relationType == RelationTypes.ManyToMany);
			weakRelationsListView.Enabled = (relationType == RelationTypes.ManyToMany);
			addWeakRelationButton.Enabled = (relationType == RelationTypes.ManyToMany);

			// Set the column header according to the selected relation type.
			if (relationType != RelationTypes.ManyToMany)
				childFieldColumnHeader.Text = Resources.PropertyInChildPersistent;
			else
				childFieldColumnHeader.Text = Resources.PropertyInWeakPersistent;

			UpdateControlsState();
		}

		/// <summary>
		/// Invoked each time the selected index of the relations list view changed.
		/// </summary>
		private void relationsListView_SelectedIndexChanged(object sender, EventArgs e)
		{
			UpdateControlsState();
		}

		/// <summary>
		/// Updates the state of the controls.
		/// </summary>
		private void UpdateControlsState()
		{
			editRelationButton.Enabled = (relationsListView.SelectedItems.Count > 0);
			deleteRelationButton.Enabled = (relationsListView.SelectedItems.Count > 0);

			RelationTypes relationType = GetSelectedRelationType();
			editWeakRelationButton.Enabled = (weakRelationsListView.SelectedItems.Count > 0 && relationType == RelationTypes.ManyToMany);
			deleteWeakRelationButton.Enabled = (weakRelationsListView.SelectedItems.Count > 0 && relationType == RelationTypes.ManyToMany);
		}

		#region addRelationButton_Click

		/// <summary>
		/// Invoked each time the new relation button is clicked.
		/// </summary>
		private void addRelationButton_Click(object sender, EventArgs e)
		{
			RelationItem relationItem = new RelationItem();

			// Get the parent table.
			Table parentTable = (Table)parentObjectsComboBox.SelectedItem;
			if (parentTable == null)
				return;

			Table childTable = null;
			// Get the child table. This one may be vary depending on the relation type.
			if (GetSelectedRelationType() == RelationTypes.ManyToMany)
				childTable = (Table)weakObjectsComboBox.SelectedItem;
			else
				childTable = (Table)childObjectsComboBox.SelectedItem;
			if (childTable == null)
				return;

			// Check if we got an OK.
			if (new RelationsItemDialog(DialogModes.New, relationItem, parentTable, childTable).ShowDialog(this) ==
				DialogResult.OK)
			{
				ListViewItem item = new ListViewItem();
	
				// Populate the list view item with the new data.
				PopulateRelationListViewItem(relationItem, item);
				relationsListView.Items.Add(item);
				_relation.Relations.Add(relationItem);
			}
		}

		#endregion

		#region editRelationButton_Click

		/// <summary>
		/// Invoked each time the edit relation button is clicked.
		/// </summary>
		private void editRelationButton_Click(object sender, EventArgs e)
		{
			RelationItem relationItem = GetSelectedRelationItem();
			if (relationItem == null)
				return;

			// Get the parent table.
			Table parentTable = (Table)parentObjectsComboBox.SelectedItem;
			if (parentTable == null)
				return;

			Table childTable = null;
			// Get the child table. This one may be vary depending on the relation type.
			if (GetSelectedRelationType() == RelationTypes.ManyToMany)
				childTable = (Table)weakObjectsComboBox.SelectedItem;
			else
				childTable = (Table)childObjectsComboBox.SelectedItem;
			if (childTable == null)
				return;

			// Check if we got an OK.
			if (new RelationsItemDialog(DialogModes.Edit, relationItem, parentTable, childTable).ShowDialog(this) ==
				DialogResult.OK)
			{
				// Populate the list view item with the new data.
				PopulateRelationListViewItem(relationItem, relationsListView.SelectedItems[0]);
			}
		}

		/// <summary>
		/// Returns the selected relation item or null if none has been selected.
		/// </summary>
		private RelationItem GetSelectedRelationItem()
		{
			if (relationsListView.SelectedItems.Count == 0)
				return null;

			return _relation.Relations[relationsListView.SelectedIndices[0]];
		}

		#endregion

		#region deleteRelationButton_Click

		/// <summary>
		/// Invoked each time the delete relation button is clicked.
		/// </summary>
		private void deleteRelationButton_Click(object sender, EventArgs e)
		{
			RelationItem relationItem = GetSelectedRelationItem();
			if (relationItem == null)
				return;

			// Asks the user if the table should be deleted.
			if (MessageBox.Show(Resources.DeleteRelationItemText, Resources.ProgramName, MessageBoxButtons.YesNo, MessageBoxIcon.Question) ==
				DialogResult.Yes)
			{
				// Removes the relationItem from the listview and the list.
				_relation.Relations.Remove(relationItem);
				relationsListView.Items.Remove(relationsListView.SelectedItems[0]);
			}
		}

		#endregion

		#region addWeakRelationButton_Click

		/// <summary>
		/// Invoked each time the add weak relations button is clicked.
		/// </summary>
		private void addWeakRelationButton_Click(object sender, EventArgs e)
		{
			RelationItem relationItem = new RelationItem();

			// Get the parent table.
			Table parentTable = (Table)weakObjectsComboBox.SelectedItem;
			if (parentTable == null)
				return;
			// Get the child table.
			Table childTable = (Table)childObjectsComboBox.SelectedItem;
			if (childTable == null)
				return;

			// Check if we got an OK.
			if (new RelationsItemDialog(DialogModes.New, relationItem, parentTable, childTable).ShowDialog(this) ==
				DialogResult.OK)
			{
				ListViewItem item = new ListViewItem();

				// Populate the list view item with the new data.
				PopulateRelationListViewItem(relationItem, item);
				weakRelationsListView.Items.Add(item);
				_relation.WeakRelations.Add(relationItem);
			}
		}

		#endregion

		#region editWeakRelationButton_Click

		/// <summary>
		/// Invoked each time the edit weak relations button is clicked.
		/// </summary>
		private void editWeakRelationButton_Click(object sender, EventArgs e)
		{
			RelationItem relationItem = GetSelectedWeakRelationItem();
			if (relationItem == null)
				return;

			// Get the parent table.
			Table parentTable = (Table)weakObjectsComboBox.SelectedItem;
			if (parentTable == null)
				return;
			// Get the child tble.
			Table childTable = (Table)childObjectsComboBox.SelectedItem;
			if (childTable == null)
				return;

			// Check if we got an OK.
			if (new RelationsItemDialog(DialogModes.Edit, relationItem, parentTable, childTable).ShowDialog(this) ==
				DialogResult.OK)
			{
				// Populate the list view item with the new data.
				PopulateRelationListViewItem(relationItem, weakRelationsListView.SelectedItems[0]);
			}
		}

		/// <summary>
		/// Returns the selected weak relation item or null if none has been selected.
		/// </summary>
		private RelationItem GetSelectedWeakRelationItem()
		{
			if (weakRelationsListView.SelectedItems.Count == 0)
				return null;

			return _relation.WeakRelations[weakRelationsListView.SelectedIndices[0]];
		}

		#endregion

		#region deleteWeakRelationButton_Click

		/// <summary>
		/// Invoked each time the delete weak realtion button is clicked.
		/// </summary>
		private void deleteWeakRelationButton_Click(object sender, EventArgs e)
		{
			RelationItem relationItem = GetSelectedWeakRelationItem();
			if (relationItem == null)
				return;

			// Asks the user if the table should be deleted.
			if (MessageBox.Show(Resources.DeleteRelationItemText, Resources.ProgramName, MessageBoxButtons.YesNo, MessageBoxIcon.Question) ==
				DialogResult.Yes)
			{
				// Removes the relationItem from the listview and the list.
				_relation.WeakRelations.Remove(relationItem);
				weakRelationsListView.Items.Remove(weakRelationsListView.SelectedItems[0]);
			}
		}

		#endregion

		/// <summary>
		/// Invoked each time the form is about to close.
		/// </summary>
		private void RelationDialog_FormClosing(object sender, FormClosingEventArgs e)
		{
			if (this.DialogResult != DialogResult.OK)
				return;

			e.Cancel = (!PopulateObject());
		}

		#region PopulateObject

		/// <summary>
		/// Populates the object with the data.
		/// </summary>
		private bool PopulateObject()
		{
			// Check if all required controls have been compiled.
			if (!CheckControls())
				return false;

            _relation.Name = nameTextBox.Text;
			_relation.ChildProperty = childPropertyTextBox.Text;
			_relation.ChildTable = (Table)childObjectsComboBox.SelectedItem;
			_relation.CreateInChildPersistent = createInChildCheckBox.Checked;
			_relation.CreateInParentPersistent = createInParentCheckBox.Checked;
			_relation.ParentProperty = parentPropertyTextBox.Text;
			_relation.ParentTable = (Table)parentObjectsComboBox.SelectedItem;

			RelationTypes relationType = GetSelectedRelationType();
			_relation.RelationType = relationType;

			if (relationType == RelationTypes.ManyToMany)
				_relation.WeakTable = (Table)weakObjectsComboBox.SelectedItem;
			else
				_relation.WeakTable = null;

			_relation.PersistBehaviour = PersistBehaviours.None;
			if (dontInsertCheckBox.Checked)
				_relation.PersistBehaviour |= PersistBehaviours.DontInsert;
			if (dontUpdateCheckBox.Checked)
				_relation.PersistBehaviour |= PersistBehaviours.DontUpdate;

			return true;
		}

		#endregion

		#region CheckControls

		/// <summary>
		/// Checks if all required controls have been compiled.
		/// </summary>
		private bool CheckControls()
		{
			// Check if the parent property has been compiled.
			if (createInParentCheckBox.Checked && parentPropertyTextBox.Text.Trim().Length == 0)
			{
				tabControl1.SelectedTab = generalTabPage;
				MessageBox.Show(Resources.MissingParentPropertyText, Resources.ProgramName, MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}
			// Check if the parent property has been compiled.
			if (createInChildCheckBox.Checked && childPropertyTextBox.Text.Trim().Length == 0)
			{
				tabControl1.SelectedTab = generalTabPage;
				MessageBox.Show(Resources.MissingChildPropertyText, Resources.ProgramName, MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}

			// A parent persistent object must be selected.
			if (parentObjectsComboBox.SelectedIndex == -1)
			{
				tabControl1.SelectedTab = generalTabPage;
				MessageBox.Show(Resources.MissingParentPersistentText, Resources.ProgramName, MessageBoxButtons.OK,
					MessageBoxIcon.Error);
				return false;
			}
			// A child persistent object must be selected.
			if (childObjectsComboBox.SelectedIndex == -1)
			{
				tabControl1.SelectedTab = generalTabPage;
				MessageBox.Show(Resources.MissingChildPersistentText, Resources.ProgramName, MessageBoxButtons.OK,
					MessageBoxIcon.Error);
				return false;
			}
			// A weak persistent object must be selected if the type of the relation is many-to-many.
			if (childObjectsComboBox.SelectedIndex == -1 && GetSelectedRelationType() == RelationTypes.ManyToMany)
			{
				tabControl1.SelectedTab = generalTabPage;
				MessageBox.Show(Resources.MissingWeakPersistentText, Resources.ProgramName, MessageBoxButtons.OK,
					MessageBoxIcon.Error);
				return false;
			}

			// We need at least one relation.
			if (relationsListView.Items.Count == 0)
			{
				tabControl1.SelectedTab = relationsTabPage;
				MessageBox.Show(Resources.MissingRelationText, Resources.ProgramName, MessageBoxButtons.OK,
					MessageBoxIcon.Error);
				return false;
			}
			// We need at least one relation.
			if (weakRelationsListView.Items.Count == 0 && GetSelectedRelationType() == RelationTypes.ManyToMany)
			{
				tabControl1.SelectedTab = weakRelationsTabPage;
				MessageBox.Show(Resources.MissingRelationText, Resources.ProgramName, MessageBoxButtons.OK,
					MessageBoxIcon.Error);
				return false;
			}

			return true;
		}

		#endregion

		/// <summary>
		/// Invoked each time the checkbox checked state changed.
		/// </summary>
		private void createInParentCheckBox_CheckedChanged(object sender, EventArgs e)
		{
			parentPropertyTextBox.Enabled = createInParentCheckBox.Checked;
		}

		/// <summary>
		/// Invoked each time the checkbox checked state changed.
		/// </summary>
		private void createInChildCheckBox_CheckedChanged(object sender, EventArgs e)
		{
			childPropertyTextBox.Enabled = createInChildCheckBox.Checked;
		}

		/// <summary>
		/// Invoked each time the selected index of the weak relations list view changed.
		/// </summary>
		private void weakRelationsListView_SelectedIndexChanged(object sender, EventArgs e)
		{
			UpdateControlsState();
		}
	}
}