#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Chili.Opf3.Wizard.Shared.Storages;
using Chili.Opf3.Wizard.Shared.UI;
using Chili.Opf3.Wizard.Properties;

#endregion

namespace Chili.Opf3.Wizard.UI.Forms
{
	/// <summary>
	/// This dialog allows to edit one relation item.
	/// </summary>
	public partial class RelationsItemDialog : BaseDialog
	{
		private RelationItem _relationItem;
		private Table _childTable;
		private Table _parentTable;

		/// <summary>
		/// Creates a new instance of the <see cref="RelationsItemDialog">RelationsItemDialog Class</see>.
		/// </summary>
		/// <param name="mode">The mode of the dialog.</param>
		/// <param name="relationItem">The item that is edited.</param>
		/// <param name="childTable">The child table.</param>
		/// <param name="parentTable">The parent table.</param>
		public RelationsItemDialog(DialogModes mode, RelationItem relationItem, Table parentTable, Table childTable)
		{
			InitializeComponent();

			if (relationItem == null)
				throw new ArgumentNullException("relationItem", "RelationItem is null.");
			if (parentTable == null)
				throw new ArgumentNullException("parentTable", "ParentTable is null.");
			if (childTable == null)
				throw new ArgumentNullException("childTable", "ChildTable is null.");

			_relationItem = relationItem;
			_parentTable = parentTable;
			_childTable = childTable;

			PopulatePropertyComboBox(relationItem.ParentField, parentTable, parentPropertyComboBox);
			PopulatePropertyComboBox(relationItem.ChildField, childTable, childPropertyComboBox);
			// Set the labels with the persistent names.
			parentTableLabel.Text = _parentTable.PersistentName;
			childTableLabel.Text = _childTable.PersistentName;

			if (mode == DialogModes.New)
				this.Text = Resources.NewRelationItemDialogTitle;
			else
			{
				this.Text = Resources.EditRelationItemDialogTitle;
				PopulateControls();
			}
		}

		/// <summary>
		/// Populates the given combobox with the fields of the given table.
		/// </summary>
		/// <param name="field">The current selected field.</param>
		/// <param name="table">The table that's fields are populated.</param>
		/// <param name="combobox">The combobox.</param>
		private void PopulatePropertyComboBox(Field field, Table table, ComboBox combobox)
		{
			// Clear all items in the combobox.
			combobox.Items.Clear();

			bool containsField = false;
			foreach (Field f in table.Fields)
			{
				if (field != null && f.PropertyName == field.PropertyName)
					containsField = true;

				combobox.Items.Add(f);
			}

			// Add the field if no longer in the list.
			if (field != null && !containsField)
				combobox.Items.Add(field);

			if (combobox.Items.Count > 0)
				combobox.SelectedIndex = 0;
		}

		/// <summary>
		/// Populates the UI controls.
		/// </summary>
		private void PopulateControls()
		{
			SelectPropertyInComboBox(_relationItem.ParentField, parentPropertyComboBox);
			SelectPropertyInComboBox(_relationItem.ChildField, childPropertyComboBox);
		}

		/// <summary>
		/// Selects the field in the combobox.
		/// </summary>
		/// <param name="field">The field that is selected.</param>
		/// <param name="comboBox">The combobox for the field.</param>
		private void SelectPropertyInComboBox(Field field, ComboBox comboBox)
		{
			for (int i = 0; i < comboBox.Items.Count; i++)
			{
				Field f = (Field)comboBox.Items[i];
				if (f.PropertyName == field.PropertyName)
				{
					comboBox.SelectedIndex = i;
					return;
				}
			}
		}

		/// <summary>
		/// Invoked each time the form is about to close.
		/// </summary>
		private void RelationsItemDialog_FormClosing(object sender, FormClosingEventArgs e)
		{
			if (this.DialogResult != DialogResult.OK)
				return;

			e.Cancel = (!PopulateObject());
		}

		/// <summary>
		/// Populates the object with data.
		/// </summary>
		private bool PopulateObject()
		{
			if (parentPropertyComboBox.SelectedIndex == -1 || childPropertyComboBox.SelectedIndex == -1)
			{
				MessageBox.Show(Resources.BothPropertiesNeedBeingCompiledText, Resources.ProgramName, MessageBoxButtons.OK,
					MessageBoxIcon.Error);
				return false;
			}

			_relationItem.ParentField = (Field)parentPropertyComboBox.SelectedItem;
			_relationItem.ChildField = (Field)childPropertyComboBox.SelectedItem;

			return true;
		}
	}
}