#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using Chili.Opf3.Wizard.Shared;
using Chili.Opf3.Wizard.Properties;
using Chili.Opf3.Wizard.Shared.Storages;
using System.Reflection;
using System.Diagnostics;

#endregion

namespace Chili.Opf3.Wizard.UI.Pages
{
	/// <summary>
	/// This class represents the page that generates the persistent objects (and the additional pages).
	/// </summary>
	public partial class GeneratePersistentPage : BasePage
	{
		private int _currentOK;
		private int _currentFailed;
		private int _count;
		private OutputLogger _logger;
		private readonly string _logFileName = Environment.CurrentDirectory + "\\opf3wiz.log";

		/// <summary>
		/// Creates a new instance of the <see cref="GeneratePersistentPage">GeneratePersistentPage Class</see>.
		/// </summary>
		public GeneratePersistentPage()
		{
			InitializeComponent();
		}

		/// <summary>
		/// Invoked each time the page is opened.
		/// </summary>
		public override void PageOpening()
		{
			_currentFailed = 0;
			_currentOK = 0;
			_count = 0;
			_logger = null;

			SetObjectsGeneratedLabel();

			// Set the title and the description of the generate persistent page header.
			SetHeaderTitle(Resources.GeneratePersistentPageHeaderTitle);
			SetHeaderDescription(Resources.GeneratePersistentPageHeaderDescription);

			// Get the wizard.
			Wizard wizard = (Wizard)this.Parent.Parent;
			// Disable all buttons.
			wizard.BackEnabled = false;
			wizard.NextEnabled = false;
			wizard.CancelEnabled = false;
			showLogButton.Enabled = false;

			// Empty list view.
			objectsListView.Items.Clear();
			objectsListView.Groups.Clear();

			// Populate the listview with the objects information.
			PopulateObjectsInListView();

			Application.DoEvents();

			// Generates the objects.
			GenerateObjects();

			// Enable the buttons.
			wizard.NextEnabled = true;
			wizard.CancelEnabled = true;
			wizard.BackEnabled = true;
			showLogButton.Enabled = true;
		}

		/// <summary>
		/// Populates the objects in the list view.
		/// </summary>
		private void PopulateObjectsInListView()
		{
			// Loop over all selected utility generators and add them to the list.
			foreach (IObjectsProvider provider in Session.Current.ObjectProvider)
			{
                List<IGeneratedObject> utilityList = provider.GetObjectsList();
				// Skip if there are no objects to generate.
				if (utilityList.Count == 0)
					continue;

				// Creates a new group and adds it to the list view.
				ListViewGroup group = new ListViewGroup(provider.Name);
				objectsListView.Groups.Add(group);

				foreach (IGeneratedObject obj in utilityList)
				{
					ListViewItem item = new ListViewItem(obj.Name, 0, group);
					item.Tag = obj.ID;
					item.SubItems.Add(Resources.GeneratingObjectPending);

					// Add the item to the listview.
					objectsListView.Items.Add(item);

					// Incement the count.
					_count++;
				}
			}
		}

		/// <summary>
		/// Generates the objects.
		/// </summary>
		private void GenerateObjects()
		{
			// Loop over all selected utility generators and add them to the list.
			foreach (IObjectsProvider provider in Session.Current.ObjectProvider)
			{
				// Register events.
				provider.GeneratedObject += new EventHandler<GeneratedObjectEventArgs>(Generator_GeneratedObject);
				provider.GeneratingObject += new EventHandler<GeneratingObjectEventArgs>(Generator_GeneratingObject);

				// Generate the utility objects.
				provider.GenerateObjects(Session.Current.Project);

				// After having generated the objects unregister the events.
				provider.GeneratedObject -= new EventHandler<GeneratedObjectEventArgs>(Generator_GeneratedObject);
				provider.GeneratingObject -= new EventHandler<GeneratingObjectEventArgs>(Generator_GeneratingObject);
			}
		}

		/// <summary>
		/// Invoked before the object is generated.
		/// </summary>
		private void Generator_GeneratingObject(object sender, GeneratingObjectEventArgs e)
		{
			ListViewItem item = GetListViewItemByTag(e.ObjectID);
			if (item == null)
				return;

			// Set the index to creating.
			item.ImageIndex = 1;
			item.SubItems[1].Text = Resources.GeneratingObjectInProgress;

			// Relax a second to update the UI.
			Application.DoEvents();
		}

		/// <summary>
		/// Returns the listview item with the given name.
		/// </summary>
		/// <param name="tag">The tag of the item.</param>
		private ListViewItem GetListViewItemByTag(string tag)
		{
			foreach (ListViewItem item in objectsListView.Items)
			{
				string t = item.Tag as string;
				if (t != null && t == tag)
					return item;
			}
			return null;
		}

		/// <summary>
		/// Invoked after the object has been generated.
		/// </summary>
		private void Generator_GeneratedObject(object sender, GeneratedObjectEventArgs e)
		{
			// Create an instance of the logger.
			if (_logger == null)
				_logger = new OutputLogger(_logFileName);

			ListViewItem item = GetListViewItemByTag(e.ObjectID);
			if (item == null)
				return;

			if (e.Success)
			{
				item.ImageIndex = 2;
				item.SubItems[1].Text = Resources.GeneratingObjectSuccessful;

				// Write in the log that the file has been generated without problems.
				_logger.AddToLog(string.Format(Resources.LogGeneratedObjectSucessfully, item.SubItems[0].Text));

				_currentOK++;
			}
			else
			{
				item.ImageIndex = 3;
				item.SubItems[1].Text = string.Format(Resources.GeneratingObjectError, e.Exception.Message);

				// Write to the log that we had a problem while generating the object.
				_logger.AddToLog(string.Format(Resources.LogGeneratedObjectProblem, item.SubItems[0].Text), e.Exception);

				_currentFailed++;
			}

			SetObjectsGeneratedLabel();

			// Relax a second to update the UI.
			Application.DoEvents();
		}

		/// <summary>
		/// Sets the persistent generated label.
		/// </summary>
		private void SetObjectsGeneratedLabel()
		{
			// Label of the status.
			generatingObjectsOKLabel.Text = string.Format(Resources.GeneratedObjectsOKLabel, _currentOK, _count);
			generatingObjectsFailedLabel.Text = string.Format(Resources.GeneratedObjectsFailedLabel, _currentFailed, _count);
		}

		/// <summary>
		/// Invoked each time the button to show the log is clicked.
		/// </summary>
		private void showLogButton_Click(object sender, EventArgs e)
		{
			Process.Start(_logFileName);
		}
	}
}
