#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Xml;
using System.Windows.Forms;
using Chili.Opf3.Wizard.Properties;
using Chili.Opf3.Wizard.Shared.Storages;
using Chili.Opf3.Wizard.Shared;
using Chili.Opf3.Wizard.Shared.UI;
using Chili.Opf3.Wizard.UI.Forms;

#endregion

namespace Chili.Opf3.Wizard.UI.Pages
{
	/// <summary>
	/// This class represents the relations page of the wizard.
	/// </summary>
	public partial class RelationsPage : BasePage
	{
		private List<Relation> _relations;

		/// <summary>
		/// Creates a new instance of the <see cref="RelationsPage">RelationsPage Class</see>.
		/// </summary>
		public RelationsPage()
		{
			InitializeComponent();
		}

		/// <summary>
		/// Invoked each time the page is opened.
		/// </summary>
		public override void PageOpening()
		{
			// Set the title and the description of the welcome page header.
			SetHeaderTitle(Resources.RelationsPageHeaderTitle);
			SetHeaderDescription(Resources.RelationsPageHeaderDescription);

			// Store the relations in a temporary list.
			_relations = Session.Current.Project.Relations;

			PopulateRelationsListView();
		}

		/// <summary>
		/// Populates the relations list view with data.
		/// </summary>
		private void PopulateRelationsListView()
		{
			relationsListView.Items.Clear();

			if (_relations == null)
				return;

			// Suspend layout to add the items.
			relationsListView.SuspendLayout();
			try
			{
				// Loop over all relations and add them to the list view.
				foreach (Relation relation in _relations)
				{
					ListViewItem item = new ListViewItem();
					PopulateListViewItem(relation, item);

					// Add the item to the listview.
					relationsListView.Items.Add(item);
				}
			}
			finally
			{
				// Resume the layout after having added all items.
				relationsListView.ResumeLayout();
			}

			UpdateControlsState();
		}

		/// <summary>
		/// Updates a listview item.
		/// </summary>
		/// <param name="relation">The relation used to update the listview item.</param>
		/// <param name="item">The item itself</param>
		private void PopulateListViewItem(Relation relation, ListViewItem item)
		{
			// Remove all subitems.
			item.SubItems.Clear();
            item.Tag = relation;

            item.Text = relation.ParentTable.PersistentName;
			item.ImageIndex = 0;
			item.Checked = relation.Generate;
			item.SubItems.Add(CreateMember(relation, true));
			item.SubItems.Add(relation.ChildTable.PersistentName);
			item.SubItems.Add(CreateMember(relation, false));
            item.SubItems.Add(relation.Name);

			// Add the weak entity if required.
			if (relation.WeakTable != null)
				item.SubItems.Add(relation.WeakTable.PersistentName);
		}

		/// <summary>
		/// Creates the parent member of the relations.
		/// </summary>
		/// <param name="parent">True to create the parent member.</param>
		/// <param name="relation">The relation that's member are created.</param>
		private string CreateMember(Relation relation, bool parent)
		{
			string result = string.Empty; ;

			foreach (RelationItem item in relation.Relations)
			{
				if (result.Length > 0)
					result += ";";

				if (parent)
					result += item.ParentField.PropertyName;
				else
					result += item.ChildField.PropertyName;
			}

			return result;
		}

		/// <summary>
		/// Invoked each time the page is closed.
		/// </summary>
		/// <param name="next">True if the next button has been clicked. False on all other.</param>
		public override bool PageClosing(bool next)
		{
			// Set the relations in the project.
			Project project = Session.Current.Project;
			project.Relations = _relations;

			// Loop over all relations and set the checked state.
            foreach (ListViewItem item in relationsListView.Items)
            {
                ((Relation)item.Tag).Generate = item.Checked;
            }

			// Allow the provider to build internal data from Storage.
			foreach (IObjectsProvider provider in Session.Current.ObjectProvider)
			{
                // apply storages changes (if any).
                provider.ProcessStorage(project);
			}
			return true;
		}


		/// <summary>
		/// Invoked each time the selected index changed.
		/// </summary>
		private void relationsListView_SelectedIndexChanged(object sender, EventArgs e)
		{
			UpdateControlsState();
		}

		/// <summary>
		/// Updates the state of the controls.
		/// </summary>
		private void UpdateControlsState()
		{
			editButton.Enabled = (relationsListView.SelectedItems.Count > 0);
			deleteButton.Enabled = (relationsListView.SelectedItems.Count > 0);
		}

		/// <summary>
		/// Returns the selected relation or null if none has been selected.
		/// </summary>
		private Relation GetSelectedRelation()
		{
			if (relationsListView.SelectedItems.Count == 0)
				return null;
			return _relations[relationsListView.SelectedIndices[0]];
		}

		/// <summary>
		/// Invoked each time the button to unselect (uncheck) all tables is clicked.
		/// </summary>
		private void unCheckAllButton_Click(object sender, EventArgs e)
		{
			// Loop over all items and uncheck.
			for (int i = 0; i < _relations.Count; i++)
			{
				relationsListView.Items[i].Checked = false;
			}
		}

		/// <summary>
		/// Invoked each time the button to select (check) all tables is clicked.
		/// </summary>
		private void checkAllButton_Click(object sender, EventArgs e)
		{
			// Loop over all items and uncheck.
			for (int i = 0; i < _relations.Count; i++)
			{
				relationsListView.Items[i].Checked = true;
			}
		}

		/// <summary>
		/// Invoked each time the edit relation button is clicked.
		/// </summary>
		private void editButton_Click(object sender, EventArgs e)
		{
			Relation relation = GetSelectedRelation();
			if (relation == null)
				return;

			// Check if we got an OK.
			if (new RelationDialog(DialogModes.Edit, relation).ShowDialog(this) == DialogResult.OK)
			{
				// Update the list view item.
				PopulateListViewItem(relation, relationsListView.SelectedItems[0]);
			}
		}


		/// <summary>
		/// Invoked each time the add button is clicked.
		/// </summary>
		private void addButton_Click(object sender, EventArgs e)
		{
			Relation relation = new Relation();
			relation.Generate = true;

			// Check if we got an OK.
			if (new RelationDialog(DialogModes.New, relation).ShowDialog(this) == DialogResult.OK)
			{
				// Create a new listview item and add it to the list view.
				ListViewItem item = new ListViewItem();
				PopulateListViewItem(relation, item);
				relationsListView.Items.Add(item);

				// Add the relation to the list of relations.
				_relations.Add(relation);
			}
		}

		/// <summary>
		/// Invoked each time the remove button is clicked.
		/// </summary>
		private void deleteButton_Click(object sender, EventArgs e)
		{
			Relation relation = GetSelectedRelation();
			if (relation == null)
				return;

			// Asks the user if the table should be deleted.
			if (MessageBox.Show(Resources.DeleteRelationText, Resources.ProgramName, MessageBoxButtons.YesNo, MessageBoxIcon.Question) ==
				DialogResult.Yes)
			{
				// Removes the relation from the listview and the list.
				_relations.Remove(relation);
				relationsListView.Items.Remove(relationsListView.SelectedItems[0]);
			}
		}
	}
}
