#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using Chili.Opf3.Wizard.Properties;
using Chili.Opf3.Wizard.Shared.Storages;
using Chili.Opf3.Wizard.Shared;
using Chili.Opf3.Wizard.UI.Forms;

#endregion

namespace Chili.Opf3.Wizard.UI.Pages
{
	/// <summary>
	/// This class represents the page that allows to specify the output of the storage schema of the wizard.
	/// </summary>
	public partial class StorageSchemaDataPage : BasePage
	{
		private StoragesManager _manager;

		/// <summary>
		/// Creates a new instance of the <see cref="StorageSchemaDataPage">StorageSchemaDataPage Class</see>.
		/// </summary>
		public StorageSchemaDataPage()
		{
			InitializeComponent();
		}

		/// <summary>
		/// Invoked each time the page is opened.
		/// </summary>
		public override void PageOpening()
		{
			// Set the title and the description page header.
			SetHeaderTitle(Resources.StorageSchemeDataPageTitle);
			SetHeaderDescription(Resources.StorageSchemaDataPageDescription);

			// Load the storages if not loaded earlier.
			if (_manager == null)
			{
				Project project = Session.Current.Project;
				// Setting the checkbox is put here, since it should only be set with the first opening of the page.
				toSingularCheckBox.Checked = project.RenameItemsToSingular;
                removefirstNumericUpDown.Value = project.RemoveFirstChars;

				_manager = new StoragesManager();
				_manager.Load();

				if (project.Storage != null)
				{
					// Check if the manager contains the storage of the project. If not, add the storge of the project.
					if (!_manager.Exists(delegate(Storage obj) { return obj.ConnectionString == project.Storage.ConnectionString; }))
						_manager.Add(new Storage(project.Storage.Alias, project.Storage.ConnectionString));
				}

				DatabindStorages();

				if (project.Storage != null)
					SelectStorageInComboBox(project.Storage);
			}
		}

		/// <summary>
		/// Invoked each time the page is closed.
		/// </summary>
		/// <param name="next">True if the next button has been clicked. False on all other.</param>
		public override bool PageClosing(bool next)
		{
			if (!next)
				return true;

			// Check if a storage has been selected.
			if (storagesComboBox.SelectedIndex == -1)
			{
				MessageBox.Show(Resources.MissingStorageText, Resources.ProgramName, MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}

			Session session = Session.Current;
			// Set the properties of the project.
			Project project = session.Project;
			project.RenameItemsToSingular = toSingularCheckBox.Checked;
            project.RemoveFirstChars = (int)removefirstNumericUpDown.Value;
			project.Storage = (Storage)storagesComboBox.SelectedItem;

			return true;
		}

		/// <summary>
		/// Bind the list of databases to the combobox.
		/// </summary>
		private void DatabindStorages()
		{
			Project project = Session.Current.Project;

			storagesComboBox.Items.Clear();

			// Loop over all items and add them to the combobox.
			// Done this way, because with simply setting the datasource there was a problem
			// during delete of a storage.
			for(int i = 0; i < _manager.Count; i++)
			{
				Storage storage = _manager[i];
				storagesComboBox.Items.Add(storage);

				if (project.Storage != null)
				{
					// Check if the list contains already the storage.
					if (storage.ConnectionString == project.Storage.ConnectionString)
						storagesComboBox.SelectedIndex = i;
				}
			}

			if (storagesComboBox.SelectedIndex == -1 && storagesComboBox.Items.Count > 0)
				storagesComboBox.SelectedIndex = 0;

			UpdateControls();
		}

		/// <summary>
		/// Updates the controls on the form.
		/// </summary>
		private void UpdateControls()
		{
			bool storageSelected = (storagesComboBox.SelectedIndex > -1);

			removeStorageButton.Enabled = storageSelected;
			metaDatabutton.Enabled = storageSelected;
			editStorageButton.Enabled = storageSelected;
		}

		/// <summary>
		/// Invoked each time the button to add a storage is clicked.
		/// </summary>
		private void addStorageButton_Click(object sender, EventArgs e)
		{
			StorageDialog storageDialog = new StorageDialog(null);
			if (storageDialog.ShowDialog(this) == DialogResult.OK)
			{
				_manager.Add(storageDialog.Storage);
				// Save the changes.
				_manager.Save();

				DatabindStorages();
				// Selects the given storage.
				SelectStorageInComboBox(storageDialog.Storage);
			}
		}

		/// <summary>
		/// Selects the given storage in the combobox.
		/// </summary>
		/// <param name="storage"></param>
		private void SelectStorageInComboBox(Storage storage)
		{
			for(int i = 0; i < storagesComboBox.Items.Count; i++)
			{
				Storage s = (Storage)storagesComboBox.Items[i];

				if (s.ConnectionString == storage.ConnectionString)
				{
					storagesComboBox.SelectedIndex = i;
					return;
				}
			}
		}

		/// <summary>
		/// Invoked each time the button to remove a storage is clicked.
		/// </summary>
		private void removeStorageButton_Click(object sender, EventArgs e)
		{
			Storage storage = (Storage)storagesComboBox.SelectedItem;
			if (storage == null)
				return;

			if (MessageBox.Show(this, Resources.RemoveStorageText, Resources.ProgramName, MessageBoxButtons.YesNo, MessageBoxIcon.Question)
				== DialogResult.Yes)
			{
				// Remove the storage and save.
				_manager.Remove(storage);
				_manager.Save();

				DatabindStorages();
			}
		}

		/// <summary>
		/// Invoked each time the meta data button is clicked.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
        private void metaDatabutton_Click(object sender, EventArgs e)
        {
            MetadataViewer viewer = new MetadataViewer((Storage)storagesComboBox.SelectedItem);
            viewer.ShowDialog(this);
        }

		/// <summary>
		/// Invoked each time the edit storage button is clicked.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void editStorageButton_Click(object sender, EventArgs e)
		{
			if (storagesComboBox.SelectedIndex == -1)
				return;

			Storage storage = (Storage)storagesComboBox.SelectedItem;

			StorageDialog storageDialog = new StorageDialog(storage);
			if (storageDialog.ShowDialog(this) == DialogResult.OK)
			{
				// Save the changes.
				_manager.Save();

				DatabindStorages();
			}
		}
	}
}
