#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using Chili.Opf3.Wizard.Properties;
using Chili.Opf3.Wizard.Shared.Storages;
using Chili.Opf3.Wizard.Shared;
using Chili.Opf3.Wizard.Shared.UI;
using System.Data.OleDb;
using Chili.Opf3.Wizard.UI.Forms;

#endregion

namespace Chili.Opf3.Wizard.UI.Pages
{
	/// <summary>
	/// This class represents the tables page of the wizard.
	/// </summary>
	public partial class TablesPage : BasePage
	{
		private OleDbConnection _connection;
		private bool _retrievedSchema;
		private bool _renameEntitiesToSingular;
		private SchemaDifferencer _schemaDiff = new SchemaDifferencer();

		/// <summary>
		/// Creates a new instance of the <see cref="TablesPage">TablesPage Class</see>.
		/// </summary>
		public TablesPage()
		{
			InitializeComponent();
		}

		/// <summary>
		/// Invoked each time the page is opened.
		/// </summary>
		public override void PageOpening()
		{
			Cursor.Current = Cursors.WaitCursor;
			fetchedLabel.Text = string.Empty;

			// Set the title and the description of the welcome page header.
			SetHeaderTitle(Resources.TablesPageHeaderTitle);
			SetHeaderDescription(Resources.TablesPageHeaderDescription);

			// Get the current selected storage and the connection to that storage.
			Project project = Session.Current.Project;
			OleDbConnection connection = new OleDbConnection(project.Storage.ConnectionString);

			// If the connection object is different then the currently stored: refesh!
			if (_renameEntitiesToSingular != project.RenameItemsToSingular || _connection == null || 
				connection.ConnectionString != _connection.ConnectionString || !_retrievedSchema)
			{
				// Hide the panel showing the tables.
				tablesPanel.Visible = false;
				// Disable buttons.
				SetButtonState(false);

				StorageSchema schema = GetStorageSchema(connection);
				if (schema != null)
				{
					// Get the schema from the storage and synchronizes the differences with the current project.
					_schemaDiff.Synchronize(schema, project.LastRetrievedStorageSchema, project);
				}

				// Set the connection of the project as the current one.
				_connection = connection;
				// Populates the tables list view with data.
				PopulateTablesListView();
				// Set the rename to singular of the current page to the one of the project.
				_renameEntitiesToSingular = project.RenameItemsToSingular;

				// Enable them again.
				SetButtonState(true);
			}

			// Show the panel for the tables.
			tablesPanel.Visible = true;
			
			Cursor.Current = Cursors.Default;
		}

		/// <summary>
		/// Sets the state of the buttons.
		/// </summary>
		/// <param name="enabled">True to enable them.</param>
		private void SetButtonState(bool enabled)
		{
			Wizard wizard = (Wizard)this.Parent.Parent;
			
			if (_retrievedSchema)
				wizard.NextEnabled = enabled;
			else
				wizard.NextEnabled = false;

			wizard.BackEnabled = enabled;
			wizard.CancelEnabled = enabled;

			// Relax a second to update the UI.
			Application.DoEvents();
		}

		/// <summary>
		/// Get the schema from the storage
		/// </summary>
		/// <param name="connection">Connection to get the schema.</param>
		private StorageSchema GetStorageSchema(OleDbConnection connection)
		{
            StorageSchemaReader schemaReader = new StorageSchemaReader(connection, Session.Current.Project);
			// Register the events.
			schemaReader.FetchedRelation += new EventHandler<FetchedRelationEventArgs>(schemaReader_FetchedRelation);
			schemaReader.FetchedTable += new EventHandler<FetchedTableEventArgs>(schemaReader_FetchedTable);

			_retrievedSchema = false;
			try
			{
				// Get the storage schema.
				StorageSchema schema = schemaReader.GetSchema();
				
				// Enable the add button.
				addButton.Enabled = true;
				_retrievedSchema = true;

				return schema;
			}
			catch (Exception ex)
			{
				// Show a messagebox with the given error if we were not able to get the storage schema.
				MessageBox.Show(string.Format(Resources.CouldNotFetchStorageSchema, ex.Message, ex.StackTrace).Replace("\\r\\n", "\r\n"), Resources.ProgramName, MessageBoxButtons.OK, MessageBoxIcon.Error);
				// Disable th add button.
				addButton.Enabled = false;
				return null;
			}
		}

		/// <summary>
		/// Invoked each time a table has been fetched.
		/// </summary>
		private void schemaReader_FetchedTable(object sender, FetchedTableEventArgs e)
		{
			fetchedLabel.Text = string.Format(Resources.TableInSchemaFound, e.Table.TableName);
			Application.DoEvents();
		}

		/// <summary>
		/// Invoked each time a relation has been fetched.
		/// </summary>
		private void schemaReader_FetchedRelation(object sender, FetchedRelationEventArgs e)
		{
			fetchedLabel.Text = string.Format(Resources.RelationInSchemaFound, e.Relation.ParentTable.TableName, e.Relation.ChildTable.TableName);
			Application.DoEvents();
		}

		/// <summary>
		/// Populates the tables listview with data.
		/// </summary>
		private void PopulateTablesListView()
		{
			tablesListView.Items.Clear();

			// Create the groups of the list view.
			ListViewGroup tablesGroup = new ListViewGroup(Resources.TablesGroupName);
			ListViewGroup viewsGroups = new ListViewGroup(Resources.ViewsGroupName);
			// Add the groups to the list view.
			tablesListView.Groups.Add(tablesGroup);
			tablesListView.Groups.Add(viewsGroups);

			Project project = Session.Current.Project;

			if (!_retrievedSchema)
				return;

			// Suspend layout to add the items.
			tablesListView.SuspendLayout();
			try
			{
				// Loop over all tables and add them to the list view.
				foreach (Table table in project.Tables)
				{
					ListViewGroup group = (table.IsView ? viewsGroups : tablesGroup);
					
					// Add the item to the listview.
					tablesListView.Items.Add(CreateListItem(table, group));
				}
			}
			finally
			{
				// Resume the layout after having added all items.
				tablesListView.ResumeLayout();
			}
		}

		/// <summary>
		/// Create a list item from the given table.
		/// </summary>
		/// <param name="table">The table used to create the list item.</param>
		/// <param name="group">Specifies the associated group.</param>
		private static ListViewItem CreateListItem(Table table, ListViewGroup group)
		{
			ListViewItem item = new ListViewItem(table.TableName, group);
            item.ImageIndex = table.IsView ? 1 : 0;
			item.SubItems.Add(table.PersistentName);
            item.SubItems.Add(table.BaseClass);
			item.Checked = table.Generate;
			item.Tag = table;
			return item;
		}

		/// <summary>
		/// Invoked each time the page is closed.
		/// </summary>
		/// <param name="next">True if the next button has been clicked. False on all other.</param>
		public override bool PageClosing(bool next)
		{
			// Set the relations and the tables in the project.
			if (_retrievedSchema)
			{
				// Loop over all tables and set the checked state.
				foreach(ListViewItem item in tablesListView.Items)
				{
					((Table)item.Tag).Generate = item.Checked;
				}
			}

			return true;
		}

		/// <summary>
		/// Invoked each time the selected index changed.
		/// </summary>
		private void tablesListView_SelectedIndexChanged(object sender, EventArgs e)
		{
			UpdateControlsState();
		}

		/// <summary>
		/// Updates the state of the controls.
		/// </summary>
		private void UpdateControlsState()
		{
			editButton.Enabled = (tablesListView.SelectedItems.Count > 0);
			deleteButton.Enabled = (tablesListView.SelectedItems.Count > 0);
		}

		/// <summary>
		/// Returns the selected table or null if none has been selected.
		/// </summary>
		private Table GetSelectedTable()
		{
			if (tablesListView.SelectedItems.Count == 0)
				return null;
			return tablesListView.SelectedItems[0].Tag as Table;
		}

		/// <summary>
		/// Invoked each time the button to unselect (uncheck) all tables is clicked.
		/// </summary>
		private void unCheckAllButton_Click(object sender, EventArgs e)
		{
			// Loop over all items and uncheck.
			foreach(ListViewItem item in tablesListView.Items)
			{
				item.Checked = false;
			}
		}

		/// <summary>
		/// Invoked each time the button to select (check) all tables is clicked.
		/// </summary>
		private void checkAllButton_Click(object sender, EventArgs e)
		{
			// Loop over all items and uncheck.
			foreach (ListViewItem item in tablesListView.Items)
			{
				item.Checked = true;
			}
		}

		/// <summary>
		/// Invoked each time the add persistent button is clicked.
		/// </summary>
		private void addButton_Click(object sender, EventArgs e)
		{
			// Create a new table.
			Table table = new Table();
			table.Generate = true;
			table.Rights = PersistentRights.All;

			Project p = Session.Current.Project;

			// Check if we got an OK.
			if (new PersistentDialog(DialogModes.New, table, p.Relations).ShowDialog(this) == DialogResult.OK)
			{
				// Add the table to the list of tables.
				p.Tables.Add(table);
				tablesListView.Items.Add(CreateListItem(table, null));
			}
		}

		/// <summary>
		/// Invoked each time the edit persistent button is clicked.
		/// </summary>
		private void editButton_Click(object sender, EventArgs e)
		{
			Table table = GetSelectedTable();
			if (table == null)
				return;

			Project p = Session.Current.Project;

			// Check if we got an OK.
            if (new PersistentDialog(DialogModes.Edit, table, p.Relations).ShowDialog(this) == DialogResult.OK)
			{
				// Update the current selected item in the list.
				ListViewItem item = tablesListView.SelectedItems[0];
				item.Text = table.TableName;
				item.SubItems[1].Text = table.PersistentName;
                item.SubItems[2].Text = table.BaseClass;
			}
		}

		/// <summary>
		/// Invoked each time the delete persistent button is clicked.
		/// </summary>
		private void deleteButton_Click(object sender, EventArgs e)
		{
			Table table = GetSelectedTable();
			if (table == null)
				return;

			// Asks the user if the table should be deleted.
			if (MessageBox.Show(Resources.DeletePersistentText, Resources.ProgramName, MessageBoxButtons.YesNo, MessageBoxIcon.Question) == 
				DialogResult.Yes)
			{
				Project p = Session.Current.Project;

				// Removes the table from the listview and the list.
				p.Tables.Remove(table);
				tablesListView.Items.Remove(tablesListView.SelectedItems[0]);
			}
		}
	}
}
