#region Developer's comment header
/* 
 * copyright:  2005-2006 Chili Software. All rights reserved.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using Chili.Opf3.Wizard.Shared;
using Chili.Opf3.Wizard.Shared.UI;
using Chili.Opf3.Wizard.Properties;
using Chili.Opf3.Wizard.Shared.Storages;
using Chili.Opf3.Wizard.UI.Forms;

#endregion

namespace Chili.Opf3.Wizard.UI.Pages
{
	/// <summary>
	/// This class allows to manage the typeconversion list. 
	/// </summary>
    public partial class TypesConversionPage : BasePage
	{

		/// <summary>
        /// Creates a new instance of the <see cref="InterfacePage">TypesConversionPage Class</see>.
		/// </summary>
        public TypesConversionPage()
		{
			InitializeComponent();
		}

        /// <summary>
        /// Invoked each time the page is opened.
        /// </summary>
        public override void PageOpening()
        {
            SetHeaderTitle(Resources.TypeConversionPageHeaderTitle);
            SetHeaderDescription(Resources.TypeConversionPageHeaderDescription);
            instructionsLabel.Text = Resources.TypeConversionPageHeaderInstructions;


            // Bind the  list.
            BindListToListView();
        }

        /// <summary>
        /// Binds the searcher list to the listview.
        /// </summary>
        private void BindListToListView()
        {
            Project project = Session.Current.Project;

            // Clear the list.
            typeconversionListView.Items.Clear();
            // Suspend the layout.
            typeconversionListView.SuspendLayout();

            try
            {
                foreach (TypeConversionPair i in project.TypeConverter.TypeConversionPairs)
                {
                    ListViewItem item = new ListViewItem();
                    // Populate the list view item.
                    PopulateListViewItem(i, item);
                    // Add the item to the listview.
                    typeconversionListView.Items.Add(item);
                }
            }
            finally
            {
                // Resume the layout.
                typeconversionListView.ResumeLayout();
            }

            UpdateControlsState();
        }

        /// <summary>
        /// Populate one list view item with data.
        /// </summary>
        /// <param name="s">The searcher itself. This searcher is used to populate the item.</param>
        /// <param name="item">The listview item that is populated.</param>
        private void PopulateListViewItem(TypeConversionPair s, ListViewItem item)
        {
            // Clear all subitems.
            item.SubItems.Clear();
            item.Tag = s;

            item.Text = s.BaseType;
            item.SubItems.Add(s.TargetType);
            item.SubItems.Add(s.Language);
            item.SubItems.Add(s.NameCondition);
            item.SubItems.Add(s.Initializer);
        }

		/// Invoked each time the page is closed.
		/// </summary>
		/// <param name="next">True if the next button has been clicked. False on all other.</param>
		public override bool PageClosing(bool next)
		{

			return true;
		}

        /// <summary>
        /// Returns the selected typeconversion or null if none is selected.
        /// </summary>
        private TypeConversionPair GetSelected()
        {
            if (typeconversionListView.SelectedItems.Count == 0)
                return null;

            return typeconversionListView.SelectedItems[0].Tag as TypeConversionPair;
        }

        /// <summary>
        /// Invoked each time the add button is clicked.
        /// </summary>
        private void addButton_Click(object sender, EventArgs e)
        {
            Project project = Session.Current.Project;
            TypeConversionPair s = new TypeConversionPair();

            // Check if we got an OK.
            if (new TypeConversionDialog(DialogModes.New, s).ShowDialog(this) == DialogResult.OK)
            {
                // Create a new listview item and add it to the list view.
                ListViewItem item = new ListViewItem();
                PopulateListViewItem(s, item);
                typeconversionListView.Items.Add(item);

                // Add the relation to the list of relations.
                project.TypeConverter.TypeConversionPairs.Add(s);
            }

        }

        /// <summary>
        /// Invoked each time the edit button is clicked.
        /// </summary>
        private void editButton_Click(object sender, EventArgs e)
        {
            TypeConversionPair s = GetSelected();
            if (s == null)
                return;

            // Check if we got an OK.
            if (new TypeConversionDialog(DialogModes.Edit, s).ShowDialog(this) == DialogResult.OK)
            {
                // Update the current selected item in the list.
                ListViewItem item = typeconversionListView.SelectedItems[0];
                PopulateListViewItem(s, item);
            }
        }

        /// <summary>
		/// Invoked each time the copy button is clicked.
		/// </summary>   
        private void copyButton_Click(object sender, EventArgs e)
        {
            Project project = Session.Current.Project;
            TypeConversionPair o = GetSelected();
            if (o == null)
                return;

            TypeConversionPair s = new TypeConversionPair();
            o.CopyTo(s); 
 
            // Check if we got an OK.
            if (new TypeConversionDialog(DialogModes.New, s).ShowDialog(this) == DialogResult.OK)
            {
                // Create a new listview item and add it to the list view.
                ListViewItem item = new ListViewItem();
                PopulateListViewItem(s, item);
                typeconversionListView.Items.Add(item);

                // Add the relation to the list of relations.
                project.TypeConverter.TypeConversionPairs.Add(s);
            }
        }

        // <summary>
        /// Invoked each time the remove button is clicked.
        /// </summary>
        private void deleteButton_Click(object sender, EventArgs e)
        {
            TypeConversionPair s = GetSelected();
            Project project = Session.Current.Project;
            if (s == null)
                return;

            // Asks the user if the table should be deleted.
            if (MessageBox.Show(Resources.DeleteTypeConversionText, Resources.ProgramName, MessageBoxButtons.YesNo, MessageBoxIcon.Question) ==
                DialogResult.Yes)
            {
                // Removes the relation from the listview and the list.
                project.TypeConverter.TypeConversionPairs.Remove(s);
                typeconversionListView.Items.Remove(typeconversionListView.SelectedItems[0]);
            }

        }

        /// <summary>
        /// Updates the state of the controls.
        /// </summary>
        private void UpdateControlsState()
        {
            editButton.Enabled = (typeconversionListView.SelectedItems.Count > 0);
            copyButton.Enabled = (typeconversionListView.SelectedItems.Count > 0);
            deleteButton.Enabled = (typeconversionListView.SelectedItems.Count > 0);
        }


        /// <summary>
        /// Invoked each time the selected index changed.
        /// </summary>
        private void typeconversionListView_SelectedIndexChanged(object sender, EventArgs e)
        {
            UpdateControlsState();
        }
	}
}
