#region Developer's comment header
/* EditBalloon.cs
 * 
 * author:    Christian Liensberger
 * copyright:  2005 Chili Software. All rights reserved.
 * 
 * history:   06.08.2005, clien: Created.
 *
 */
#endregion
#region Using directives

using System;
using System.Windows.Forms;
using System.Runtime.InteropServices;

#endregion

namespace Chili.Windows.Forms
{
	/// <summary>
	/// IMPORTANT: Call Application.EnableVisualStyles() BEFORE you use the balloon class! It won't work
	/// without that call!
	/// A sample class to manipulate ballon tooltips. Windows XP balloon-tips if used properly can 
	/// be very helpful. This class creates a balloon tooltip in the form of a message.
	/// This becomes useful for showing important information quickly to the user.
	/// Ever so often we need to avoid certain invalid characters in input textboxes. Though we do a 
	/// good job of filtering the characters, we dont necessarily do a good job of informing the user.
	/// This class helps immensely under such scenarios. This also helps in a shorter learning cycle of the 
	/// application.
	/// NOTE: The difference between this and other balloon classes is that this is made specifically 
	/// for the edit control and displays the balloon at the caret position within the textbox.
	/// </summary>
	public class EditBalloon 
	{
		Control _Parent = null;
		String _Text = "", _Title = "";
		ToolTipIcon _TitleIcon = ToolTipIcon.None;

		const Int32 ECM_FIRST = 0x1500, EM_SHOWBALLOONTIP = ECM_FIRST + 3;

		[DllImport("User32", SetLastError=true)]
		private static extern int SendMessage(IntPtr hWnd, Int32 Msg, Int32 wParam, IntPtr lParam);

		[StructLayout(LayoutKind.Sequential, CharSet=CharSet.Unicode)]
		private struct EDITBALLOONTIP
		{
			public Int32 cbStruct;
			public String pszTitle;
			public String pszText;
			public Int32 ttiIcon;
		}

		/// <summary>
		/// Empty constructor.
		/// </summary>
		public EditBalloon() {}

		/// <summary>
		/// Constructor that allows to set the title, text, icon and parent of the balloon.
		/// </summary>
		/// <param name="Title">Title of the balloon.</param>
		/// <param name="Text">Text shown in the balloon.</param>
		/// <param name="Icon">Icon of the balloon.</param>
		/// <param name="Parent">Parent control of the balloon.</param>
		public EditBalloon(String Title, String Text, ToolTipIcon Icon, Control Parent)
		{
			_Parent = Parent;
			_Text = Text;
			_Title = Title;
			_TitleIcon = Icon;
		}

		/// <summary>
		/// Static function to show the balloon. Set the title, text, icon and a parent control,
		/// where to show the balloon. 
		/// </summary>
		/// <param name="Title">Title of the balloon.</param>
		/// <param name="Text">Text of the balloon.</param>
		/// <param name="Icon">Icon of the balloon. Can be None, Warning, Information, Error.</param>
		/// <param name="Parent">Parent control of the balloon. The balloon is pointed at this control.</param>
		public static void Show(String Title, String Text, ToolTipIcon Icon, Control Parent)
		{
			new EditBalloon(Title, Text, Icon, Parent).Show();
		}

		/// <summary>
		/// Show a balloon tooltip for edit control.
		/// </summary>
		public void Show()
		{
			// Check if the version of the OS is major then 5.1 - Windows XP
			if (((Environment.OSVersion.Version.Major == 5 && Environment.OSVersion.Version.Minor >= 1)
				|| Environment.OSVersion.Version.Major > 5 && Environment.OSVersion.Version.Minor >= 0) &&
				_Parent is TextBox)
			{
				// Show the balloon if we are in win xp or higher.
				ShowBalloon();
			}
			else
			{
				// Show the messagebox in minor versions then win xp.
				ShowMessageBox();
			}
		}

		/// <summary>
		/// Shows the balloon. Only called in WinXP or higher.
		/// </summary>
		private void ShowBalloon()
		{
			// Create a new balloon structure and populate it.
			EDITBALLOONTIP ebt = new EDITBALLOONTIP();
			ebt.cbStruct = Marshal.SizeOf(ebt);
			ebt.pszText = _Text;
			ebt.pszTitle = _Title;
			ebt.ttiIcon = (Int32)_TitleIcon;

			// Get the memory for the balloon structure.
			IntPtr ptrStruct = Marshal.AllocHGlobal(Marshal.SizeOf(ebt));
			try
			{
				Marshal.StructureToPtr(ebt, ptrStruct, false);
				// Send the message to show the balloon.
				int ret = SendMessage(_Parent.Handle, EM_SHOWBALLOONTIP, 0, ptrStruct);
			}
			finally
			{
				// Free the structure used for the balloon.
				Marshal.FreeHGlobal(ptrStruct);
			}
		}

		/// <summary>
		/// Show a messagebox if the version of the OS is lower then the WinXP version.
		/// </summary>
		private void ShowMessageBox()
		{
			// Get the right icon for the message box.
			MessageBoxIcon Icon = MessageBoxIcon.Error;
			if (_TitleIcon == ToolTipIcon.Info || _TitleIcon == ToolTipIcon.None)
				Icon = MessageBoxIcon.Information;
			else if (_TitleIcon == ToolTipIcon.Warning)
				Icon = MessageBoxIcon.Warning;

			// Show the message box.
			MessageBox.Show(_Text, _Title, MessageBoxButtons.OK, Icon);

			// Set the focus to the parent control.
			_Parent.Focus();
		}

		/// <summary>
		/// Sets or gets the Title.
		/// </summary>
		public String Title
		{
			get { return _Title; }
			set { _Title = value; }
		}

		/// <summary>
		/// Sets or gets the display icon.
		/// </summary>
		public ToolTipIcon ToolTipIcon
		{
			get { return _TitleIcon; }
			set { _TitleIcon = value; }
		}

		/// <summary>
		/// Sets or gets the display text.
		/// </summary>
		public String Text
		{
			get { return _Text; }
			set { _Text = value; }
		}

		/// <summary>
		/// Sets or gets the parent control of the balloon.
		/// </summary>
		public Control Parent
		{
			get { return _Parent; }
			set { _Parent = value; }
		}
	}
}
