#region Developer's comment header
/* PersistWindowState.cs
 * 
 * author:    Christian Liensberger
 * copyright:  2005 Chili Software. All rights reserved.
 * 
 * history:   21.07.2005, clien: Created.
 *
 */
#endregion
#region Using directives

using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;
using System.Windows.Forms;
using System.Reflection;
using System.Xml;
using System.IO;
using System.Diagnostics;
using System.Drawing;

#endregion

namespace Chili.Windows.Forms
{
	/// <summary>
	/// This class allows to save and restore the state of a form to an xml file.
	/// </summary>
	public class PersistWindowState : Component
	{
		#region XmlFormatManager

		/// <summary>
		/// This class manages the xml format used during the save and load process.
		/// </summary>
		private sealed class XmlFormatManager
		{
			private const string STATEINFO = "stateinfo";
			private const string WINDOW = "window";
			private const string LEFT = "left";
			private const string TOP = "top";
			private const string HEIGHT = "height";
			private const string WIDTH = "width";
			private const string WINDOWSTATE = "windowstate";
			private const string CUSTOM = "custom";
			private const string ITEM = "item";
			private const string NAME = "name";
			private const string VALUE = "value";

			private FormInfo _info;
			private PersistWindowState _state;

			/// <summary>
			/// Creates a new instance of the <see cref="XmlFormatManager">XmlFormatManager Class</see>.
			/// </summary>
			/// <param name="state">The window state itself.</param>
			/// <param name="info">The status information about the form.</param>
			internal XmlFormatManager(PersistWindowState state, FormInfo info)
			{
				_state = state;
				_info = info;
			}

			/// <summary>
			/// Saves the file to the given path.
			/// </summary>
			/// <param name="path">Full path used to save the state.</param>
			internal void SaveToXml(string path)
			{
				if (_state.IsInDesignTime)
					return;

				// Creates the directory.
				string simplePath = System.IO.Path.GetDirectoryName(path);
				Directory.CreateDirectory(simplePath);

				// Get the xml.
				string xml = CompileXml();
				// Create the file and write the xml.
				using (TextWriter writer = File.CreateText(path))
				{
					writer.Write(xml);
				}
			}

			/// <summary>
			/// Restores the file to the given path.
			/// </summary>
			/// <param name="path">Full path used to restore the state.</param>
			internal void RestoreFormXml(string path)
			{
				if (_state.IsInDesignTime)
					return;

				if (!File.Exists(path))
				{
					// Center on the parent form if no information is available.
					_info.Form.SetDesktopLocation(SystemInformation.WorkingArea.Width / 2 - _state.Form.Width / 2,
						SystemInformation.WorkingArea.Height / 2 - _state.Form.Height / 2);
					return;
				}

				XmlDocument xml = new XmlDocument();
				try
				{
					// Load the xml.
					xml.Load(path);
				}
				catch
				{
				} // Ignore exceptions.

				// Restores the window size and position.
				RestoreWindowSizeAndPos(xml);
				// Restores the custom states.
				RestoreCustomStates(xml);
			}

			#region Xml compiler

			/// <summary>
			/// Compiles and returns the xml, which is saved.
			/// </summary>
			private string CompileXml()
			{
				XmlDocument xml = new XmlDocument();
				XmlElement stateInfoElement = xml.CreateElement(STATEINFO);
				xml.AppendChild(stateInfoElement);

				XmlElement windowElement = xml.CreateElement(WINDOW);
				stateInfoElement.AppendChild(windowElement);

				// Add left element and text.
				XmlElement leftElement = xml.CreateElement(LEFT);
				windowElement.AppendChild(leftElement);
				leftElement.AppendChild(xml.CreateTextNode(_info.Location.X.ToString()));

				// Add the top element and text.
				XmlElement topElement = xml.CreateElement(TOP);
				windowElement.AppendChild(topElement);
				topElement.AppendChild(xml.CreateTextNode(_info.Location.Y.ToString()));

				// Check if the form is sizable. Save the height and width only then.
				if (_info.Form.FormBorderStyle == FormBorderStyle.Sizable || _info.Form.FormBorderStyle == FormBorderStyle.SizableToolWindow)
				{
					// Add the height element and text.
					XmlElement heightElement = xml.CreateElement(HEIGHT);
					windowElement.AppendChild(heightElement);
					heightElement.AppendChild(xml.CreateTextNode(_info.Size.Height.ToString()));

					XmlElement widthElement = xml.CreateElement(WIDTH);
					windowElement.AppendChild(widthElement);
					widthElement.AppendChild(xml.CreateTextNode(_info.Size.Width.ToString()));
				}

				// Add the window state to the xml.
				XmlElement windowStateElement = xml.CreateElement(WINDOWSTATE);
				windowElement.AppendChild(windowStateElement);
				windowStateElement.AppendChild(xml.CreateTextNode(_info.Form.WindowState.ToString()));

				// Compiles the custom state information.
				CompileCustomStates(stateInfoElement, xml);

				// Return the compiled xml.
				return xml.OuterXml;
			}

			/// <summary>
			/// Compiles the custom state elements.
			/// </summary>
			/// <param name="root">The root element of the custom states.</param>
			/// <param name="xml">The xml document itself.</param>
			private void CompileCustomStates(XmlElement root, XmlDocument xml)
			{
				XmlElement customElement = xml.CreateElement(CUSTOM);
				root.AppendChild(customElement);

				// Loop over all custom states and create the appropriate nodes.
				foreach (KeyValuePair<string, string> item in _state.CustomStates)
				{
					// Create the item element.
					XmlElement itemElement = xml.CreateElement(ITEM);
					customElement.AppendChild(itemElement);

					// Create the name.
					XmlElement nameElement = xml.CreateElement(NAME);
					itemElement.AppendChild(nameElement);
					nameElement.AppendChild(xml.CreateTextNode(item.Key));

					XmlElement valueElement = xml.CreateElement(VALUE);
					itemElement.AppendChild(valueElement);
					valueElement.AppendChild(xml.CreateCDataSection(item.Value));
				}
			}

			#endregion

			#region Xml parser

			/// <summary>
			/// Restores the custom state information from the xml file.
			/// </summary>
			/// <param name="xml">XmlDocument that contains the data.</param>
			private void RestoreCustomStates(XmlDocument xml)
			{
				// Clear the custom states here.
				_state.CustomStates.Clear();

				XmlNodeList nodeList = xml.SelectNodes(string.Format("//{0}/*", CUSTOM));
				foreach (XmlNode node in nodeList)
				{
					string name = null;
					string value = null;

					// Loop over all values of the current item (in the custom states).
					foreach (XmlNode childNode in node.ChildNodes)
					{
						switch (childNode.Name)
						{
							case NAME:
								name = childNode.InnerText;
								break;
							case VALUE:
								value = childNode.InnerText;
								break;
						}
					}

					if (name == null || value == null)
						throw new NullReferenceException("Name or value of an item in the custom states was null.");

					// Add the name and value to the custom states.
					_state.CustomStates.Add(name, value);
				}
			}

			/// <summary>
			/// Restores the windows position and size.
			/// </summary>
			/// <param name="xml">XmlDocument that contains the position and the size.</param>
			private void RestoreWindowSizeAndPos(XmlDocument xml)
			{
				XmlNodeList nodeList = xml.SelectNodes(string.Format("//{0}/*", WINDOW));
				foreach (XmlNode node in nodeList)
				{
					switch (node.Name)
					{
						case LEFT:
							_info.Form.Location = new Point(int.Parse(node.InnerText), _info.Form.Location.Y);
							break;
						case TOP:
							_info.Form.Location = new Point(_info.Form.Location.X, int.Parse(node.InnerText));
							break;
						case HEIGHT:
							_info.Form.Size = new Size(_info.Form.Size.Width, int.Parse(node.InnerText));
							break;
						case WIDTH:
							_info.Form.Size = new Size(int.Parse(node.InnerText), _info.Form.Size.Height);
							break;
						case WINDOWSTATE:
							{
								FormWindowState state = (FormWindowState)Enum.Parse(typeof(FormWindowState), node.InnerText);
								// Set the window state only if maximized is set.
								if (state == FormWindowState.Maximized)
									_info.Form.WindowState = state;
							}
							break;
					}
				}
			}

			#endregion
		}

		#endregion

		#region FormInfo

		/// <summary>
		/// This class groups the various form information.
		/// </summary>
		private sealed class FormInfo
		{
			private Size _size;
			private Point _location;
			private Form _form;

			/// <summary>
			/// Specifies the size of the form.
			/// </summary>
			public Size Size
			{
				get { return _size; }
				set { _size = value; }
			}

			/// <summary>
			/// Specifies the location of the form.
			/// </summary>
			public Point Location
			{
				get { return _location; }
				set { _location = value; }
			}

			/// <summary>
			/// Specifies the form associated with the component.
			/// </summary>
			public Form Form
			{
				get { return _form; }
				set { _form = value; }
			}
		}

		#endregion

		private const string EXTENSION = "xml";
		private const string APPLICATION_DATA = "%APPLICATION_DATA%";
		private const string COMPANY = "%COMPANY%";
		private const string PRODUCT = "%PRODUCT%";

		private string _path = string.Format(@"{0}\{1}\{2}\", APPLICATION_DATA, COMPANY, PRODUCT);
		private bool _autoSave = true;
		private bool _autoRestore = true;
		private FormInfo _info = new FormInfo();
		private IDictionary<string, string> _customStates = new Dictionary<string, string>();

		#region Events

		/// <summary>
		/// Event is fired after the state has been restored.
		/// </summary>
		public event EventHandler<EventArgs> StateRestored;

		/// <summary>
		/// Event if fired before the state is saved.
		/// </summary>
		public event EventHandler<EventArgs> StateSaving;

		/// <summary>
		/// Invoked by the control to fire the <see cref="StateRestored">StateRestored</see> event.
		/// </summary>
		protected virtual void OnStateRestored(EventArgs e)
		{
			if (StateRestored != null)
				StateRestored(this, e);
		}

		/// <summary>
		/// Invoked by the control to fire the <see cref="StateSaving">StateSaving</see> event.
		/// </summary>
		protected virtual void OnStateSaving(EventArgs e)
		{
			if (StateSaving != null)
				StateSaving(this, e);
		}

		#endregion

		/// <summary>
		/// Returns a dictionary with custom states.
		/// </summary>
		[Browsable(false)]
		public IDictionary<string, string> CustomStates
		{
			get { return _customStates; }
		}

		/// <summary>
		/// Gets or sets the form for the <see cref="PersistWindowState">PersistWindowState</see>. 
		/// </summary>
		[Description("Gets or sets the form for the PersistWindowState.")]
		public Form Form
		{
			get { return _info.Form; }
			set
			{
				// Unregister form closing event.
				if (_info.Form != null)
				{
					_info.Form.FormClosing -= new FormClosingEventHandler(value_FormClosing);
					_info.Form.Load -= new EventHandler(value_Load);
					_info.Form.LocationChanged -= new EventHandler(newForm_LocationChanged);
					_info.Form.SizeChanged -= new EventHandler(newForm_SizeChanged);
				}
				// Register form closing event.
				if (value != null)
				{
					value.FormClosing += new FormClosingEventHandler(value_FormClosing);
					value.Load += new EventHandler(value_Load);
					value.LocationChanged += new EventHandler(newForm_LocationChanged);
					value.SizeChanged += new EventHandler(newForm_SizeChanged);
				}

				_info.Form = value;
			}
		}

		/// <summary>
		/// Invoked each time the form associated with the current <see cref="PersistWindowState">PersistWindowState</see> is closed.
		/// </summary>
		private void value_FormClosing(object sender, FormClosingEventArgs e)
		{
			if (_autoSave)
				SaveToXml();
		}

		/// <summary>
		/// Invoked each time the size of the associated form changed.
		/// </summary>
		private void newForm_SizeChanged(object sender, EventArgs e)
		{
			if (_info.Form.WindowState == FormWindowState.Normal)
				_info.Size = _info.Form.Size;
		}

		/// <summary>
		/// Invoked each time the location of the associated form changed.
		/// </summary>
		private void newForm_LocationChanged(object sender, EventArgs e)
		{
			if (_info.Form.WindowState == FormWindowState.Normal)
				_info.Location = _info.Form.Location;
		}

		/// <summary>
		/// Invoked each time the form associated with the current <see cref="PersistWindowState">PersistWindowState</see> is loaded.
		/// </summary>
		private void value_Load(object sender, EventArgs e)
		{
			if (_autoRestore)
				RestoreFromXml();
		}

		/// <summary>
		/// Gets or sets the path where the state is saved.
		/// </summary>
		[Description("%APPLICATION_DATA% = application data folder; %COMPANY% = company name; %PRODUCT% = product name")]
		public string Path
		{
			get { return _path; }
			set
			{
				if (value == null)
					throw new ArgumentNullException("Path", "Path is null. Path must be a string that represents the path to the xml file.");
				_path = value;
			}
		}

		/// <summary>
		/// Specifies whether to save automatically the state of the window.
		/// </summary>
		[Description("Specifies whether to save automatically the state of the window.")]
		public bool AutoSave
		{
			get { return _autoSave; }
			set { _autoSave = value; }
		}

		/// <summary>
		/// Specifies whether to restore automatically the state of the window.
		/// </summary>
		[Description("Specifies whether to restore automatically the state of the window.")]
		public bool AutoRestore
		{
			get { return _autoRestore; }
			set { _autoRestore = value; }
		}

		/// <summary>
		/// Restores the state from the given xml file.
		/// </summary>
		/// <param name="path">Path of the file. This path is the full path of the file used to restore the state.</param>
		public void RestoreFromXml(string path)
		{
			XmlFormatManager manager = new XmlFormatManager(this, _info);
			manager.RestoreFormXml(path);

			// Invoke the event.
			OnStateRestored(EventArgs.Empty);
		}

		/// <summary>
		/// Restores the window state from xml. Uses the path specifies in the Path property.
		/// </summary>
		public void RestoreFromXml()
		{
			// Only if we are not in design time.
			if (!IsInDesignTime)
				RestoreFromXml(GetFullPath());
		}

		/// <summary>
		/// Saves the state to the given xml file.
		/// </summary>
		/// <param name="path">Path of the file. This path is the full path of the file used to save the state.</param>
		public void SaveToXml(string path)
		{
			// Invoke the event.
			OnStateSaving(EventArgs.Empty);

			XmlFormatManager manager = new XmlFormatManager(this, _info);
			manager.SaveToXml(path);
		}

		/// <summary>
		/// Saves the window state to xml. Uses the path specifies in the Path property.
		/// </summary>
		public void SaveToXml()
		{
			// Only if we are not in design time.
			if (!IsInDesignTime)
				SaveToXml(GetFullPath());
		}

		/// <summary>
		/// Returns the full path for the file used to store the data.
		/// </summary>
		/// <returns></returns>
		private string GetFullPath()
		{
			string path = _path.Replace(APPLICATION_DATA, Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData));
			path = path.Replace(PRODUCT, Assembly.GetEntryAssembly().GetName().Name);

			// Get the company.
			AssemblyCompanyAttribute[] companies = (AssemblyCompanyAttribute[])
				Assembly.GetEntryAssembly().GetCustomAttributes(typeof(AssemblyCompanyAttribute), true);

			if (companies.Length == 0)
				throw new NotImplementedException("Cannot save window state without company information. Add the company information to the AssemblyInfo file.");

			path = path.Replace(COMPANY, companies[0].Company);

			if (_info.Form == null)
				throw new ArgumentNullException("Form", "Form property is null. This property must be set to use save the window state.");

			// Adds the filename to the path and returns the full path.
			return string.Format(@"{0}\{1}.{2}", path, _info.Form.GetType().FullName, EXTENSION);
		}

		/// <summary>
		/// Returns whether the component is in design time.
		/// </summary>
		[Browsable(false)]
		public bool IsInDesignTime
		{
			get
			{
				// HACK: There is no other way to get the design time state.
				return (Assembly.GetEntryAssembly() == null);
			}
		}
	}
}
