//--------------------------------------------------------------------------
// 
//  Copyright (c) Chili Software.  All rights reserved. 
// 
//  File: TypeRelationInfoCollection.cs
//
//  Description: Holds all relations to a type of persistent objects.
// 
//--------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using Chili.Opf3.Relations;

namespace Chili.Opf3.Mapping
{
	/// <summary>
	/// Class that represents a collection of <see cref="RelationInfo">RelationInfo</see> objects.
	/// </summary>
	/// <remarks>
	/// This class represents a collection of <see cref="RelationInfo">RelationInfo</see> objects. This class
	/// contains all relations related with a given type.
	/// </remarks>
	internal sealed class TypeRelationInfoCollection : ITypeRelationInfoCollection
	{
		private Dictionary<string, RelationInfo> _attributeRelationMappings = new Dictionary<string, RelationInfo>();
		private List<IRelationInfo> _relations = new List<IRelationInfo>();
		private Type _type;

		/// <summary>
		/// Creates a new instance of the <see cref="TypeRelationInfoCollection">TypeRelationInfoCollection Class</see>.
		/// </summary>
		/// <param name="type">The type associated with the collection.</param>
		public TypeRelationInfoCollection(Type type)
		{
			if (type == null)
				throw new ArgumentNullException("type");
			_type = type;
		}

		/// <summary>
		/// Adds the <see cref="IRelationInfo">IRelationInfo</see> object to the list.
		/// </summary>
		/// <param name="info">The object that is added.</param>
		internal void Add(RelationInfo info)
		{
			if (info == null)
				throw new ArgumentNullException("info");

			if (!_attributeRelationMappings.ContainsKey(info.RelationAttribute.Name))
			{
				_attributeRelationMappings.Add(info.RelationAttribute.Name, info);
				_relations.Add(info);
			}
		}

		/// <summary>
		/// Returns the type of the <see cref="TypeRelationInfoCollection">TypeRelationInfoCollection</see>.
		/// This type is the type of the related objects.
		/// </summary>
		public Type RelatedType
		{
			get { return _type; }
		}

		/// <summary>
		/// Returns the <see cref="RelationInfo">RelationInfo</see> object associated with the 
		/// <see cref="RelationAttribute">RelationAttribute</see>'s name.
		/// </summary>
		/// <param name="name">The relation's name. The name is specified by setting 
		/// the <see cref="RelationAttribute">RelationAttribute</see>'s name property.</param>
		public IRelationInfo GetRelationMapping(string name)
		{
			if (name == null)
				throw new ArgumentNullException("name");

			RelationInfo info = null;
			_attributeRelationMappings.TryGetValue(name, out info);

			return info;
		}

		/// <summary>
		/// Determines whether an element is in the collection.
		/// </summary>
		/// <param name="item">The object to locate in the collection.</param>
		/// <returns>True if item is found in the collection; otherwise, false.</returns>
		public bool Contains(IRelationInfo item)
		{
			return _relations.Contains(item);
		}

		/// <summary>
		/// Copies the entire collection to a compatible one-dimensional array, starting at the specified index of the target array.
		/// </summary>
		/// <param name="array">The one-dimensional System.Array that is the destination of the elements copied from collection. The System.Array must have zero-based indexing.</param>
		/// <param name="arrayIndex">The zero-based index in array at which copying begins.</param>
		public void CopyTo(IRelationInfo[] array, int arrayIndex)
		{
			_relations.CopyTo(array, arrayIndex);
		}

		/// <summary>
		/// Gets the number of elements actually contained in the collection.
		/// </summary>
		public int Count
		{
			get { return _relations.Count; }
		}

		/// <summary>
		/// Returns whether the collection is read-only.
		/// </summary>
		public bool IsReadOnly
		{
			get { return true; }
		}

		/// <summary>
		/// Returns the item at the given index.
		/// </summary>
		/// <param name="index">The index of the item that is returned.</param>
		/// <returns>The item at the given index.</returns>
		public IRelationInfo this[int index]
		{
			get { return _relations[index]; }
		}

		#region IEnumerable<IMemberInfo> Members

		/// <summary>
		/// Returns an enumerator that iterates through the collection.
		/// </summary>
		/// <returns> A enumerator that can be used to iterate through the collection.</returns>
		public IEnumerator<IRelationInfo> GetEnumerator()
		{
			return _relations.GetEnumerator();
		}

		#endregion

		#region IEnumerable Members

		/// <summary>
		/// Returns an enumerator that iterates through a collection.
		/// </summary>
		/// <returns>An enumerator object that can be used to iterate through the collection.</returns>
		System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
		{
			return ((IEnumerable)_relations).GetEnumerator();
		}

		#endregion
	}
}
