﻿//--------------------------------------------------------------------------
// 
//  Copyright (c) Chili Software.  All rights reserved. 
// 
//  File: ObjectPersistedEventArgs.cs
//
//  Description: EventArgs for ObjectContext events fired after persist.
// 
//--------------------------------------------------------------------------

using System;
using System.Reflection;
using Chili.Opf3.Query;

namespace Chili.Opf3
{
	/// <summary>
	/// EventArgs used in the <see cref="Opf3.ObjectContext">ObjectContext</see> events
	/// that are invoked after the persist process of an object.
	/// </summary>
	/// <remarks>
	/// This class is used by the <see cref="Opf3.ObjectContext">ObjectContext</see> in
	/// events that are invoked after the persist process of an object to send additional event
	/// data. The data contains the <see cref="Opf3.ObjectInfo">ObjectInfo</see> object that
	/// contains further information about the state of the object, the object itself and if the
	/// operation has been performed sucessfully.
	/// </remarks>
	[Serializable]
	public sealed class ObjectPersistedEventArgs : EventArgs
	{
		private ObjectInfo _objectInfo;
		private object _obj;
		private bool _successfully;

		/// <summary>
		/// Creates a new instance of the<see cref="ObjectPersistedEventArgs">ObjectPersistedEventArgs Class</see>.
		/// </summary>
		/// <param name="obj">The object that is affected by the current operation. This object is persisted to the storage. </param>
		/// <param name="objectInfo"><see cref="Opf3.ObjectInfo">ObjectInfo</see> object that contains state information about the persistent object.</param>
		/// <param name="successfully">Specifies whether the operation was successfully.</param>
		/// <exception cref="System.ArgumentNullException">Query is null, obj is null or objectInfo is null.</exception>
		internal ObjectPersistedEventArgs(object obj, ObjectInfo objectInfo, bool successfully)
		{
			if (obj == null)
				throw new ArgumentNullException("obj");
			if (objectInfo == null)
				throw new ArgumentNullException("objectInfo");

			_objectInfo = objectInfo;
			_obj = obj;
			_successfully = successfully;
		}

		/// <summary>
		/// Returns the <see cref="Opf3.ObjectInfo">ObjectInfo</see> object that 
		/// contains state information about the persistent object.
		/// </summary>
		/// <value>
		/// The <see cref="ObjectInfo">ObjectInfo</see> object that contains state
		/// information about the persistent object.
		/// </value>
		public ObjectInfo ObjectInfo
		{
			get { return _objectInfo; }
		}

		/// <summary>
		/// Returns the object that is affected by the current operation. This object is
		/// persisted to the storage.
		/// </summary>
		/// <value>The object affected by the current operation.</value>
		public object Object
		{
			get { return _obj; }
		}

		/// <summary>
		/// Returns whether the operation has been performed successfully.
		/// </summary>
		/// <value>
		/// True if the operation was successfully, otherwise false.
		/// </value>
		public bool Successfully
		{
			get { return _successfully; }
		}
	}
}
