﻿//--------------------------------------------------------------------------
// 
//  Copyright (c) Chili Software.  All rights reserved. 
// 
//  File: ObjectPersistingEventArgs.cs
//
//  Description: EventArgs for ObjectContext events fired before persist.
// 
//--------------------------------------------------------------------------

using System;
using System.Reflection;
using Chili.Opf3.Query;

namespace Chili.Opf3
{
	/// <summary>
	/// EventArgs used in the <see cref="Opf3.ObjectContext">ObjectContext</see> events
	/// that are invoked before the persist process of an object.
	/// </summary>
	/// <remarks>
	/// This class is used by the <see cref="Opf3.ObjectContext">ObjectContext</see> in
	/// events that are invoked before the persist process of an object to send additional event
	/// data. The data contains the <see cref="Opf3.ObjectInfo">ObjectInfo</see> object that
	/// contains further information about the state of the object, the object itself and the
	/// <see cref="Opf3.Query.IQuery">IQuery</see> object that represents the query of the
	/// current operation.
	/// </remarks>
	[Serializable]
	public sealed class ObjectPersistingEventArgs : EventArgs
	{
		private ObjectInfo _objectInfo;
		private object _obj;
		private IQuery _query;

		/// <summary>
		/// Creates a new instance of the<see cref="ObjectPersistingEventArgs">ObjectPersistingEventArgs Class</see>.
		/// </summary>
		/// <param name="query"><see cref="Opf3.Query.IQuery">IQuery</see> used by for the current operation.</param>
		/// <param name="obj">The object that is affected by the current operation. This object is persisted to the storage. </param>
		/// <param name="objectInfo"><see cref="Opf3.ObjectInfo">ObjectInfo</see> object that contains state information about the persistent object.</param>
		/// <exception cref="System.ArgumentNullException">Query is null, obj is null or objectInfo is null.</exception>
		internal ObjectPersistingEventArgs(IQuery query, object obj, ObjectInfo objectInfo)
		{
			if (query == null)
				throw new ArgumentNullException("query");
			if (obj == null)
				throw new ArgumentNullException("obj");
			if (objectInfo == null)
				throw new ArgumentNullException("objectInfo");

			_objectInfo = objectInfo;
			_obj = obj;
			_query = query;
		}

		/// <summary>
		/// Returns the <see cref="Opf3.ObjectInfo">ObjectInfo</see> object that 
		/// contains state information about the persistent object.
		/// </summary>
		/// <value>
		/// The <see cref="ObjectInfo">ObjectInfo</see> object that contains state
		/// information about the persistent object.
		/// </value>
		public ObjectInfo ObjectInfo
		{
			get { return _objectInfo; }
		}

		/// <summary>
		/// Returns the object that is affected by the current operation. This object is
		/// persisted to the storage.
		/// </summary>
		/// <value>The object affected by the current operation.</value>
		public object Object
		{
			get { return _obj; }
		}

		/// <summary>
		/// Specifies the <see cref="Opf3.Query.IQuery">IQuery</see> of the current
		/// operation.
		/// </summary>
		/// <value>
		/// The <see cref="Opf3.Query.IQuery">IQuery</see> of the current
		/// operation.<see cref="ObjectPersistingEventArgs">ObjectPersistingEventArgs
		/// Class</see><see cref="ObjectNotificationEventArgs">ObjectNotificationEventArgs
		/// Class</see>
		/// </value>
		public IQuery Query
		{
			get { return _query; }
			set { _query = value; }
		}
	}
}
