﻿//--------------------------------------------------------------------------
// 
//  Copyright (c) Chili Software.  All rights reserved. 
// 
//  File: ObjectSetHolder_1.cs
//
//  Description: Holds an ObjectSet of releated persistent objects.
// 
//--------------------------------------------------------------------------

using System;
using System.Collections;
using System.Collections.Generic;
using System.Reflection;
using System.Runtime.Serialization;
using System.Xml.Serialization;
using Chili.Opf3.Query;
using Chili.Opf3.Query.OPathParser;

namespace Chili.Opf3.Relations
{
	/// <summary>
	/// Holds an <see cref="Opf3.ObjectSet{T}">ObjectSet</see> of objects that are connected with the parent object.
	/// </summary>
	/// <remarks>
	/// This class is used in a persistent class to hold an <see cref="Opf3.ObjectSet{T}">ObjectSet</see> of objects 
	/// that are connected with the parent object. When accessing the first time the 
	/// <see cref="Opf3.Relations.ObjectHolder{T}.InnerObject">InnerObject</see> property the related
	/// objects are loaded. This is called delayed loading and reduces roundtrips to the storage.
	/// </remarks>
	/// <example>
	/// The following example shows how to use the ObjectSetHolder in your persistent objects. In the
	/// example each User has associated Calls.
	/// <code lang="CS">
	/// [Persistent("USER", PoolSize = 50)]
	/// public class User
	/// {
	///		// The class contains an ObjectSetHolder and a RelationAttribute
	///		// that specifies the relation between the parent object and the 
	///		// related objects. It demonstrates also how to use a condition to 
	///		// delimit the items in the list and to sort the result.
	///		[Relation("Id", "UserId")]
	///		private ObjectSetHolder&lt;Call&gt; _calls = 
	///			new ObjectSetHolder&lt;Call&gt;("Duration &lt; {0} SortBy Duration Asc", "50");
	///	
	///		public ObjectSet&lt;Call&gt; Calls
	///		{
	///			get { return _calls.InnerObject; }
	///			set { _calls.InnerObject = value; }
	///		}
	///		
	///		// ... Other properties and code.
	/// }
	/// </code>
	/// </example>
	[Serializable]
	public sealed class ObjectSetHolder<T> : ObjectHolderBase
	{
		/// <summary>
		/// Creates a new instance of the <see cref="ObjectSetHolder{T}">ObjectSetHolder Class</see>.
		/// </summary>
		public ObjectSetHolder() : base()
		{
		}

		/// <summary>
		/// Creates a new instance of the <see cref="ObjectSetHolder{T}">ObjectSetHolder Class</see>.
		/// </summary>
		/// <param name="conditions">OPath that is used when retrieving the objects. For more information about OPath
		/// look at the <see cref="Opf3.Query.ObjectQuery{T}">ObjectQuery</see> class.</param>
		/// <param name="parameters">A list of parameters that are used in the OPath expression.</param>
		public ObjectSetHolder(string conditions, params object[] parameters) : base(new ObjectQuery<T>(conditions, parameters))
		{
		}

		/// <summary>
		/// Creates a new instance of the <see cref="ObjectSetHolder{T}">ObjectSetHolder Class</see>.
		/// </summary>
		/// <param name="query"><see cref="Opf3.Query.IQuery">IQuery</see> that is executed to load the associated 
		/// objects. This parameter allows you to use any kind of query without using the build in functionality.</param>
		public ObjectSetHolder(IQuery query) : base(query)
		{
		}

		/// <summary>
		/// Creates a new instance of the <see cref="ObjectSetHolder{T}">ObjectSetHolder Class</see>.
		/// </summary>
		/// <param name="info"></param>
		/// <param name="context"></param>
		private ObjectSetHolder(SerializationInfo info, StreamingContext context) : base(info, context)
		{
		}

		/// <summary>
		/// Event is fired before the data from the storage is used to populate an object.
		/// </summary>
		/// <remarks>
		/// This event is fired before the data from the storage is used to populate an object. It allows the user
		/// to specify another type for the row (item) that is populated with the data. It allows also to have 
		/// certain items ignored. This event can be used in certain scenarios of inheritance.
		/// </remarks>
		/// <example>
		/// The following example introduces to the use of the FecthingStorageRecord event. 
		/// <code lang="CS">
		/// [Persistent("USER", PoolSize = 50)]
		/// public class User
		/// {
		///		// The class contains an ObjectSetHolder and a RelationAttribute
		///		// that specifies the relation between the parent object and the 
		///		// related objects. It demonstrates also how to use a condition to 
		///		// delimit the items in the list and to sort the result.
		///		[Relation("Id", "UserId")]
		///		private ObjectSetHolder&lt;Call&gt; _calls = 
		///			new ObjectSetHolder&lt;Call&gt;("Duration &lt; {0} SortBy Duration Asc", "50");
		///	
		///     public User()
		///     {
		///         _calls.FetchingStorageRecord += PersistentTypeSelector(MySelection);
		///     }
		/// 
		///     private void MySelection(object sender, FetchingStorageRecordEventArgs e)
		///     {
		///        if (e.DataRecord["Type"] == "1")
		///        {
		///            // Switch the type of the persistent that is created to User1.
		///	           e.Type = typeof(User1);
		///        }
		///        else if ((bool)e.DataRecord["Ignore"])
		///        {
		///            // Tell the framework to ignore this item.
		///            e.Ignore = true;
		///        }
		///        // Have a persistent of type User created.
		///        e.Type = typeof(User);
		///     }
		///
		///		public ObjectSet&lt;Call&gt; Calls
		///		{
		///			get { return _calls.InnerObject; }
		///			set { _calls.InnerObject = value; }
		///		}
		///		
		///		// ... Other properties and code.
		/// }
		/// 
		/// // ...
		/// 
		/// ObjectSet&lt;User&gt; objectSet = context.GetObjectSet&lt;User&gt;();
		/// foreach(User user in objectSet)
		/// {
		///     foreach(Call call in user.Calls)
		///     {
		///         if (call.GetType() == typeof(Call1))
		///             Console.WriteLine("Call1 found.");
		///         else
		///             Console.WriteLine("Call found.");
		///     }
		/// }
		/// 
		/// </code>
		/// </example>
		public event PersistentTypeSelector FetchingStorageRecord;

		/// <summary>
		/// Invokes the FetchingStorageRecord event.
		/// </summary>
		private void OnFetchingStorageRecord(object sender, FetchingStorageRecordEventArgs e)
		{
			if (FetchingStorageRecord != null)
				FetchingStorageRecord(this, e);
		}

		/// <summary>
		/// Gets and sets the list of objects that are related to the parent object.
		/// </summary>
		/// <value>A <see cref="Opf3.ObjectSet{T}">ObjectSet</see> that contains the related persistent objects.</value>
		[XmlIgnore]
		public ObjectSet<T> Value
		{
			get { return this.InnerObject; }
			set { this.InnerObject = value; }
		}

		/// <summary>
		/// Depreciated: Use the Value property instead!
		/// Gets and sets the list of objects that are related to the parent object.
		/// </summary>
		/// <value>A <see cref="Opf3.ObjectSet{T}">ObjectSet</see> that contains the related persistent objects.</value>
		[XmlIgnore]
		public ObjectSet<T> InnerObject
		{
			get
			{
				if (!AutomaticLoad)
					return (ObjectSet<T>)this.ManagedObject;

				// Get the ObjectSet if not loaded.
				if (!IsLoaded)
				{
					InnerLoad(this.ObjectContext);

					Tracing.Current.TraceStorageIO("Loaded persistent's related objects from storage.");
					// Set the ObjectSetHolder as loaded.
					IsLoaded = true;
				}

				return (ObjectSet<T>)this.ManagedObject;
			}
			set 
			{
				this.ManagedObject = value;
				// Set the ObjectSetHolder as loaded if not null is set.
				if (value != null)
					IsLoaded = true;
			}
		}

		/// <summary>
		/// Loads the related object(s) from the storage.
		/// </summary>
		protected override void InnerLoad(ObjectContext context)
		{
			// Check if the parent implements the IRelationsPopulateHelper.
			IRelationsPopulateHelper helper = this.Parent as IRelationsPopulateHelper;
			// Check if the relation is handled by the parent.
			if (helper != null)
			{
				// Create the relation information.
				RelationInformation info = new RelationInformation(this.Parent, typeof(T), this.RelationAttribute, context);
				if (helper.RelationIsHandled(this.RelationAttribute.Name, info))
				{
					// Get the object from the parent if requested.
					this.ManagedObject = helper.GetRelatedObject(this.RelationAttribute.Name, info);
					return;
				}
			}

			// Creates the query.
			IQuery query = ObjectHolderUtility.PrepareQuery<T>(this.RelationAttribute, this.Query, this.Parent);
			if (query != null && context != null)
			{
				// This check is here for remoting. On some privilege level the delegate can't be passed to the server.
				// If the delegate is null it is not passed ot the server.
				if (FetchingStorageRecord == null)
					this.ManagedObject = context.GetObjectSet<T>(query);
				else
					this.ManagedObject = context.GetObjectSet<T>(new PersistentTypeSelector(OnFetchingStorageRecord), query);
			}
			else
				this.ManagedObject = new ObjectSet<T>();
		}

		/// <summary>
		/// This method is invoked to persist the changes on the objects in the ObjectHolder.
		/// </summary>
		/// <param name="context"><see cref="Opf3.ObjectContext">ObjectContext</see> used to persist 
		/// the changes. It is the same <see cref="Opf3.ObjectContext">ObjectContext</see> as used by the
		/// parent object.</param>
		/// <param name="persistDepth">Specifies the depth of the loaded object graph that is persisted.</param>
		internal override void PersistChanges(ObjectContext context, PersistDepths persistDepth)
		{
			this.ObjectContext = context;

			// Persist the object using the internal context.
			if (this.ManagedObject == null)
				return;

			// Return if the persist depth is flat.
			if (PersistDepths.Flat == persistDepth)
				return;

			bool isMany2Many = (this.RelationAttribute.WeakPersistent != null);
			if (isMany2Many)
				ObjectHolderUtility.DeleteManyToMany<T>((ObjectSet<T>)this.ManagedObject, this.RelationAttribute, this.Parent, context);

			// Set to flat since there is only one level left.
			if (persistDepth == PersistDepths.FirstLevel)
				context.PersistChanges(this.ManagedObject, PersistDepths.Flat);
			else
				context.PersistChanges(this.ManagedObject, PersistDepths.FullObjectGraph);

			// Save only if the parent object has not been deleted.
			if (isMany2Many && context.GetObjectInfo(Parent).ObjectState != ObjectStates.Deleted)
				ObjectHolderUtility.SaveManyToMany<T>((ObjectSet<T>)this.ManagedObject, this.RelationAttribute, this.Parent, context);
		}

		/// <summary>
		/// This method connects the foreign-key properties in the child persistent objects with the identifier
		/// properties of the parent object.
		/// </summary>
		internal override void ConnectPropertiesInChildObjects()
		{
			if (this.RelationAttribute.WeakPersistent != null)
				return;
			if (this.ManagedObject == null)
				return;

			ObjectSet<T> os = (ObjectSet<T>)this.ManagedObject;
			foreach (T obj in os)
			{
				// Updates the connected property.
				UpdateConnectedProperty(obj);
			}
		}

		/// <summary>
		/// This method is invoked to check if any of the ObjectHolder's objects have been modified.
		/// </summary>
		/// <param name="context"><see cref="Opf3.ObjectContext">ObjectContext</see> used to check for modified 
		/// objects. It is the same <see cref="Opf3.ObjectContext">ObjectContext</see> as used by the
		/// parent object.</param>
		/// <returns>True if an object has been modified; false othewise.</returns>
		internal override bool HasModifiedObjects(ObjectContext context)
		{
			if (this.ManagedObject == null)
				return false;

			ObjectSet<T> os = (ObjectSet<T>)this.ManagedObject;
			// loop over all objects and check them for being modified.
			foreach (T obj in os)
			{
				if (context.IsModified(obj, true))
					return true;
			}

			return false;
		}

		/// <summary>
		/// This method is invoked to check if the <see cref="ObjectSetHolder{T}" /> has removed objects.
		/// </summary>
		/// <returns>True if an object has been removed; false otherwise</returns>
		internal override bool HasRemovedObjects()
		{
			if (this.ManagedObject == null)
				return false;

			ObjectSet<T> os = (ObjectSet<T>)this.ManagedObject;
			return (os.RemovedList.Count > 0);
		}

		/// <summary>
		/// <para>
		/// Invalidates the current instance. The next request of the InnerObject property reloads all the 
		/// related objects. This is only true if the automatic load of the related objects has been enabled. 
		/// Otherwise the reload method needs to be invoked (after calling Invalidate) to do a manual reload.
		/// </para>
		/// <para>
		/// The call of this method removes the objects also from the identity map if one is in place. This makes
		/// sure that the objects are really reloaded from the storage and not from the identity map again.
		/// </para>
		/// </summary>
		public override void Invalidate()
		{
			// set the inner object back to null.
			this.ManagedObject = null;
			base.Invalidate();
		}

		/// <summary>
		/// Gets the enumerator for the given loaded related data. This method doesn't load
		/// data from the storage but rather returns the currently loaded data.
		/// </summary>
		internal override IEnumerable GetRawEnumerator()
		{
			if (this.ManagedObject == null)
				yield break;

			ObjectSet<T> os = (ObjectSet<T>)this.ManagedObject;
			foreach (object obj in os)
			{
				yield return obj;
			}
		}
	}
}
